/**
 * @file common.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created April 20, 2011
 * @modified June 18, 2011
 * @version 0.28
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include "parms.h"
#include <errno.h>
#include <error.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

const char info_category[] = "Info";
const char warning_category[] = "Warning";
const char error_category[] = "Error";
const char unknown_category[] = "?";
//
// from libev/ev.c
//
#ifdef __linux
# include <sys/utsname.h>
#endif

unsigned int
ev_linux_version (void) {
#ifdef __linux
  unsigned int v = 0;
  struct utsname buf;
  int i;
  char *p = buf.release;

  if (uname (&buf))
    return 0;

  for (i = 3+1; --i; )
    {
      unsigned int c = 0;

      for (;;)
        {
          if (*p >= '0' && *p <= '9')
            c = c * 10 + *p++ - '0';
          else
            {
              p += *p == '.';
              break;
            }
        }

      v = (v << 8) | c;
    }

  return v;
#else
  return 0;
#endif
}
//
/* clipped from ping.c */
//
/* Copyright (c)1987 Regents of the University of California.
* All rights reserved.
*
* Redistribution and use in source and binary forms are permitted
* provided that the above copyright notice and this paragraph are
* dupliated in all such forms and that any documentation, advertising
* materials, and other materials related to such distribution and use
* acknowledge that the software was developed by the University of
* California, Berkeley. The name of the University may not be used
* to endorse or promote products derived from this software without
* specific prior written permission. THIS SOFTWARE IS PROVIDED ``AS
* IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
* WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHATIBILITY AND
* FITNESS FOR A PARTICULAR PURPOSE
*/
//
unsigned short
in_cksum(unsigned short *ptr, int nbytes) {
    register long sum;		/* assumes long == 32 bits */
    u_short oddbyte;
    register u_short answer;		/* assumes u_short == 16 bits */
    //
    /*
     * Our algorithm is simple, using a 32-bit accumulator (sum),
     * we add sequential 16-bit words to it, and at the end, fold back
     * all the carry bits from the top 16 bits into the lower 16 bits.
     */
    //
    sum = 0;
    while (nbytes > 1)  {
        sum += *ptr++;
	nbytes -= 2;
    }
    //
    /* mop up an odd byte, if necessary */
    //
    if (nbytes == 1) {
        oddbyte = 0;		/* make sure top half is zero */
	*((u_char *) &oddbyte) = *(u_char *)ptr;   /* one byte only */
	sum += oddbyte;
    }
    //
    /*
     * Add back carry outs from top 16 bits to low 16 bits.
     */
    //
    sum  = (sum >> 16) + (sum & 0xffff);	/* add high-16 to low-16 */
    sum += (sum >> 16);			/* add carry */
    answer = ~sum;		/* ones-complement, then truncate to 16 bits */
    return(answer);
}
//
// logit() function
//
void
logit(int msg_category, char * msg, int options)
{
    char timess[96];
    char category_label[16];
    //
    set_time_snapshot(timess, TIME_FORMAT);
    switch (msg_category) {
        case _LOG_INFO:
            strcpy(category_label, info_category);
            break;
        case _LOG_WARNING:
            strcpy(category_label, warning_category);
            break;
        case _LOG_ERROR:
            strcpy(category_label, error_category);
            break;
        default:
            strcpy(category_label, unknown_category);
            break;
    }
    printf("%s: %s %s\n", category_label, timess, msg);
    if (get_uselogfile() && get_logfilename()) {
        FILE *plog = fopen(get_logfilename(), "a");
        if (plog != NULL) {
            logfileit(msg_category, msg, plog, options);
            fclose(plog);
        }
    }
}
//
// logfileit() function
//
void
logfileit(int msg_category, char *msg, FILE *pfile, int options)
{
    char timess[96];
    char category_label[16];
    char str[MAX_LOG_ENTRY_LEN];
    memset(str, 0x0, sizeof(str));
    //
    set_time_snapshot(timess, TIME_FORMAT);
    switch (msg_category) {
        case _LOG_INFO:
            strcpy(category_label, info_category);
            break;
        case _LOG_WARNING:
            strcpy(category_label, warning_category);
            break;
        case _LOG_ERROR:
            strcpy(category_label, error_category);
            break;
        default:
            strcpy(category_label, unknown_category);
            break;
    }
    sprintf(str, "%s: %s %s\n", category_label, timess, msg);
    fputs(str, pfile);
    if (options == 1) {
        fflush(pfile);
    }
}
//
// convert count inchar characters to outchar hex character representation
//
void
set_char_to_hex(char * inchar, char * outchar, int countchar, int prefix) {
    char onehex[3] = "";
    int i, j;
    memset(onehex, '\0', 3);
    if (prefix == 0) {
        // no prefix
        j=0;
    } else {
        // add 0x prefix to generated hex char representation
        outchar[0] = '0';
        outchar[1] = 'x';
        j=2;
    }
    for(i=0; i < countchar; i++) {
        sprintf(onehex, "%02X", inchar[i]);
        outchar[j++] = onehex[0];
        outchar[j++] = onehex[1];
    }
    outchar[j] = '\0';
}
//
// convert inint integer to outchar hex character representation
//
void
set_int_to_hex(int inint, char * outchar, int prefix) {
    char onehex[3] = "";
    int j;
    memset(onehex, '\0', 3);
    if (prefix == 0) {
        // no prefix
        j=0;
    } else {
        // add 0x prefix to generated hex char representation
        outchar[0] = '0';
        outchar[1] = 'x';
        j=2;
    }
    sprintf(onehex, "%02X", (char)inint);
    outchar[j++] = onehex[0];
    outchar[j++] = onehex[1];
    outchar[j] = '\0';
}
//
// get_time_snapshot() function
//
void
get_time_snapshot(char * result, int format, struct timeval mtime, struct tm * Tmx)
{
    time_t ltimex = time(NULL);
    gettimeofday(&mtime, NULL);
    localtime_r(&ltimex, Tmx);
    char * resultx = (char *)asctime(Tmx);
    if (strlen(resultx) > 0) {
        resultx[strlen(resultx)-1] = 0; // remove trailing \n
    }
    if (format == 1) {
        sprintf(result, "%s", resultx);
    } else if (format == 2) {
        sprintf(result, "%d %d %d %d:%d:%d",
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday,
            Tmx->tm_hour,
            Tmx->tm_min,
            Tmx->tm_sec);
    } else if (format == 3) {
        sprintf(result, "%d %d %d %d:%d:%d.%d",
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday,
            Tmx->tm_hour,
            Tmx->tm_min,
            Tmx->tm_sec,
            (int)mtime.tv_usec);
    } else if (format == 4) {
        sprintf(result, "%d %d %d",
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday);
    } else {
        sprintf(result, "%s %d %d %d %d:%d:%d.%d",
            resultx,
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday,
            Tmx->tm_hour,
            Tmx->tm_min,
            Tmx->tm_sec,
            (int)mtime.tv_usec);
    }
    return;
}
//
// set_time_snapshot() function
//
void
set_time_snapshot(char * result, int format)
{
    time_t ltimex;
    //struct tm *Tmx;
    struct tm *Tmx = malloc(sizeof(struct tm));
    struct timeval micro_time;
    //
    ltimex = time(NULL);
    gettimeofday(&micro_time, NULL);
    //Tmx = (struct tm *)localtime_r(&ltimex, (struct tm *)Tmx2);
    localtime_r(&ltimex, Tmx);
    char * resultx = (char *)asctime(Tmx);
    if (strlen(resultx) > 0) {
        resultx[strlen(resultx)-1] = 0; // remove trailing \n
    }
    if (format == 1) {
        sprintf(result, "%s", resultx);
    } else if (format == 2) {
        sprintf(result, "%d %d %d %d:%d:%d",
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday,
            Tmx->tm_hour,
            Tmx->tm_min,
            Tmx->tm_sec);
    } else if (format == 3) {
        sprintf(result, "%d %d %d %d:%d:%d.%d",
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday,
            Tmx->tm_hour,
            Tmx->tm_min,
            Tmx->tm_sec,
            (int)micro_time.tv_usec);
    } else if (format == 4) {
        sprintf(result, "%d %d %d",
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday);
    } else {
        sprintf(result, "%s %d %d %d %d:%d:%d.%d",
            resultx,
            Tmx->tm_year+1900,
            Tmx->tm_mon+1,
            Tmx->tm_mday,
            Tmx->tm_hour,
            Tmx->tm_min,
            Tmx->tm_sec,
            (int)micro_time.tv_usec);
    }
    free(Tmx);
    Tmx = NULL;
}
//
/* Run en external command returning exit status, and optionally filling
 * provided  buffer with STDOUT output up to the size provided.
 *
 * Note: XXX: We are not using the timeout parameter at present. We still need
 *       to implement a reliable timeout mechanism.
*/
int
run_extcmd_inviz(uid_t user_uid, char *cmd, char *so_buf, size_t so_buf_sz, int timeout) {
    FILE   *ipt;
    int     retval = 0;
    char    so_read_buf[IO_READ_BUF_LEN];
    char    logmsg[512];
    pid_t   pid;

    if(so_buf == NULL) {

        /* Since we do not have to capture output, we will fork here (which we
         * would have to do anyway if we are running as another user as well).
        */
        pid = fork();
        if(pid == -1) {
            sprintf(logmsg, "unable to fork to exec command [%s]", strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return(EXTCMD_FORK_ERROR);
        } else if (pid == 0) {
            /* We are the child */
            /* If user is not null, then we setuid to that user before running the
            * command.
            */
            if(user_uid > 0) {
                if(setuid(user_uid) < 0) {
                    exit(EXTCMD_SETUID_ERROR);
                }
            }
            exit(WEXITSTATUS(system(cmd)));
        }

        /* Retval is forced to 0 as we don't care about the exit status of
         * the child (for now)>
        */
        retval = 0;
    } else {
        /* Looking for output use popen and fill the buffer to its limit.
        */
        ipt = popen(cmd, "r");

        if(ipt == NULL) {
            sprintf(logmsg, "unable to popen() to exec command [%s]", strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            retval = -1;
        } else {
            memset(so_buf, 0x0, so_buf_sz);

            while((fgets(so_read_buf, IO_READ_BUF_LEN, ipt)) != NULL) {
                if(so_buf == NULL) {
                    continue;
                }
                strlcat(so_buf, so_read_buf, so_buf_sz);

                if(strlen(so_buf) >= so_buf_sz-1) {
                    break;
                }
            }
            pclose(ipt);
        }
    }
    return(retval);
}
//
// erase and wipe file from disc wipetimes
//
int
wipefile(const char *fname, int wipetimes) {
    int ires, itimes;
    long fsize;
    FILE *fd;
    char logmsg[512];
    fd = fopen(fname, "rb"); // read, binary
    if (fd == NULL) {
        sprintf(logmsg, "unable to open file (%s) to be wiped [%s]", fname, strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        return ( EXIT_FAILURE );
    }
    fseek(fd, 0, SEEK_END);
    if (ferror(fd)) {
        sprintf(logmsg, "unable to read file (%s) to be wiped [%s]", fname, strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        fclose(fd);
        return ( EXIT_FAILURE );
    } else {
        fsize = ftell(fd);
    }
    fclose(fd);
    if (wipetimes < MIN_WIPE_FILE_TIMES || wipetimes > MAX_WIPE_FILE_TIMES) {
        sprintf(logmsg, "unable to wipe file %s %i times", (char *)fname, wipetimes);
        logit(_LOG_ERROR, logmsg, 0);
        return ( EXIT_FAILURE );
    }
    int shredit;
    snprintf(cmd_buf, CMD_BUFSIZE-1, "which shred");
    ires = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);
    if(!EXTCMD_IS_SUCCESS(ires)) {
        //
        // try this then
        //
        fd = fopen("/usr/bin/shred", "rb");
        shredit = (fd != NULL);
    } else {
        //
        // read outbuf buffer
        //
        cmd_out[STANDARD_CMD_OUT_BUFSIZE] = '\0';
        char *ndx = strchr(cmd_out, '\n');
        if (ndx != NULL) {
            cmd_out[ndx-cmd_out] = '\0';
        }
        fd = fopen(cmd_out, "rb");
        shredit = (fd != NULL);
    }
    if (shredit) {
        if (fd != NULL) {
            fclose(fd);
            fd = NULL;
        }
        if (get_verbose()) {
            sprintf(logmsg, "starting shred (wipe) of %.0lf byte %s file...",
                    (double)fsize,
                    (char *)fname);
            logit(_LOG_INFO, logmsg, 0);
        }

        // add -u to remove after wiping
        // add -z to do final wipe with zeroes
        snprintf(cmd_buf, CMD_BUFSIZE-1, "shred -f --iterations=%i --random-source=%s --size=%.0lf -u %s 2>&1",
                wipetimes, (char *)get_devrandom(), (double)fsize, (char *)fname);
        if (get_verbose() > 1) {
            sprintf(logmsg, "%s", cmd_buf);
            logit(_LOG_INFO, logmsg, 0);
        }
        ires = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);
        if(!EXTCMD_IS_SUCCESS(ires)) {
            if (get_verbose()) {
                sprintf(logmsg, "unable to exec %s [%i]", cmd_buf, ires);
                logit(_LOG_ERROR, logmsg, 0);
            }
            return (EXIT_FAILURE);
        }
        fd = fopen(fname, "rb"); // read, binary
        if (fd == NULL) {
            //
            // wiped successfully
            //
            return (EXIT_SUCCESS);
        } else {
            fclose(fd);
            fd = NULL;
        }
    }
    //
    int ddit;
    snprintf(cmd_buf, CMD_BUFSIZE-1, "which dd");
    ires = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);
    if(!EXTCMD_IS_SUCCESS(ires)) {
        //
        // try this then
        //
        fd = fopen("bin/dd", "rb");
        ddit = (fd != NULL);
    } else {
        //
        // read outbuf buffer
        //
        cmd_out[STANDARD_CMD_OUT_BUFSIZE] = '\0';
        char *ndx = strchr(cmd_out, '\n');
        if (ndx != NULL) {
            cmd_out[ndx-cmd_out] = '\0';
        }
        fd = fopen(cmd_out, "rb");
        ddit = (fd != NULL);
    }
    if (!ddit) {
        sprintf(logmsg, "unable to wipe file %s with either shred or dd", (char *)fname);
        logit(_LOG_ERROR, logmsg, 0);
        return (EXIT_FAILURE);
    }
    if (fd != NULL) {
        fclose(fd);
        fd = NULL;
    }
    if (get_verbose()) {
        sprintf(logmsg, "starting dd (wipe) of %.0lf byte %s file...",
                (double)fsize, (char *)fname);
        logit(_LOG_INFO, logmsg, 0);
    }
    for (itimes=0; itimes<wipetimes; itimes++) {
        snprintf(cmd_buf, CMD_BUFSIZE-1, "dd if=%s of=%s bs=%.0lf count=1 conv=notrunc 2>&1",
                (char *)get_devrandom(), (char *)fname, (double)fsize);
        if (get_verbose() > 1) {
            sprintf(logmsg, "%s", cmd_buf);
            logit(_LOG_INFO, logmsg, 0);
        }
        ires = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);
        if(!EXTCMD_IS_SUCCESS(ires)) {
            if (get_verbose()) {
                sprintf(logmsg, "unable to exec %s [%i]", cmd_buf, ires);
                logit(_LOG_ERROR, logmsg, 0);
            }
            return (EXIT_FAILURE);
        }
    }
    if (remove(fname) != 0) {
        if (get_verbose()) {
            sprintf(logmsg, "unable to delete file %s after wiping it [%s]", fname, strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
        }
        return (EXIT_FAILURE);
    }
    fd = fopen(fname, "rb"); // read, binary
    if (fd == NULL) {
        //
        // wiped successfully
        //
        return (EXIT_SUCCESS);
    } else {
        fclose(fd);
        fd = NULL;
    }
    return (EXIT_FAILURE);
}

//
/* Generic resolver (for ip4 address or host name) from unknown source */
//
unsigned int
host_convert_ip4(char *hostname) {
    static struct in_addr i;
    struct hostent *h;
    i.s_addr = inet_addr(hostname);
    if(i.s_addr == -1) {
        h = gethostbyname(hostname);
        if(h == NULL) {
            char logmsg[512];
            sprintf(logmsg, "cannot resolve hostname %s", hostname);
            logit(_LOG_ERROR, logmsg, 0);
            return( EXIT_FAILURE );
        }
        bcopy(h->h_addr, (char *)&i.s_addr, h->h_length);
    }
    return i.s_addr;
}
//
// courtesty of libevent2/evutil.c
//
const unsigned char UTIL_TOUPPER_TABLE[256] = {
  0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
  16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
  32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,
  48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
  64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,
  80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
  96,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,
  80,81,82,83,84,85,86,87,88,89,90,123,124,125,126,127,
  128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
  144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
  160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
  176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
  192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
  208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
  224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
  240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,
};
const unsigned char UTIL_TOLOWER_TABLE[256] = {
  0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
  16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
  32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,
  48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
  64,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
  112,113,114,115,116,117,118,119,120,121,122,91,92,93,94,95,
  96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
  112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
  128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
  144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
  160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
  176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
  192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
  208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
  224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
  240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,
};
//
// courtesty of libevent2/evutil.c
//
char
UTIL_TOLOWER(char c)
{
    return ((char)UTIL_TOLOWER_TABLE[(uint8_t)c]);
}
//
// courtesy of libevent2/evutil.c
//
char
UTIL_TOUPPER(char c)
{
    return ((char)UTIL_TOUPPER_TABLE[(uint8_t)c]);
}
//
// courtesty of libevent2/evutil.c
//
int
util_ascii_strcasecmp(const char *s1, const char *s2) {
	char c1, c2;
	while (1) {
		c1 = UTIL_TOLOWER(*s1++);
		c2 = UTIL_TOLOWER(*s2++);
		if (c1 < c2)
			return -1;
		else if (c1 > c2)
			return 1;
		else if (c1 == 0)
			return 0;
	}
}
//
// courtesy of libevent2/evutil.c
//
int
util_ascii_strncasecmp(const char *s1, const char *s2, size_t n) {
	char c1, c2;
	while (n--) {
		c1 = UTIL_TOLOWER(*s1++);
		c2 = UTIL_TOLOWER(*s2++);
		if (c1 < c2)
			return -1;
		else if (c1 > c2)
			return 1;
		else if (c1 == 0)
			return 0;
	}
	return 0;
}
#ifndef HAVE_STRLCPY
//
/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
//
int
strlcpy(char *dst, const char *src, int siz) {
        char *d = dst;
        const char *s = src;
        int n = siz;

        /* Copy as many bytes as will fit */
        if (n != 0 && --n != 0) {
                do {
                        if ((*d++ = *s++) == 0)
                                break;
                } while (--n != 0);
        }

        /* Not enough room in dst, add NUL and traverse rest of src */
        if (n == 0) {
                if (siz != 0)
                        *d = '\0';              /* NUL-terminate dst */
                while (*s++)
                        ;
        }

        return(s - src - 1);    /* count does not include NUL */
}
#endif

#ifndef HAVE_STRLCAT
//
/*
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
 * Returns strlen(src) + MIN(siz, strlen(initial dst)).
 * If retval >= siz, truncation occurred.
 */
int
strlcat(char *dst, const char *src, int siz) {
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}
#endif
//
//
//
int
make_dir_path(const char *run_dir) {
    struct stat     st;
    int             res, len;
    char            tmp_path[MAX_PATH_LEN];
    char            *ndx;
    char            logmsg[512];

    strlcpy(tmp_path, run_dir, MAX_PATH_LEN);
    len = strlen(tmp_path);
    //
    /* Strip any trailing dir sep char.
    */
    if(tmp_path[len-1] == PATH_SEP) {
        tmp_path[len-1] = '\0';
    }
    //
    for(ndx = tmp_path+1; *ndx; ndx++) {
        if(*ndx == '/') {
            *ndx = '\0';
            //
            /* Stat this part of the path to see if it is a valid directory.
             * If it does not exist, attempt to create it. If it does, and
             * it is a directory, go on.  Otherwise, any other error cause it
             * to bail.
            */
            if(stat(tmp_path, &st) != 0) {
                if(errno == ENOENT) {
                    res = mkdir(tmp_path, S_IRWXU);
                }
                if(res != 0) {
                    return res;
                }
            }
            //
            if(! S_ISDIR(st.st_mode)) {
                sprintf(logmsg, "component: %s of %s is NOT a directory", tmp_path, run_dir);
                logit(_LOG_ERROR, logmsg, 0);
                return(ENOTDIR);
            }
            *ndx = '/';
        }
    }
    res = mkdir(tmp_path, S_IRWXU);
    return(res);
}
//
/* Ensure the specified directory exists.  If not, create it or die.
 * based on check_dir_path() from fwknop project
 * (Damien S. Stuart <dstuart@dstuart.org> and Michael Rash <mbr@cipherdyne.org>)
*/
int
check_dir_path(const char *filepath, const char *fp_desc, unsigned char use_basename) {
    struct stat     st;
    int             res;
    char            tmp_path[MAX_PATH_LEN];
    char            *ndx;
    char            logmsg[512];
    /*
     * FIXME:  We shouldn't use a hard-coded dir-separator here.
    */
    /* But first make sure we are using an absolute path.
    */
    if(*filepath != PATH_SEP) {
        sprintf(logmsg, "%s directory (%s) is not an absolute path", fp_desc, filepath);
        logit(_LOG_ERROR, logmsg, 0);
        return (EXIT_FAILURE);
    }
    //
    /* If this is a file path that we want to use only the basename, strip
     * the trailing filename here.
    */
    if(use_basename && ((ndx = strrchr(filepath, PATH_SEP)) != NULL)) {
        strlcpy(tmp_path, filepath, (ndx-filepath)+1);
    } else {
        strcpy(tmp_path, filepath);
    }
    //
    /* At this point, we should make the path is more than just the
     * PATH_SEP.  If it is not, silently return.
    */
    if(strlen(tmp_path) < 2) {
        return (EXIT_SUCCESS);
    }
    //
    /* Make sure we have a valid directory.
    */
    res = stat(tmp_path, &st);
    if(res != 0) {
        if(errno == ENOENT) {
            sprintf(logmsg, "%s directory (%s) does not exist - will attempt to create it", fp_desc, tmp_path);
            logit(_LOG_WARNING, logmsg, 0);
            //
            /* Directory does not exist, so attempt to create it.
            */
            res = make_dir_path(tmp_path);
            if(res != 0) {
                sprintf(logmsg, "Unable to create %s directory: %s (error: %i)", fp_desc, tmp_path, errno);
                logit(_LOG_ERROR, logmsg, 0);
                return(EXIT_FAILURE);
            }
            if (get_verbose()) {
                sprintf(logmsg, "Successfully created %s directory: %s", fp_desc, tmp_path);
                logit(_LOG_ERROR, logmsg, 0);
            }
        } else {
            sprintf(logmsg, "Stat of %s returned error %i", tmp_path, errno);
            logit(_LOG_ERROR, logmsg, 0);
            return(EXIT_FAILURE);
        }
    } else {
        //
        /* It is a file, but is it a directory?
        */
        if(! S_ISDIR(st.st_mode)) {
            sprintf(logmsg, "Specified %s directory (%s) is NOT a directory", fp_desc, tmp_path);
            logit(_LOG_ERROR, logmsg, 0);
            return(EXIT_FAILURE);
        }
    }
    return (EXIT_SUCCESS);
}
//
// return -1 on error
// return 0 on success
// return >0 if existing pid is found
//
int
write_pid_file(const char * pid_filename) {
    pid_t   old_pid, my_pid;
    int     op_fd, lck_res, num_bytes;
    char    buf[6]  = {0};
    char    logmsg[512];
    //
    /* Reset errno (just in case)
    */
    errno = 0;
    //
    if (!pid_filename) {
        return 0; // nothing to do here
    }
    //
    /* Open the PID file
    */
    op_fd = open((char *)pid_filename, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR);
    //
    if(op_fd == -1) {
        sprintf(logmsg, "unable to open pid file %s [%s]", (char *)pid_filename, strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        return (-1);
    }
    //
    /* Attempt to lock the PID file.  If we get an EWOULDBLOCK
     * error, another instance already has the lock. So we grab
     * the pid from the existing lock file, complain and bail.
    */
    lck_res = lockf(op_fd, F_TLOCK, 0);
    if(lck_res == -1) {
        if(errno != EAGAIN) {
            sprintf(logmsg, "unexpected error attempting to lock pid file %s [%s]",
                    (char *)pid_filename, strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return (-1);
        }

        close(op_fd);
        //
        /* Look for an existing lock holder. If we get a pid return it.
        */
        old_pid = get_running_pid((char *)pid_filename);
        if(old_pid) {
            return (old_pid);
        }
        //
        /* Otherwise, consider it an error.
        */
        sprintf(logmsg, "unable read existing pid file %s [%s]", (char *)pid_filename, strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        return (-1);
    }
    //
    /* Write our PID to the file
    */
    my_pid = getpid();
    snprintf(buf, 6, "%i\n", my_pid);
    //
    if(get_verbose() > 1) {
        sprintf(logmsg, "writing pid (%i) to pid file %s", my_pid, (char *)pid_filename);
        logit(_LOG_INFO, logmsg, 0);
    }
    num_bytes = write(op_fd, buf, strlen(buf));
    //
    if( errno || (num_bytes != strlen(buf)) ) {
        sprintf(logmsg, "write error on pid file %s [%s]", (char *)pid_filename, strerror(errno));
        logit(_LOG_WARNING, logmsg, 0);
    }
    //
    /* Sync/flush regardless...
    */
    fsync(op_fd);
    //
    /* Put the lock file discriptor in out options struct so any
     * child processes we may spawn can close and release it.
    */
    set_pidfd(op_fd);
    //
    return 0;
}
//
//
//
pid_t
get_running_pid(const char * pid_filename) {
    int     op_fd;
    char    buf[6]  = {0};
    pid_t   rpid    = 0;
    char    logmsg[512];

    op_fd = open((char *)pid_filename, O_RDONLY);

    if(op_fd > 0) {
        if (read(op_fd, buf, 6) > 0) {
            rpid = (pid_t)atoi(buf);
            if (get_verbose()) {
                sprintf(logmsg, "existing pid (%i) found in pid file %s", (int)rpid, (char *)pid_filename);
                logit(_LOG_INFO, logmsg, 0);
            }
        }
        close(op_fd);
    }
    return(rpid);
}
//
/* Become a daemon: fork(), start a new session, chdir "/",
 * and close unneeded standard filehandles.
*/
int
daemonize_process(const char * pid_filename) {
    pid_t pid, old_pid;
    char logmsg[512];
    //
    /* Reset our umask
    */
    umask(0);
    //
    if ((pid = fork()) < 0) {
        sprintf(logmsg, "unable to fork [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        return(EXIT_FAILURE);
    } else if (pid != 0) {
        /* parent */
        if (get_verbose()) {
            sprintf(logmsg, "successful fork of child process %i", (int)pid);
            logit(_LOG_INFO, logmsg, 0);
        }
        return(EXIT_SUCCESS);
    }
    //
    /* Child process from here on out */
    if (get_verbose()) {
        sprintf(logmsg, "new instance of program started (pid %i)", (int)pid);
        logit(_LOG_INFO, logmsg, 0);
    }
    //
    /* Start a new session
    */
    setsid();
    //
    /* Create the PID file (or be blocked by an existing one).
    */
    if (pid_filename) {
        old_pid = write_pid_file((char *)pid_filename);
        if(old_pid > 0) {
            sprintf(logmsg, "another instance of program appears to be running already (pid %i)", (int)old_pid);
            logit(_LOG_WARNING, logmsg, 0);
            return(EXIT_FAILURE);
        } else if(old_pid < 0) {
            // error messages already written
        }
    }
    //
    /* Chdir to the root of the filesystem
    */
    if ((chdir("/")) < 0) {
        sprintf(logmsg, "unable to chdir() to / [%s]", strerror(errno));
        logit(_LOG_WARNING, logmsg, 0);
        return(EXIT_FAILURE);
    }
    //
    /* Close un-needed file handles
    */
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);

    return (EXIT_SUCCESS);
}
