/**
 * @file conf.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created April 20, 2011
 * @modified May 26, 2011
 * @version: 0.11
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @credits Jon Mayo
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <error.h>
#include "conf.h"
#include "parms.h"
//
struct parameter {
    char * name;
    char * value;
    struct parameter * next;
};
//
struct section {
    char * name;
    struct parameter * head;
    struct section * next;
};
//
struct ini_st {
    struct section * head;
    char * filename;
    unsigned line;
    struct section * curr_section;
    struct parameter * curr_parameter;
};
//
// ini_free() function to cleanup ini file stuff when done
//
void
ini_free(struct ini_st * ii) {
    struct section * curr_section, * next_section;
    struct parameter * curr_parameter, * next_parameter;
    if (ii) {
        for (curr_section=ii->head; curr_section; curr_section=next_section) {
            for (curr_parameter=curr_section->head; curr_parameter; curr_parameter=next_parameter) {
                next_parameter = curr_parameter->next;
                free (curr_parameter->name);
                free (curr_parameter->value);
                free (curr_parameter);
            }
            next_section = curr_section->next;
            free (curr_section->name);
            free (curr_section);
        }
        free (ii->filename);
        free (ii);
    }
}
//
// ini_load() function to load an existing ini file contents into ini_st struct
//
struct ini_st *
ini_load(const char * filename) {
    char buf[512], *s;
    char msg[128];
    struct ini_st *result;
    struct section *curr_section, **prev_section;
    struct parameter *curr_parameter, **prev_parameter;
    unsigned line;
    FILE *pFile;
    //
    pFile = fopen(filename, "r");
    if (!pFile) {
        sprintf(msg, "unable to open config file (%s) - %s", filename, strerror(errno));
        logit(_LOG_ERROR, msg, 0);
        return 0;
    }
    /* create a new ini_info */
    result = malloc(sizeof *result);
    result->filename = strdup(filename);
    result->line = 0;
    result->curr_section = 0;
    result->curr_parameter = 0;
    //
    prev_section = &result->head;
    curr_parameter = 0;
    prev_parameter = 0;
    line = 0;
    while (fgets(buf, sizeof(buf), pFile)) {
        line++;
        s = buf;
        while (isspace(*s)) {
            s++; /* ignore leading whitespace */
        }
        if (*s == ';' || *s == '#' || *s == 0) {
            /* comment or blank line */
        } else if(*s == '[') { /* new section */
            char * tmp;
            /* parse the section */
            s = buf+1;
            tmp = strchr(s, ']');
            if (tmp) {
                *tmp = 0;
            } else { /* error : unterminated section names */
                sprintf(msg, "%s:%d:unterminated section name", filename, line);
                logit(_LOG_ERROR, msg, 0);
                ini_free (result);
                result = 0; /* failure */
                break;
            }
            /* create section */
            curr_section = malloc(sizeof *curr_section);
            curr_section->next = 0;
            curr_section->head = 0;
            curr_section->name = strdup(s);
            /* append it to the end */
            *prev_section = curr_section;
            prev_section = &curr_section->next;
            prev_parameter = &curr_section->head;
            curr_parameter = *prev_parameter;
        } else if (prev_parameter) { /* parameter inside section */
            char *value, *tmp;
            /* trim trailing comment */
            tmp = strchr(s, ';');
            if (tmp) {
                *tmp = 0;
            }
            /* trim trailing whitespace */
            tmp = (s + strlen(s));
            while (tmp!=s) {
                tmp--;
                if (!isspace(*tmp)) {
                    break;
                }
                *tmp = 0;
            }
            /* look for '=' sign to mark the start of the value */
            value = s;
            while (*value && (*value != '=')) {
                value++;
            }
            if (*value) {
                /* build the value string */
                *value = 0; /* chop parameter string - now s is parameter */
                value++;
                while (isspace(*value)) {
                    value++; /* ignore leading whitespace */
                }
                /* build the parameter string */
                /* trim trailing whitespace */
                tmp = (s+strlen(s));
                while (tmp != s) {
                    tmp--;
                    if (!isspace(*tmp)) {
                        break;
                    }
                    *tmp = 0;
                }
                /* create parameter */
                curr_parameter = malloc(sizeof * curr_parameter);
                curr_parameter->name = strdup(s);
                curr_parameter->value = strdup(value);
                curr_parameter->next = 0;
                /* append it to the end */
                *prev_parameter = curr_parameter;
                prev_parameter = &curr_parameter->next;
            } else { /* error: no value found */
                sprintf(msg, "%s:%d:no value found for parameter", filename, line);
                logit(_LOG_WARNING, msg, 0);
                ini_free (result);
                result = 0; /* failure */
                break;
            }
        } else { /* error: parameters outside section */
            sprintf(msg, "%s:%d:parse error in INI file", filename, line);
            logit(_LOG_ERROR, msg, 0);
            /* it's important to always keep the structure in a consistant
            * state so that we can use the standard free utility */
            ini_free(result);
            result = 0; /* failure */
            break;
        }
    }
    fclose(pFile);
    return result;
}
//
// sequential i/o
// ini_next_section() function
// return section name, NULL on end of file
//
const char *
ini_next_section(struct ini_st * ii) {
    ii->curr_parameter = 0; /* always reset the parameter */
    //
    if (ii->curr_section) {
        ii->curr_section = ii->curr_section->next;
    } else {
        ii->curr_section = ii->head;
    }
    //
    if (ii->curr_section) {
        return (ii->curr_section->name);
    } else {
        return 0; /* end */
    }
}
//
// sequential i/o
// ini_next_parameter() function
// return value of parameter. NULL on end of section
//
const char *
ini_next_parameter(struct ini_st * ii, const char ** parameter) {
    /* go to next parameter, or the first if set to 0 */
    if (ii->curr_parameter) {
        ii->curr_parameter = ii->curr_parameter->next;
    } else {
        ii->curr_parameter = ii->curr_section->head;
    }
    //
    if (ii->curr_parameter) {
        if (parameter) {
            *parameter = ii->curr_parameter->name;
        }
        return (ii->curr_parameter->value);
    } else {
        return 0; /* end */
    }
}
//
// rewind the pointers so the ini_next_xxx functions can start from the top */
//
void
ini_rewind(struct ini_st * ii) {
    if (ii) {
        ii->curr_parameter = 0;
        ii->curr_section = 0;
    }
}
//
// random access ini_get() function
// return value for a parameter
const char *
ini_get(struct ini_st * ii, const char * section, const char * parameter) {
    struct section * curr_section;
    struct parameter * curr_parameter;

    for (curr_section=ii->head; curr_section; curr_section=curr_section->next) {
        if (strcasecmp(curr_section->name, section)==0) {
            break;
        }
    }
    if (!curr_section) {
        return 0;
    }
    for (curr_parameter=curr_section->head; curr_parameter; curr_parameter=curr_parameter->next) {
        if (strcasecmp(curr_parameter->name, parameter)==0) {
            return curr_parameter->value;
        }
    }
    return 0;
}
