/**
 * @file main.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created April 20, 2011
 * @based on Covert_TCP 1.0 written by Craig H. Rowland (crowland@psionic.com)
 * @modified June 18, 2011
 * @version 0.49
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include "parms.h"
#include "hostnames.h"
#ifdef LIBEV
#include "libevwrap.h"
#endif
#include "hpingwrap.h"
#include "pkcommon.h"
//
// function prototypes
//
int run(int argc, char** argv);
//
// main() function
//
int
main(int argc, char** argv) {
    sig_interrupted = 0;
    int p;
    char logmsg[128];
    sprintf(logmsg, "invizible %s", INVIZ_VERSION);
    logit(_LOG_INFO, logmsg, 0);
    sprintf(logmsg, "see LICENSE file for copyright, terms and conditions");
    logit(_LOG_INFO, logmsg, 0);
    //
    // initialize default parameters
    //
    if ( init_parms() == EXIT_FAILURE ) {
        return ( EXIT_FAILURE );
    }
    set_originalpsname();
    set_programname(argv[0], strlen(argv[0]));
    set_defaultpsmask();
    //
    // now init embedded domain names to be used for lookups
    //
    set_hostnames();
    //
#ifdef LIBEV
    loop = EV_DEFAULT;
#endif
    if ( valparms(argc, argv) == EXIT_SUCCESS ) {
        setvbuf(stdout, NULL, _IOLBF, 0); // set stdout to line buffered
        //
        if (get_verbose()) {
            sprintf(logmsg, "running %s", get_uname());
            logit(_LOG_INFO, logmsg, 0);
        }
        //
        // now mask process name
        //
        char *parms = malloc(INVIZ_MAX_COMMANDLINE_LEN+1);
        memset(parms, 0, INVIZ_MAX_COMMANDLINE_LEN+1);
        strncpy(parms, argv[0], strlen(argv[0]));
        memset(argv[0], 0x0, strlen(argv[0]));
        int pslen = strlen(get_psmask());
        strncpy(argv[0], get_psmask(), pslen);
        set_psmask(argv[0], strlen(argv[0]));
        //
        // and hide parameters (from ps)
        // (they are not needed/used by run() function)
        //
        for (p=1; p<argc; p++) {
            strncat(parms, " ", 1);
            strncat(parms, argv[p], strlen(argv[p]));
            memset(argv[p], 0, strlen(argv[p]));
        }
        memset(logmsg, 0, sizeof(logmsg));
        sprintf(logmsg, "%s", parms);
        logit(_LOG_INFO, logmsg, 0);
        free(parms);
        parms = NULL;
        if (get_verbose()) {
            sprintf(logmsg, "invizible process %i now masked as %s",
                (int)getpid(),
                (char *)get_psmask());
            logit(_LOG_INFO, logmsg, 0);
        }
        //
        //
        //
        if (get_wipeprogram()) {
            if (wipefile(get_programname(), get_iwipetimes()) == EXIT_SUCCESS) {
                sprintf(logmsg, "invizible program now wiped and deleted from disk");
                logit(_LOG_INFO, logmsg, 0);
            } else {
                sprintf(logmsg, "unable to wipe invizible program from disk");
                logit(_LOG_WARNING, logmsg, 0);
            }
        }
        //
        // now run mainline
        //
        int run_exit = ( run(argc, argv) );
        //
        // now cleanup
        //
        cleanup_parms();
        return ( run_exit );
    } else {
        return ( EXIT_FAILURE );
    }
}
//
// run() function
// called from main()
//
int
run(int argc, char** argv) {
    unsigned int source_host=0, dest_host=0;
    int ipid=0, seq=0, ack=0;
    int spoofing_source = FALSE;
    int i, rc;
    char logmsg[512];
    int server = (get_mode() == INVIZ_MODE_LISTENER);
    //
#ifdef LIBEV
    //
    // setup signal handler for Ctrl-C
    //
    ev_signal_init(&signal_watcher, sigint_cb, SIGINT);
    ev_signal_start(loop, &signal_watcher);
#endif
    //
    /* Can they run this? */
    //
    if(geteuid() != 0) {
        logit(_LOG_ERROR, "you must have root privileges to run this program", 0);
        return ( EXIT_FAILURE );
    }
    //
    // dest_host is receiving end of file transfer (ultimate destination) from client
    // or receiving end of port knocking (ultimate destination) from client
    // or originating client (from server) if sending back confirmation packets
    // or if sending back port knocking command results
    //
    // source_host is dns/bounce server in file transfer (from both client and server)
    // or spoofed ip (if specified) by client in port knocking
    // or originating client or server in port knocking
    //
    if (!server) {
        //
        // is client - always sending to server as ultimate destination
        //
        set_destinationip((char *)get_serverip(), strlen(get_serverip()));
    } else if ((get_iconfirminterval() > 0) || get_usepk()) {
        //
        // is server and sending back confirmation packets to specified client
        // or sending back port knocking results to specified client
        // then client should be specified with --client= parameter
        // if the client is not sending directly to the server
        // (i.e. is bouncing packets to server)
        //
        if (get_clientip() && (strlen(get_clientip()) > 0)) {
            set_destinationip((char *)get_clientip(), strlen(get_clientip()));
        }
    }
    if (!get_usepk()) {
        //
        // doing file transfer - bounced off dns server
        // source_host must be same on both client and server programs
        // source_host is where dns requests are sent from client
        // source_host is where dns replies are coming to server
        // "source" in this context is from perspective of ultimate destination
        // (not from perspective of actual, originating client or server)
        //
        set_sourceip((char *)get_dnsip(), strlen(get_dnsip()));
    }
    //
    // if using ipv6? which host_convert equivalent?
    //
    //
    /* Do the dirty work */
    //
    if (!sig_interrupted) {
        //
        // is port knocking to be used?
        // then init context
        //
        if (get_usepk()) {
            //
            // initialize port knocking "context"
            //
            if (set_pkcontext() == EXIT_FAILURE) {
                return ( EXIT_FAILURE );
            }
            if (get_spoofip() && (strlen(get_spoofip()) > 0)) {
                //
                // spoofed ip specified?
                // if so, this is similar to bouncing file transfer packets
                // off dns server
                //
                set_sourceip((char *)get_spoofip(), strlen(get_spoofip()));
                spoofing_source = TRUE;
            } else {
                //
                // must use own public/external IP address if not spoofed
                // if not yet set, then set to "auto" so can be resolved
                // in next step of this logic
                //
                if ( !get_pkprivatesubnet() && (!get_externalip4() || (strlen(get_externalip4()) == 0)) ) {
                    set_externalip4(INVIZ_AUTO_DESC, strlen(INVIZ_AUTO_DESC));
                }
            }
        }
        //
        // must public ip address be automatically determined?
        // signaled by parameter --self=auto
        // usually just required by client to learn what IP address the server must allow access to
        // but maybe needed by server
        // if public ip address already known with certainty then it can be specified as
        // --self=123.123.123.123 which will ensure no outgoing tcp connection is made to
        // a site such as showmyip.com or ippages.com or whatismyip.com etc to determine it
        // Note: resolve_ip_http_inviz() does a set_externalip4() if successful
        //
        int do_auto_self_resolve = ( get_externalip4() &&
            (strncmp(get_externalip4(), INVIZ_AUTO_DESC, strlen(INVIZ_AUTO_DESC)) == 0) );
        if (do_auto_self_resolve) {
            rc = resolve_ip_http_inviz(HTTP_RESOLVE_HOST, HTTP_RESOLVE_URL);
            if (rc != FKO_SUCCESS) {
#ifndef HTTP_RESOLVE_HOSTS
                sprintf(logmsg, "unable to get external/public ip4 address [%s]", fko_errstr(rc));
                logit(_LOG_ERROR, logmsg, 0);
                return ( EXIT_FAILURE );
#else
                rc = resolve_ip_http_inviz(HTTP_RESOLVE_HOST3, HTTP_RESOLVE_URL3);
                if (rc != FKO_SUCCESS) {
                    rc = resolve_ip_http_inviz(HTTP_RESOLVE_HOST4, HTTP_RESOLVE_URL4);
                    if (rc != FKO_SUCCESS) {
                        sprintf(logmsg, "unable to get external/public ip4 address [%s]", fko_errstr(rc));
                        logit(_LOG_ERROR, logmsg, 0);
                        return ( EXIT_FAILURE );
                    }
                }
#endif
            }
            if ( get_pktype()  &&
                    get_pkallowip4() &&
                    (strncmp(get_pkallowip4(), INVIZ_AUTO_DESC, strlen(INVIZ_AUTO_DESC)) == 0) ) {
                //
                // pk access is specified
                // and pk access ip address is set to "auto"
                // so
                // set pk access ip address to auto calculated external ip address
                // just resolved above
                //
                set_pkallowip4(get_externalip4(), strlen(get_externalip4()));
            }
        }
        if (get_usepk() && !spoofing_source && !get_pkprivatesubnet()) {
            // source_host for port knocking is the originating "source"
            // source_host for file transfers is the bounced DNS server
            set_sourceip((char *)get_externalip(), strlen(get_externalip()));
            //
            // if doing pk on local private subnet then set source ip later
            // when we know it
            //
        }
        if (get_useip4()) {
            if (get_sourceip()) {
                source_host = host_convert_ip4((char *)get_sourceip());
            }
            if (get_destinationip()) {
                dest_host = host_convert_ip4((char *)get_destinationip());
            }
        } else {
            sprintf(logmsg, "incomplete support for IPv6 at this time");
            logit(_LOG_ERROR, logmsg, 0);
            return ( EXIT_FAILURE );
        }
        doinvizible(source_host, dest_host, (char *)get_filename(), server, ipid, seq, ack);
        if (get_usepk()) {
            fko_destroy(get_pkcontext());
        }
    }
    //
    // cleanup
    //
    for (i=0; i < HOST_COM_COUNT; i++) {
        if (hostcoms[i]) {
            free(hostcoms[i]);
            hostcoms[i] = NULL;
        }
    }
    for (i=0; i < HOST_NET_COUNT; i++) {
        if (hostnets[i]) {
            free(hostnets[i]);
            hostnets[i] = NULL;
        }
    }
    for (i=0; i < HOST_ORG_COUNT; i++) {
        if (hostorgs[i]) {
            free(hostorgs[i]);
            hostorgs[i] = NULL;
        }
    }
    for (i=0; i < HOST_OTHER_COUNT; i++) {
        if (hostothers[i]) {
            free(hostothers[i]);
            hostothers[i] = NULL;
        }
    }
    return ( EXIT_SUCCESS );
}
