/**
 * @file mypcap.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created April 30, 2011
 * @modified June 15, 2011
 * @version 0.42
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include "mypcap.h"
#include "parms.h"

//
// based on libpcap/inet.c
// lookup specified device
//
/*
 * Return the name of a network interface attached to the system, or NULL
 * if none can be found.  The interface must be configured up; the
 * lowest unit number is preferred; loopback is ignored.
 */
char *
pcap_lookupdevspec(char * specdev) {
    pcap_if_t *alldevs, *d;
/* for old BSD systems, including bsdi3 */
#ifndef IF_NAMESIZE
#define IF_NAMESIZE IFNAMSIZ
#endif
    static char device[IF_NAMESIZE + 1];
    char * ret;
    char * errbuf;
    int isloopback, islb, isconfigured;
    char logmsg[512];
    //
    if (pcap_findalldevs(&alldevs, errbuf) == -1) {
        // none found
        return (NULL);
    }
    //
    if (alldevs == NULL) {
        /*
	 * There are no devices on the list, or the first device
	 * on the list is a loopback device, which means there
	 * are no non-loopback devices on the list.  This means
	 * we can't return any device.
	 *
	 * XXX - why not return a loopback device?  If we can't
	 * capture on it, it won't be on the list, and if it's
	 * on the list, there aren't any non-loopback devices,
	 * so why not just supply it as the default device?
	 */
	(void)strlcpy(errbuf, "no suitable device found", PCAP_ERRBUF_SIZE);
        ret = NULL;
    } else {
        /*
         * Return the name of the first device on the list?
	 */
        isloopback = (alldevs->flags & PCAP_IF_LOOPBACK);
        isconfigured = (alldevs->addresses != 0x0);
        ret = alldevs->name;
        if (isloopback || !isconfigured ||
                (specdev && (specdev[0] != '\0') && (strcmp(device, specdev) != 0))) {
            ret = NULL;
            //
            // now check list of all devs if first device is not one specified
            // or if first device is loopback
            // but ignore loopback devs
            //
            for (d=alldevs; d; d=d->next) {
                islb = (d->flags & PCAP_IF_LOOPBACK);
                isconfigured = (d->addresses != 0x0);
                if (!islb && isconfigured && (strcmp(d->name, specdev) == 0)) {
                    ret = d->name;
                    break;
                } else if (!islb && isconfigured && (strcmp(specdev, INVIZ_ANY_DESC) == 0)) {
                    ret = d->name;
                    break;
                } else if (!islb && isconfigured) {
                    sprintf(logmsg, "additional interface %s available on this device", d->name);
                    logit(_LOG_INFO, logmsg, 0);
                }
            }
        }
    }
    //
    //pcap_freealldevs(alldevs);
    return (ret);
}
//
// This is the callback function that is passed to pcap_loop(..) and called each time
// a packet is received by the server
//
void
pcap_server_packet_callback(u_char * args, const struct pcap_pkthdr * pkthdr, const u_char * packet) {
    static uint pcap_packet_count = 0;
    u_int16_t type = handle_ethernet(args, pkthdr, packet);
    //
    if(type == ETHERTYPE_IP) {
        // handle the IP packet
    	handle_IP(args, pkthdr, packet, TRUE);
    } else if (type == ETHERTYPE_ARP) {
        // handle the ARP packet
    } else if (type == ETHERTYPE_REVARP) {
        // handle reverse arp packet
    }
    //
    pcap_packet_count++;
}
//
// This is the callback function that is passed to pcap_loop(..) and called each time
// a packet is received by the client
//
void
pcap_client_packet_callback(u_char * args, const struct pcap_pkthdr * pkthdr, const u_char * packet) {
    static uint pcap_packet_count = 0;
    u_int16_t type = handle_ethernet(args, pkthdr, packet);
    //
    if(type == ETHERTYPE_IP) {
        // handle the IP packet
    	handle_IP(args, pkthdr, packet, FALSE);
    } else if (type == ETHERTYPE_ARP) {
        // handle the ARP packet
    } else if (type == ETHERTYPE_REVARP) {
        // handle reverse arp packet
    }
    //
    pcap_packet_count++;
}
//
// name_parse() based on libevent2/evdns.c
//
int
name_parse(u8 *packet, int length, int *idx, char *name_out, int name_out_len) {
	int name_end = -1;
	int j = *idx;
	int ptr_count = 0;
        char logmsg[512];
#define GET32(x) do { if (j + 4 > length) goto err; memcpy(&_t32, packet + j, 4); j += 4; x = ntohl(_t32); } while (0)
#define GET16(x) do { if (j + 2 > length) goto err; memcpy(&_t, packet + j, 2); j += 2; x = ntohs(_t); } while (0)
#define GET8(x) do { if (j >= length) goto err; x = packet[j++]; } while (0)

	char *cp = name_out;
	const char *const end = (name_out + name_out_len);

	/* Normally, names are a series of length prefixed strings terminated */
	/* with a length of 0 (the lengths are u8's < 63). */
	/* However, the length can start with a pair of 1 bits and that */
	/* means that the next 14 bits are a pointer within the current */
	/* packet. */

	for (;;) {
		u8 label_len;
		if (j >= length) {
                    if (get_verbose()) {
                        sprintf(logmsg, "invalid length (%d %d) encountered in DNS packet name_parse [1]", j, length);
                        logit(_LOG_WARNING, logmsg, 0);
                    }
                    return EXIT_FAILURE;
                }
		GET8(label_len);
		if (!label_len) {
                    break;
                }
		if (label_len & 0xc0) {
			u8 ptr_low;
			GET8(ptr_low);
			if (name_end < 0) {
                            name_end = j;
                        }
			j = (((int)label_len & 0x3f) << 8) + ptr_low;
			/* Make sure that the target offset is in-bounds. */
			if (j < 0 || j >= length) {
                            if (get_verbose()) {
                                sprintf(logmsg, "invalid length (%d %d) encountered in DNS packet name_parse [2]", j, length);
                                logit(_LOG_WARNING, logmsg, 0);
                            }
                            return EXIT_FAILURE;
                        }
			/* If we've jumped more times than there are characters in the
			 * message, we must have a loop. */
			if (++ptr_count > length) {
                            if (get_verbose()) {
                                sprintf(logmsg, "invalid length (%d %d) encountered in DNS packet name_parse [3]", ptr_count, length);
                                logit(_LOG_WARNING, logmsg, 0);
                            }
                            return EXIT_FAILURE;
                        }
			continue;
		}
		if (label_len > 63) {
                    if (get_verbose()) {
                        sprintf(logmsg, "invalid length (%d) encountered in DNS packet name_parse [4]", label_len);
                        logit(_LOG_WARNING, logmsg, 0);
                    }
                    return EXIT_FAILURE;
                }
		if (cp != name_out) {
			if (cp + 1 >= end) {
                            if (get_verbose()) {
                                sprintf(logmsg, "invalid length (%d %d) encountered in DNS packet name_parse [5]", (int)cp, (int)end);
                                logit(_LOG_WARNING, logmsg, 0);
                            }
                            return EXIT_FAILURE;
                        }
			*cp++ = '.';
		}
		if (cp + label_len >= end) {
                    if (get_verbose()) {
                        sprintf(logmsg, "invalid length (%d %d %d) encountered in DNS packet name_parse [6]", (int)cp, label_len, (int)end);
                        logit(_LOG_WARNING, logmsg, 0);
                    }
                    return EXIT_FAILURE;
                }
		memcpy(cp, packet + j, label_len);
		cp += label_len;
		j += label_len;
	}
	if (cp >= end) {
            if (get_verbose()) {
                sprintf(logmsg, "invalid length (%d %d) encountered in DNS packet name_parse [7]", (int)cp, (int)end);
                logit(_LOG_WARNING, logmsg, 0);
            }
            return EXIT_FAILURE;
        }
	*cp = '\0';
	if (name_end < 0) {
            *idx = j;
        } else {
            *idx = name_end;
        }
	return EXIT_SUCCESS;
 err:
	return EXIT_FAILURE;
}
//
// reply_parse() of received dns packet (based on libevent2/evdns.c)
//
int
reply_parse(u8 *packet, int length, int serious_error) {
	int j = 0, k = 0;  /* index into packet */
	u16 _t;	 /* used by the macros */
	u32 _t32;  /* used by the macros */
	char tmp_name[256], cmp_name[256]; /* used by the macros */
        char logmsg[512];
	int name_matches = FALSE; // dns reply host name matches originating dns query host name?
	u16 trans_id, questions, answers, authority, additional;
	u16 flags = 0;
        u16 dns_query_type, dns_query_class;
	u32 ttl_r = 0xffffffff;
	struct reply dns_reply;
	struct request *dns_query = NULL;
	unsigned int i;

	GET16(trans_id);
	GET16(flags);
	GET16(questions);
	GET16(answers);
	GET16(authority);
	GET16(additional);
	(void) authority; /* suppress "unused variable" warnings. */
	(void) additional; /* suppress "unused variable" warnings. */

#ifndef DNS_SKIP_MATCH_TO_ORIGINAL
        // find originating dns query
        dns_query = request_find_from_trans_id(base, trans_id);
	if (!dns_query) return EXIT_FAILURE;
#endif
        //
	memset(&dns_reply, 0, sizeof(dns_reply));

	/* If it's not an answer, it doesn't correspond to any request. */
	if (!(flags & 0x8000)) {
            sprintf(logmsg, "DNS reply packet flags indicate it is not an answer packet" );
            logit(_LOG_WARNING, logmsg, 0);
            serious_error = TRUE;
            return EXIT_FAILURE;  /* must be an answer */
        }
	if (flags & 0x020f) {
            /* there was an error */
            sprintf(logmsg, "DNS reply packet flags indicate error");
            logit(_LOG_WARNING, logmsg, 0);
            serious_error = TRUE;
            return EXIT_FAILURE;
	}
        //
	/* This macro skips a name in the DNS reply. */
        //
#define SKIP_NAME                                                           \
	do {                                                                \
            tmp_name[0] = '\0';                                             \
            if (name_parse(packet, length, &j, tmp_name, sizeof(tmp_name)) != EXIT_SUCCESS) { \
                sprintf(logmsg, "DNS reply packet skip_name error (%i %i %s)", length, j, tmp_name);        \
                logit(_LOG_WARNING, logmsg, 0);                             \
                serious_error = FALSE;                                      \
                return EXIT_FAILURE;                                        \
            }                                                               \
	} while (0)
        //
        /* This macro compares name in reply to name in request */
        //
#define TEST_NAME                                                           \
	do {                                                                \
            tmp_name[0] = '\0';                                             \
            cmp_name[0] = '\0';                                             \
            k = j;                                                          \
            if (name_parse(packet, length, &j, tmp_name, sizeof(tmp_name)) != EXIT_SUCCESS) { \
                sprintf(logmsg, "DNS reply packet test_name error (%i %i %s)", length, j, tmp_name); \
                logit(_LOG_WARNING, logmsg, 0);                             \
                serious_error = FALSE;                                      \
                return EXIT_FAILURE;                                        \
            }                                                               \
            if (dns_query != NULL && (name_parse(dns_query->request, dns_query->request_len, &k, cmp_name, sizeof(cmp_name)) != EXIT_SUCCESS)) \
                goto err;                                                   \
            if (get_randomizecase()) {                                      \
                if (strcmp(tmp_name, cmp_name) == 0) {                      \
                    name_matches = TRUE;                                    \
                }                                                           \
            } else {                                                        \
                if (util_ascii_strcasecmp(tmp_name, cmp_name) == 0) {       \
                    name_matches = TRUE;                                    \
                }                                                           \
            }                                                               \
	} while (0)

        if (dns_query != NULL) {
            dns_reply.type = dns_query->request_type;
        }

	/* skip over each question in the reply */
	for (i = 0; i < questions; ++i) {
		/* the question looks like
		 *   <label:name><u16:type><u16:class>
		 */
		TEST_NAME;
                GET16(dns_query_type);
                GET16(dns_query_class);
                if (dns_query == NULL) {
                    dns_reply.type = dns_query_type;
                }
		if (j >= length) {
                    sprintf(logmsg, "DNS reply packet invalid length (%i)", length);
                    logit(_LOG_WARNING, logmsg, 0);
                    serious_error = TRUE;
                    goto err;
                }
	}

	if ( (dns_query != NULL) && !name_matches ) {
            goto err;
        }

	/* now we have the answer section which looks like
	 * <label:name><u16:type><u16:class><u32:ttl><u16:len><data...>
	 */

	for (i = 0; i < answers; ++i) {
            u16 dns_answer_type, dns_answer_class, dns_answer_datalength;
            u32 dns_answer_ttl;
            SKIP_NAME;
            GET16(dns_answer_type);
            GET16(dns_answer_class);
            GET32(dns_answer_ttl);
            GET16(dns_answer_datalength);
            if (dns_query_type == DNS_TYPE_A && dns_query_class == DNS_CLASS_INET) {
                int addrcount, addrtocopy;
		if (dns_query != NULL && (dns_query->request_type != DNS_TYPE_A)) {
                    j += dns_answer_datalength;
                    continue;
		}
		if ((dns_answer_datalength & 3) != 0) {
                    /* not an even number of As. */
                    //sprintf(logmsg, "unrecognized DNS type A reply packet dns_answer_datalength (%i)", dns_answer_datalength);
                    //logit(_LOG_WARNING, logmsg, 0);
                    //serious_error = FALSE;
                    //goto err;
                }
		addrcount = dns_answer_datalength >> 2;
		addrtocopy = MIN(DNS_MAX_IP4_ADDRS - dns_reply.data.a.addrcount, (unsigned)addrcount);
                ttl_r = MIN(ttl_r, dns_answer_ttl);
		/* we only bother with the first four addresses returned */
		if (j + 4*addrtocopy > length) {
                    sprintf(logmsg, "more than 4 addresses in DNS type A reply packet");
                    logit(_LOG_WARNING, logmsg, 0);
                    serious_error = FALSE;
                    goto err;
                }
		memcpy(&dns_reply.data.a.addresses[dns_reply.data.a.addrcount], packet + j, 4*addrtocopy);
		j += 4*addrtocopy;
		dns_reply.data.a.addrcount += addrtocopy;
		dns_reply.have_answer = 1;
		if (dns_reply.data.a.addrcount == DNS_MAX_IP4_ADDRS) {
                    break;
                }
            } else if (dns_query_type == DNS_TYPE_PTR && dns_query_class == DNS_CLASS_INET) {
                if (dns_query != NULL && dns_query->request_type != DNS_TYPE_PTR) {
                    j += dns_answer_datalength;
                    continue;
		}
		if (name_parse(packet, length, &j, dns_reply.data.ptr.name, sizeof(dns_reply.data.ptr.name)) != EXIT_SUCCESS) {
                    serious_error = TRUE;
                    goto err;
                }
		ttl_r = MIN(ttl_r, dns_answer_ttl);
		dns_reply.have_answer = 1;
		break;
            } else if (dns_query_type == DNS_TYPE_CNAME) {
                char cname[NI_MAXHOST];
		if (dns_query != NULL && (!dns_query->put_cname_in_ptr || *dns_query->put_cname_in_ptr)) {
                    j += dns_answer_datalength;
                    continue;
		}
		if (name_parse(packet, length, &j, cname, sizeof(cname)) != EXIT_SUCCESS) {
                    serious_error = TRUE;
                    goto err;
                }
                if (dns_query != NULL) {
                    *dns_query->put_cname_in_ptr = strdup(cname);
                }
            } else if (dns_query_type == DNS_TYPE_AAAA && dns_query_class == DNS_CLASS_INET) {
                int addrcount, addrtocopy;
		if (dns_query != NULL && dns_query->request_type != DNS_TYPE_AAAA) {
                    j += dns_answer_datalength;
                    continue;
		}
		if ((dns_answer_datalength & 15) != 0) {
                    /* not an even number of AAAAs. */
                    serious_error = FALSE;
                    goto err;
                }
		addrcount = dns_answer_datalength >> 4;  /* each address is 16 bytes long */
		addrtocopy = MIN(DNS_MAX_IP6_ADDRS - dns_reply.data.aaaa.addrcount, (unsigned)addrcount);
		ttl_r = MIN(ttl_r, dns_answer_ttl);
                /* we only bother with the first four addresses. */
		if (j + 16*addrtocopy > length) {
                    serious_error = FALSE;
                    goto err;
                }
		memcpy(&dns_reply.data.aaaa.addresses[dns_reply.data.aaaa.addrcount], packet + j, 16*addrtocopy);
		dns_reply.data.aaaa.addrcount += addrtocopy;
		j += 16*addrtocopy;
		dns_reply.have_answer = 1;
		if (dns_reply.data.aaaa.addrcount == DNS_MAX_IP6_ADDRS) {
                    break;
                }
            } else {
		/* skip over any other type of resource */
                j += dns_answer_datalength;
            }
	}
	return ( EXIT_SUCCESS );
 err:
	return ( EXIT_FAILURE );
}
//
// courtesy of libevent2/evdns.c
/* Parse a raw request (packet,length) sent to a nameserver port (port) from */
/* a DNS client (addr,addrlen), and if it's well-formed, call the corresponding */
/* callback. */
//
int
request_parse(u8 *packet, int length, struct sockaddr *addr, socklen_t addrlen)
{
	int j = 0;	/* index into packet */
	u16 _t;	 /* used by the macros */
	char tmp_name[256]; /* used by the macros */

	int i;
	u16 trans_id, flags, questions, answers, authority, additional;
	struct server_request *server_req = NULL;

	/* Get the header fields */
	GET16(trans_id);
	GET16(flags);
	GET16(questions);
	GET16(answers);
	GET16(authority);
	GET16(additional);

	if (flags & 0x8000) return -1; /* Must not be an answer. */
	flags &= 0x0110; /* Only RD and CD get preserved. */

	server_req = malloc(sizeof(struct server_request));
	if (server_req == NULL) return -1;
	memset(server_req, 0, sizeof(struct server_request));

	server_req->trans_id = trans_id;
	memcpy(&server_req->addr, addr, addrlen);
	server_req->addrlen = addrlen;

	server_req->base.flags = flags;
	server_req->base.nquestions = 0;
	server_req->base.questions = calloc(sizeof(struct dns_server_question *), questions);
	if (server_req->base.questions == NULL)
		goto err;

	for (i = 0; i < questions; ++i) {
		u16 type, class;
		struct dns_server_question *q;
		int namelen;
		if (name_parse(packet, length, &j, tmp_name, sizeof(tmp_name)) != EXIT_SUCCESS)
			goto err;
		GET16(type);
		GET16(class);
		namelen = (int)strlen(tmp_name);
		q = malloc(sizeof(struct dns_server_question) + namelen);
		if (!q)
			goto err;
		q->type = type;
		q->dns_question_class = class;
		memcpy(q->name, tmp_name, namelen+1);
		server_req->base.questions[server_req->base.nquestions++] = q;
	}

	/* Ignore answers, authority, and additional. */

	//server_req->port = port;
	//port->refcnt++;

	/* Only standard queries are supported. */
	if (flags & 0x7800) {
		//evdns_server_request_respond(&(server_req->base), DNS_ERR_NOTIMPL);
		return ( EXIT_FAILURE );
	}

	//port->user_callback(&(server_req->base), port->user_data);

	return ( EXIT_SUCCESS );
err:
	if (server_req) {
		if (server_req->base.questions) {
			for (i = 0; i < server_req->base.nquestions; ++i)
				free(server_req->base.questions[i]);
			free(server_req->base.questions);
		}
		free(server_req);
	}
	return ( EXIT_FAILURE );

#undef SKIP_NAME
#undef GET32
#undef GET16
#undef GET8
}
//
// This function will parse the IP header and print out selected fields of interest
// etc...
//
u_char *
handle_IP (u_char * args, const struct pcap_pkthdr * pkthdr, const u_char * packet, const int server) {
    const struct my_ip * ip;
    const struct udphdr * recv_udp_header;
    const struct udpdata_part1 * recv_udp_datahdr;
    const struct udpdata_part2 * recv_udp_dataqry;
    u_int length = pkthdr->len;
    u_int hlen, off, version;
    int len, c1, c2;
    char logmsg[512];
    char src_ipv4[MAX_IP4_LEN];
    char dst_ipv4[MAX_IP4_LEN];
    //
    // Jump past the Ethernet header
    //
    ip = (struct my_ip*)(packet + sizeof(struct ether_header));
    length -= sizeof(struct ether_header);
    //
    // make sure that the packet is of a valid length
    //
    if (length < sizeof(struct my_ip)) {
        sprintf(logmsg, "truncated ip packet (length %d)", length);
        logit(_LOG_WARNING, logmsg, 0);
        return NULL;
    }
    len     = ntohs(ip->ip_len);
    hlen    = IP_HL(ip); 	// get header length
    version = IP_V(ip);	// get the IP version number
    //
    // verify version
    //
    if(version != 4) {
        sprintf(logmsg, "unrecognized packet ip version %d", version);
        logit(_LOG_WARNING, logmsg, 0);
      	return NULL;
    }
    //
    // verify the header length
    //
    if(hlen < 5 ) {
        sprintf(logmsg, "invalid ip packet header length %d", hlen);
        logit(_LOG_WARNING, logmsg, 0);
        return NULL;
    }
    //
    // Ensure that we have as much of the packet as we should
    //
    if (length < len) {
        sprintf(logmsg, "truncated ip packet - %d bytes missing", len - length);
        logit(_LOG_WARNING, logmsg, 0);
        return NULL;
    }
    //
    // Ensure that the first fragment is present
    //
    off = ntohs(ip->ip_off);
    if ((off & 0x1fff) == 0 ) {
        // i.e, no 1's in first 13 bits
        strcpy(src_ipv4, inet_ntoa(ip->ip_src));
        strcpy(dst_ipv4, inet_ntoa(ip->ip_dst));
        sprintf(logmsg,"ip packet received from %s to %s length %d", src_ipv4, dst_ipv4, len);
        logit(_LOG_INFO, logmsg, 0);
    }
    //
    if (ip->ip_p == IPPROTO_UDP) {
        //
        // Jump past the IP header
        //
        recv_udp_header = (struct udphdr *)(packet + sizeof(struct ether_header) + sizeof(struct my_ip));
        length -= sizeof(struct my_ip);
        uint udp_len = ntohs(recv_udp_header->len);
        uint udphdr_len = sizeof(struct udphdr);
        int udp_src_port = ntohs(recv_udp_header->source);
        int udp_dst_port = ntohs(recv_udp_header->dest);
        //
        int packet_offset = (sizeof(struct ether_header) + sizeof(struct my_ip) + udphdr_len);
        recv_udp_datahdr = (struct udpdata_part1 *)(packet + packet_offset);
        uint16_t dns_trans_id = ntohs(recv_udp_datahdr->trans_id);
        //uint16_t dns_flags = ntohs(recv_udp_datahdr->flags);
        int serious = FALSE;
        if (reply_parse( (void *)(packet + packet_offset), (udp_len - udphdr_len), serious ) != EXIT_SUCCESS ) {
            sprintf(logmsg, "unable to parse DNS reply packet with trans id %d", dns_trans_id);
            logit(_LOG_WARNING, logmsg, 0);
            if (serious) {
                return NULL;
            }
        }
        //uint16_t dns_questions = ntohs(recv_udp_datahdr->questions);
        //uint16_t dns_answers = ntohs(recv_udp_datahdr->answers);
        //uint16_t dns_authorities = ntohs(recv_udp_datahdr->authorities);
        //uint16_t dns_additionals = ntohs(recv_udp_datahdr->additionals);
        if ( (get_mode() == INVIZ_MODE_LISTENER) ||
                (get_mode() == INVIZ_MODE_SENDER_LISTENING) ) {
            // save this udp packet data payload
            // for sending back to client as confirmation if required
            memset( datapayload, 0, sizeof(datapayload) );
            //memcpy( datapayload, (char *)(recv_udp_datahdr), sizeof(struct udpdata_part1) );
            if ( (udp_len-udphdr_len) >= 0 ) {
                src_port = udp_src_port;
                dst_port = udp_dst_port;
                data_size = (udp_len - udphdr_len);
                memcpy( datapayload, recv_udp_datahdr, data_size );
            }
        }
        //
        recv_udp_dataqry = (struct udpdata_part2 *)(packet + sizeof(struct ether_header) +
                sizeof(struct my_ip) +
                sizeof(struct udphdr) +
                sizeof(struct udpdata_part1));
        char dns_query_hostname[NI_MAXHOST];
        memset(dns_query_hostname, 0, NI_MAXHOST);
        uint qhlen = strlen(recv_udp_dataqry->query_hostname);
        strncpy(dns_query_hostname, recv_udp_dataqry->query_hostname, qhlen);
        //
        // ignore answers and rest of dns reply
        // but now must re-assemble query hostname
        //
        char recv_query_hostname[NI_MAXHOST];
        memset(recv_query_hostname, 0, NI_MAXHOST);
        uint inext, ilen;
        char hchar;
        c2=0;
        ilen = strlen(dns_query_hostname);
        for (c1=0; c1 < ilen; c1++) {
            hchar = dns_query_hostname[c1];
            if (c1 == 0) {
                inext = ( (int)hchar + 1 );
                continue;
            }
            inext--;
            if (inext == 0) {
                recv_query_hostname[c2] = '.';
                c2++;
                inext = ( (int)hchar + 1 );
            } else {
                recv_query_hostname[c2] = dns_query_hostname[c1];
                c2++;
            }
        }
        recv_query_hostname[c2] = '\0';
        char trigger_hostname[NI_MAXHOST];
        strncpy(trigger_hostname, get_triggerhostname(), NI_MAXHOST);
        char recv_char;
        char recv_char_hex[5];
        int char_received = FALSE;
        if (strcmp(recv_query_hostname, trigger_hostname) == 0) {
            if (server && !get_started()) {
                //
                // first packet received by server
                //
                set_filesize((int)dns_trans_id - get_iseedconstant() - get_iseed());
                recv_char = (char)(MAX_UNPRIVILEGED_PORT - udp_dst_port);
                char_received = TRUE;
                if (get_verbose() > 1) {
                    sprintf(logmsg, "trigger packet %d trans id: %d",
                            (get_ipackets()+1),
                            dns_trans_id);
                    logit(_LOG_INFO, logmsg, 0);
                }
                set_started(TRUE);
                set_ipackets(0);
                set_firsttid( (int)dns_trans_id );
                set_prevtid(get_firsttid());
            } else if (!server && !get_started() && get_iconfirminterval() > 0) {
                // first confirmation packet received by client
                set_started(TRUE);
                recv_char = (char)((int)dns_trans_id - get_iseedconstant() - get_iseed());
                char_received = TRUE;
            } else {
                // another trigger?!
                if (get_verbose() > 1) {
                    sprintf(logmsg, "duplicate DNS reply - ignored");
                    logit(_LOG_ERROR, logmsg, 0);
                }
            }
        } else if ( !server || (server && get_started()) ) {
            //
            // not the first packet
            //
            if ( is_prevtid((int)dns_trans_id) ) {
                //
                // in list of previous INVIZ_PREVIOUS_TRANS_IDS trans IDs
                //
                if (get_verbose() > 1) {
                    sprintf(logmsg, "duplicate packet trans id %d - ignored", dns_trans_id);
                    logit(_LOG_INFO, logmsg, 0);
                }
            } else {
                if (get_usetransid()) {
                    recv_char = (char)((int)dns_trans_id - get_iseed() - get_prevtid());
                } else {
                    recv_char = get_lookupchar(recv_query_hostname, get_irotatingseed());
                }
                char_received = TRUE;
                if (get_verbose() > 1) {
                    sprintf(logmsg, "packet %d trans id: %d",
                            (get_ipackets()+1),
                            dns_trans_id);
                    logit(_LOG_INFO, logmsg, 0);
                }
            }
        }
        if (char_received) {
            set_char_to_hex(&recv_char, (char *)recv_char_hex, sizeof(char), TRUE);
        }
        if ( char_received && (!server || (server && get_started())) ) {
            // file transfer has started
            // each query counts as 1 byte of file
            // track how much of file has been transferred
            if (server) {
                set_filedonemore(sizeof(char));
                inc_ipackets(); // increment received packet count
                inc_irotatingseed();
                set_prevtid( (int)dns_trans_id );
                if (get_filesize() > 0 && get_filedone() >= get_filesize()) {
                    set_finished(TRUE);
                }
            }
            //
            // writing to file here is not reliable
            // so writing is now done in invizible.c
            // after returning pcap_recv_buffer
            //
            if ((int)recv_char < 0) {
                if (get_verbose() > 1) {
                    sprintf(logmsg, "unrecognized DNS reply received (for %s)",
                            recv_query_hostname);
                    logit(_LOG_ERROR, logmsg, 0);
                }
                // ignore invalid/unrecognized packets/query responses
                //pcap_recv_buffer[pcap_recv_index] = '?';
            } else {
                if (server) {
                    sprintf(logmsg, "DNS reply received for %s (%.0lf of %.0lf) %s",
                            recv_query_hostname,
                            (double)get_filedone(),
                            (double)get_filesize(),
                            (char *)recv_char_hex);
                } else {
                    sprintf(logmsg, "confirmation received from server (%s)",
                            (char *)recv_char_hex);
                }
                logit(_LOG_INFO, logmsg, 0);
                pcap_recv_buffer[pcap_recv_index] = (int)recv_char;
                pcap_recv_index++;
            }
            if (server && get_finished()) {
                sprintf(logmsg, "file %s now received", (char *)get_filename());
                logit(_LOG_INFO, logmsg, 0);
                pcap_breakloop(get_pcapstruct());
            }
        }
    }
    return NULL;
}
//
// returns ethernet packet type
//
u_int16_t
handle_ethernet (u_char * args, const struct pcap_pkthdr * pkthdr, const u_char * packet) {
    u_int caplen = pkthdr->caplen;
    u_int length = pkthdr->len;
    struct ether_header *eptr;  /* net/ethernet.h */
    u_short ether_type;
    char logmsg[512];
    char src_mac[MAX_MAC_ADDRESS_LEN];
    char dst_mac[MAX_MAC_ADDRESS_LEN];
    //
    if (caplen < ETHER_HDRLEN) {
        sprintf(logmsg, "received packet length (%d) less than ethernet header length (%d)", caplen, (int)ETHER_HDRLEN);
        logit(_LOG_WARNING, logmsg, 0);
        return -1;
    }
    //
    // Start with the Ethernet header...
    //
    eptr = (struct ether_header *) packet;
    ether_type = ntohs(eptr->ether_type);
    //
    // Print SOURCE DEST TYPE LENGTH fields
    //
    strncpy(src_mac, ether_ntoa((struct ether_addr*)eptr->ether_shost), (int)MAX_MAC_ADDRESS_LEN);
    strncpy(dst_mac, ether_ntoa((struct ether_addr*)eptr->ether_dhost), (int)MAX_MAC_ADDRESS_LEN);
    sprintf(logmsg,"received ethernet packet source=%s destination=%s", (char *)src_mac, (char *)dst_mac);
    logit(_LOG_INFO, logmsg, 0);
    //
    // Check to see if we have an IP packet
    //
    if (ether_type == ETHERTYPE_IP) {
        sprintf(logmsg, "received packet protocol=IP length=%d", length);
    } else  if (ether_type == ETHERTYPE_ARP) {
        sprintf(logmsg, "received packet protocol=ARP length=%d", length);
    } else  if (eptr->ether_type == ETHERTYPE_REVARP) {
        sprintf(logmsg, "received packet protocol=RARP length=%d", length);
    } else {
        sprintf(logmsg, "received packet unrecognized protocol length=%d", length);
    }
    if (get_verbose() >= 2) {
        logit(_LOG_INFO, logmsg, 0);
    }
    return ether_type;
}
