/**
 * @file pkclient.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created May 15, 2011
 * @based on fwknop (Damien S. Stuart <dstuart@dstuart.org>
 * and Michael Rash <mbr@cipherdyne.org>)
 * @modified June 18, 2011
 * @version 0.30
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include "parms.h"
#include "pkcommon.h"
#include "pkclient.h"

#ifdef WIN32
  #include <conio.h>
#else
  #include <termios.h>
#endif

/* Function to generate a header checksum.
*/
unsigned short
chksum_inviz(unsigned short *buf, int nbytes)
{
    unsigned int   sum;
    unsigned short oddbyte;

    sum = 0;
    while (nbytes > 1)
    {
        sum += *buf++;
        nbytes -= 2;
    }

    if (nbytes == 1)
    {
        oddbyte = 0;
        *((unsigned short *) &oddbyte) = *(unsigned short *) buf;
        sum += oddbyte;
    }

    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);

    return (unsigned short) ~sum;
}

/* Function for accepting password input from users
*/
char*
getpasswd_inviz(const char *prompt)
{
    static char     pwbuf[INVIZ_MAX_PW_LEN + 1] = {0};
    char           *ptr;
    int             c;

#ifndef WIN32
    FILE           *fp;
    sigset_t        sig, old_sig;
    struct termios  ts, old_ts;

    if((fp = fopen(ctermid(NULL), "r+")) == NULL)
        return(NULL);

    setbuf(fp, NULL);

    /* Setup blocks for SIGINT and SIGTSTP and save the original signal
     * mask.
    */
    sigemptyset(&sig);
    sigaddset(&sig, SIGINT);
    sigaddset(&sig, SIGTSTP);
    sigprocmask(SIG_BLOCK, &sig, &old_sig);

    /* Save current tty state for later restoration after we disable echo
     * of characters to the tty.
    */
    tcgetattr(fileno(fp), &ts);
    old_ts = ts;
    ts.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
    tcsetattr(fileno(fp), TCSAFLUSH, &ts);

    fputs(prompt, fp);
#endif

    /* Read in the password.
    */
    ptr = pwbuf;

#ifdef WIN32
    _cputs(prompt);
    while((c = _getch()) != '\r')
	{
		/* Handle a backspace without backing up too far.
		*/
		if(c == '\b')
		{
			if(ptr != pwbuf)
				*ptr--;

			continue;
		}

		/* Handle a Ctrl-U to clear the password entry and start over
		 * (like it works under Unix).
	    */
		if(c == 0x15)
		{
			ptr = pwbuf;
			continue;
		}
#else
	while((c = getc(fp)) != EOF && c != '\n') {
#endif
        if(ptr < &pwbuf[INVIZ_MAX_PW_LEN])
            *ptr++ = c;
	}

	/* Null terminate the password.
    */
    *ptr = 0;

#ifndef WIN32
    /* we can go ahead and echo out a newline.
    */
    putc('\n', fp);

    /* Restore our tty state and signal handlers.
    */
    tcsetattr(fileno(fp), TCSAFLUSH, &old_ts);
    sigprocmask(SIG_BLOCK, &old_sig, NULL);

    fclose(fp);
#else
	/* In Windows, it would be a CR-LF
	*/
	_putch('\r');
	_putch('\n');
#endif

    return(pwbuf);
}

/* Function for accepting password input from from a file
*/
char*
getpasswd_file_inviz(const char *pw_file, const char *server_str)
{
    FILE           *pwfile_ptr;
    unsigned int    numLines = 0, i = 0, found_dst;

    static char     pwbuf[INVIZ_MAX_PW_LEN + 1]     = {0};
    char            conf_line_buf[MAX_LINE_LEN] = {0};
    char            tmp_char_buf[MAX_LINE_LEN]  = {0};
    char           *lptr;

    if ((pwfile_ptr = fopen(pw_file, "r")) == NULL)
    {
        fprintf(stderr, "Could not open password file: %s\n", pw_file);
        exit(1);
    }

    while ((fgets(conf_line_buf, MAX_LINE_LEN, pwfile_ptr)) != NULL)
    {
        numLines++;
        conf_line_buf[MAX_LINE_LEN-1] = '\0';
        lptr = conf_line_buf;

        memset(tmp_char_buf, 0x0, MAX_LINE_LEN);

        while (*lptr == ' ' || *lptr == '\t' || *lptr == '=')
            lptr++;

        /* Get past comments and empty lines.
        */
        if (*lptr == '#' || *lptr == '\n' || *lptr == '\r' || *lptr == '\0' || *lptr == ';')
            continue;

        /* Look for a line like "<SPA destination IP>: <password>" - this allows
        * multiple keys to be placed within the same file, and the client will
        * reference the matching one for the SPA server we are contacting
        */
        found_dst = 1;
        for (i=0; i < strlen(server_str); i++)
            if (*lptr++ != server_str[i])
                found_dst = 0;

        if (! found_dst)
            continue;

        if (*lptr == ':')
            lptr++;
        else
            continue;

        /* Skip whitespace until we get to the password
        */
        while (*lptr == ' ' || *lptr == '\t' || *lptr == '=')
            lptr++;

        i = 0;
        while (*lptr != '\0' && *lptr != '\n') {
            pwbuf[i] = *lptr;
            lptr++;
            i++;
        }
        pwbuf[i] = '\0';
    }

    fclose(pwfile_ptr);

    if (pwbuf[0] == '\0') {
        fprintf(stderr, "Could not get password for IP: %s from: %s\n",
            server_str, pw_file);
        exit(1);
    }

    return pwbuf;
}
//
/* Prompt for and receive a user password.
*/
char *
get_user_pw_inviz(int crypt_op) {
    char        *pw_ptr = NULL;
    static char *no_pw  = "";
    char logmsg[512];

    /* First of all if we are using GPG and GPG_AGENT
     * then there is no password to return.
    */
    if(get_usegpg() && (get_usegpg_agent() || (crypt_op == CRYPT_OP_ENCRYPT && get_gpg_signerkey() == NULL))) {
        return(no_pw);
    }

    /* If --get-key file was specified grab the key/password from it.
     * the server IP address should be found on 1 line of the file
     * with the corresponding password beside it in the format
     * IP address:password
     * (see fwknop/client/getpasswd.c)
    */
    if (get_gpg_keyfile() && strlen(get_gpg_keyfile()) > 0) {
        if (get_useip4() && get_serverip4() && (strlen(get_serverip4()) > 0)) {
            pw_ptr = getpasswd_file_inviz(get_gpg_keyfile(), get_serverip4());
        } else if (get_useip6() && get_serverip6() && (strlen(get_serverip6()) > 0)) {
            pw_ptr = getpasswd_file_inviz(get_gpg_keyfile(), get_serverip6());
        } else {
            logit(_LOG_ERROR, "server IPv4 or IPv6 address must be specified to support GPG key file", 0);
            return(no_pw);
        }
    }
    else if (get_usegpg()) {
        if(crypt_op == CRYPT_OP_DECRYPT) {
            pw_ptr = getpasswd_inviz("Enter passphrase for secret key: ");
        } else if(get_gpg_signerkey() && strlen(get_gpg_signerkey()) > 0) {
            pw_ptr = getpasswd_inviz("Enter passphrase for signing: ");
        } else {
            pw_ptr = no_pw;
        }
    } else {
        if(crypt_op == CRYPT_OP_ENCRYPT) {
            pw_ptr = getpasswd_inviz("Enter encryption password: ");
        } else if(crypt_op == CRYPT_OP_DECRYPT) {
            pw_ptr = getpasswd_inviz("Enter decryption password: ");
        } else {
            pw_ptr = getpasswd_inviz("Enter password: ");
        }
    }

    /* Empty password is allowed, NULL password is not.
    */
    if (pw_ptr == NULL) {
        sprintf(logmsg, "no password data received");
        logit(_LOG_ERROR, logmsg, 0);
        return '\0';
    }

    return pw_ptr;
}
//
/* Send the SPA data via UDP packet.
*/
int
send_spa_packet_tcp_or_udp_inviz(char *spa_data, int sd_len, char *destination, struct sockaddr *remote, int specified_socket) {
    int     sock, res=0, error;
    struct  addrinfo *result, *rp, hints;
    char    port_str[MAX_PORT_STR_LEN];
    char    logmsg[512];

    if (get_testing()) {
        sprintf(logmsg, "test mode enabled, SPA packet not actually sent");
        logit(_LOG_INFO, logmsg, 0);
        return res;
    }

    memset(&hints, 0, sizeof(struct addrinfo));

    if (get_useip4()) {
        hints.ai_family = AF_INET;
    } else if (get_useip6()) {
        hints.ai_family = AF_INET6;
    } else {
        hints.ai_family = AF_UNSPEC; /* Allow IPv4 or IPv6 */
    }

    if (get_useudp()) {
        /* Send the SPA data packet via an single UDP packet - this is the
         * most common usage.
        */
        hints.ai_socktype = SOCK_DGRAM;
        hints.ai_protocol = IPPROTO_UDP;
    } else if (get_usetcp()) {
        /* Send the SPA data packet via an established TCP connection.
        */
        hints.ai_socktype = SOCK_STREAM;
        hints.ai_protocol = IPPROTO_TCP;
    } else {
        sprintf(logmsg, "send_spa_packet_tcp_or_udp function requires use of udp or tcp");
        logit(_LOG_ERROR, logmsg, 0);
        return FKO_ERROR_UNSUPPORTED_FEATURE;
    }
    //
    // is socket already ready for use?
    //
    if (specified_socket > 0) {
        sock = specified_socket;
    } else {
        if (get_pkportrandom() && !get_pkport() && get_pkportfirst() && get_pkportlast()) {
            //
            // using random destination ports
            // and needs to be randomized here
            // because get_pkport() is 0
            //
            set_pkport((rand() % (get_pkportlast() - get_pkportfirst())) + get_pkportfirst());
        }
        sprintf(port_str, "%d", get_pkport());
        error = getaddrinfo(destination, port_str, &hints, &result);
        if (error != 0) {
            sprintf(logmsg, "error in send_spa_packet_tcp_or_udp getaddrinfo for server %s address [%s]",
                    get_serverip(), gai_strerror(error));
            logit(_LOG_ERROR, logmsg, 0);
            return FKO_ERROR_UNKNOWN;
        }
        for (rp = result; rp != NULL; rp = rp->ai_next) {
            sock = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
            if (sock < 0) {
                continue;
            }
            if ((error = (connect(sock, rp->ai_addr, rp->ai_addrlen) != -1))) {
                break;  /* made it */
            }

#ifdef WIN32
            closesocket(sock);
#else
            close(sock);
#endif
        }
        if (rp == NULL) {
            perror("send_spa_packet_tcp_or_udp: Could not create socket: ");
            exit(EXIT_FAILURE);
        }
        freeaddrinfo(result);
    }

    if (get_usetcp()) {
        res = send(sock, spa_data, sd_len, 0);
    } else if (get_useudp()) {
        res = sendto( sock, spa_data, sd_len, 0, (const struct sockaddr *)remote, (socklen_t)sizeof(*remote) );;
    }

    if(res < 0) {
        perror("send_spa_packet_tcp_or_udp: write error: ");
    } else if(res != sd_len) {
        //fprintf(stderr, "[#] Warning: bytes sent (%i) not spa data length (%i).\n", res, sd_len);
    }
    //
    // if socket previously specified then do not close it here
    //
    if (specified_socket <= 0) {
#ifdef WIN32
        closesocket(sock);
#else
        close(sock);
#endif
    }

    return(res);
}
//
/* Send the SPA data via raw TCP packet.
*/
// supports either tcp or udp at this time
//
int
send_spa_packet_raw_inviz(char *spa_data, int sd_len, struct sockaddr_in *saddr,
    struct sockaddr_in *daddr, int specified_socket) {
#ifdef WIN32
    fprintf(stderr,
        "send_spa_packet_tcp_raw: raw packets are not yet supported.\n");
    return(-1);
#else
    int sock, res = 0;
    int max_pkt_size = (sd_len + 512);
    //char pkt_data[2048] = {0}; /* Should be enough for our purposes */
    char * pkt_data = malloc(max_pkt_size+1);
    memset(pkt_data, 0, max_pkt_size+1);

    struct fko_iphdr  *iph  = (struct fko_iphdr *) pkt_data;
    struct fko_tcphdr *tcph = (struct fko_tcphdr *) (pkt_data + sizeof (struct fko_iphdr));
    struct fko_udphdr *udph = (struct fko_udphdr *) (pkt_data + sizeof (struct fko_iphdr));

    int hdrlen_tcp_only = sizeof(struct fko_tcphdr);
    int hdrlen_tcp = sizeof(struct fko_iphdr) + hdrlen_tcp_only;
    int hdrlen_udp_only = sizeof(struct fko_udphdr);
    int hdrlen_udp = sizeof(struct fko_iphdr) + hdrlen_udp_only;

    /* Values for setsockopt.
    */
    int         one     = 1;
    const int  *so_val  = &one;

    if (get_testing()) {
        fprintf(stderr, "test mode enabled, SPA packet not actually sent.\n");
        return res;
    }
    if (specified_socket > 0) {
        sock = specified_socket;
    } else {
        sock = socket (PF_INET, SOCK_RAW, IPPROTO_RAW);
        if (sock < 0) {
            perror("send_spa_packet_raw: create socket: ");
            return(sock);
        } else {
            set_sockraw(sock);
        }
    }
    //
    /* Put the spa data in place.
    */
    if (get_usetcp()) {
        memcpy((pkt_data + hdrlen_tcp), spa_data, sd_len);
    } else if (get_useudp()) {
        memcpy((pkt_data + hdrlen_udp), spa_data, sd_len);
    }

    /* Construct our own header by filling in the ip/tcp header values,
     * starting with the IP header values.
    */
    iph->ihl        = 5;
    iph->version    = 4;
    iph->tos        = 0;
    /* The value here does not matter */
    iph->id         = random() & 0xffff;
    iph->frag_off   = 0;
    iph->ttl        = IP_TTL_DEFAULT;
    if (get_usetcp()) {
        iph->protocol = IPPROTO_TCP;
        /* Total size is header plus payload */
        iph->tot_len = (hdrlen_tcp + sd_len);
    } else if (get_useudp()) {
        iph->protocol = IPPROTO_UDP;
        /* Total size is header plus payload */
        iph->tot_len = (hdrlen_udp + sd_len);
    }
    iph->check      = 0;
    iph->saddr      = saddr->sin_addr.s_addr;
    iph->daddr      = daddr->sin_addr.s_addr;

    if (get_usetcp()) {
        /* Now the TCP header values.
        */
        if (saddr->sin_port == 0) {
            tcph->source = htons(get_packetsrcport());
        } else {
            tcph->source = saddr->sin_port;
        }
        tcph->dest      = daddr->sin_port;
        tcph->seq       = htonl(1);
        tcph->ack_seq   = 0;
        tcph->doff      = 5;
        tcph->res1      = 0;
        /* TCP flags */
        tcph->fin       = 0;
        tcph->syn       = 1;
        tcph->rst       = 0;
        tcph->psh       = 0;
        tcph->ack       = 0;
        tcph->urg       = 0;

        tcph->res2      = 0;
        tcph->window    = htons(INVIZ_TCP_SYN_WINDOW_SIZE); // 32767?
        tcph->check     = 0;
        tcph->urg_ptr   = 0;
    } else if (get_useudp()) {
        udph->check = 0;
        if (saddr->sin_port == 0) {
            udph->source = htons(get_packetsrcport());
        } else {
            udph->source = saddr->sin_port;
        }
        udph->dest = daddr->sin_port;
        udph->len = htons(hdrlen_udp_only + sd_len);
    }

    /* No we can compute our checksum.
    */
    iph->check = chksum_inviz((unsigned short *)pkt_data, iph->tot_len);

    if (specified_socket <= 0) {
        /* Make sure the kernel knows the header is included in the data so it
         * doesn't try to insert its own header into the packet.
        */
        if (setsockopt (sock, IPPROTO_IP, IP_HDRINCL, so_val, sizeof(one)) < 0) {
            perror("send_spa_packet_tcp_raw: setsockopt HDRINCL: ");
        }
    }
    //
    int duplicate_count = get_duplicates();
    int duplicate_sleep = get_iduplicateinterval();
    while (TRUE) {
        res = sendto(sock, pkt_data, iph->tot_len, 0, (struct sockaddr *)daddr, sizeof(*daddr));
        if (duplicate_count <= 0) {
            break;
        }
        duplicate_count--;
        if (duplicate_sleep) {
            sleep(duplicate_sleep);
        }
    }
    if (pkt_data) {
        free(pkt_data);
        pkt_data = NULL;
    }
    if(res < 0) {
        perror("send_spa_packet_tcp_raw: sendto error: ");
    } else if(res != sd_len + iph->tot_len) {
        /* account for the header ?*/
        //fprintf(stderr, "[#] Warning: bytes sent (%i) not spa data length (%i).\n", res, sd_len);
    }
    if ( (specified_socket <= 0) && (get_sockraw() == 0) ) {
        close(sock);
    }

    return(res); // number of bytes sent

#endif /* !WIN32 */
}

/* Send the SPA data via ICMP packet.
*/
int
send_spa_packet_icmp_inviz(char *spa_data, int sd_len, struct sockaddr_in *saddr, struct sockaddr_in *daddr, int specified_socket) {
#ifdef WIN32
    fprintf(stderr, "send_spa_packet_icmp: raw packets are not yet supported.\n");
    return(-1);
#else
    int res = 0, sock;
    char pkt_data[2048] = {0};

    struct fko_iphdr  *iph    = (struct fko_iphdr *) pkt_data;
    struct fko_icmphdr *icmph = (struct fko_icmphdr *) (pkt_data + sizeof (struct fko_iphdr));

    int hdrlen = sizeof(struct fko_iphdr) + sizeof(struct fko_icmphdr);

    /* Values for setsockopt.
    */
    int         one     = 1;
    const int  *so_val  = &one;

    if (get_testing()) {
        fprintf(stderr, "test mode enabled, SPA packet not actually sent.\n");
        return res;
    }
    if (specified_socket > 0) {
        sock = specified_socket;
    } else {
        sock = socket (PF_INET, SOCK_RAW, IPPROTO_RAW);
        if (sock < 0) {
            perror("send_spa_packet_icmp: create socket: ");
            return(sock);
        }
    }

    /* Put the spa data in place.
    */
    memcpy((pkt_data + hdrlen), spa_data, sd_len);

    /* Construct our own header by filling in the ip/icmp header values,
     * starting with the IP header values.
    */
    iph->ihl        = 5;
    iph->version    = 4;
    iph->tos        = 0;
    /* Total size is header plus payload */
    iph->tot_len    = hdrlen + sd_len;
    /* The value here does not matter */
    iph->id         = random() & 0xffff;
    iph->frag_off   = 0;
    iph->ttl        = IP_TTL_DEFAULT;
    iph->protocol   = IPPROTO_ICMP;
    iph->check      = 0;
    iph->saddr      = saddr->sin_addr.s_addr;
    iph->daddr      = daddr->sin_addr.s_addr;

    /* Now the ICMP header values.
    */
    icmph->type     = ICMP_ECHOREPLY; /* Make it an echo reply */
    icmph->code     = 0;
    icmph->checksum = 0;

    /* No we can compute our checksum.
    */
    iph->check = chksum_inviz((unsigned short *)pkt_data, iph->tot_len);
    icmph->checksum = chksum_inviz((unsigned short *)icmph, sizeof(struct fko_icmphdr) + sd_len);

    if (specified_socket <= 0) {
        /* Make sure the kernel knows the header is included in the data so it
         * doesn't try to insert its own header into the packet.
        */
        if (setsockopt (sock, IPPROTO_IP, IP_HDRINCL, so_val, sizeof(one)) < 0) {
            perror("send_spa_packet_icmp: setsockopt HDRINCL: ");
        }
    }

    res = sendto (sock, pkt_data, iph->tot_len, 0, (struct sockaddr *)daddr, sizeof(*daddr));

    if(res < 0) {
        perror("send_spa_packet_icmp: sendto error: ");
    } else if(res != sd_len + hdrlen) {
        /* account for icmp header */
        //fprintf(stderr, "[#] Warning: bytes sent (%i) not spa data length (%i).\n", res, sd_len);
    }
    if (specified_socket <= 0) {
        close(sock);
    }

    return(res);

#endif /* !WIN32 */
}

/* Send the SPA data packet via an HTTP request
*/
int
send_spa_packet_http_inviz(char *spa_data, int sd_len, char *destination, struct sockaddr *remote, int specified_socket) {
    char http_buf[HTTP_MAX_REQUEST_LEN], *spa_data_copy = NULL;
    char *ndx = get_proxy();
    int  i, proxy_port = 0;
    char logmsg[512];
    if (!get_useragent() || (strlen(get_useragent()) <= 0)) {
        set_useragent(INVIZ_DEFAULT_UA, strlen(INVIZ_DEFAULT_UA));
    }

    spa_data_copy = malloc(sd_len+1);
    if (spa_data_copy == NULL) {
        exit(EXIT_FAILURE);
    }
    memcpy(spa_data_copy, spa_data, sd_len+1);

    /* Change "+" to "-", and "/" to "_" for HTTP requests (the server
     * side will translate these back before decrypting)
    */
    for (i=0; i < sd_len; i++) {
        if (spa_data_copy[i] == '+') {
            spa_data_copy[i] = '-';
        }
        else if (spa_data_copy[i] == '/') {
            spa_data_copy[i] = '_';
        }
    }

    if(!get_usehttpproxy()) {
        snprintf(http_buf, HTTP_MAX_REQUEST_LEN,
            "GET /%s HTTP/1.0\r\nUser-Agent: %s\r\nAccept: */*\r\n"
            "Host: %s\r\nConnection: Keep-Alive\r\n\r\n",
            spa_data_copy,
            (char *)get_useragent(),
            (char *)get_serverip()  /* hostname or IP */
        );
    } else {
        /* we are sending the SPA packet through an HTTP proxy */
        /* Extract the hostname if it was specified as a URL. Actually,
         * we just move the start of the hostname to the begining of the
         * original string.
        */
        if(strncasecmp(ndx, "http://", 7) == 0)
            memmove(ndx, ndx+7, strlen(ndx)+1);

        /* If there is a colon assume the proxy hostame or IP is on the left
         * and the proxy port is on the right. So we make the : a \0 and
         * extract the port value.
        */
        ndx = strchr((char *)get_proxy(), ':');
        if(ndx) {
            *ndx = '\0';
            proxy_port = atoi(ndx+1);
        }

        /* If we have a valid port value, use it.
        */
        if(proxy_port) {
            set_pkport(proxy_port);
        }

        snprintf(http_buf, HTTP_MAX_REQUEST_LEN,
            "GET http://%s/%s HTTP/1.0\r\nUser-Agent: %s\r\nAccept: */*\r\n"
            "Host: %s\r\nConnection: Keep-Alive\r\n\r\n",
            (char *)get_serverip(),
            spa_data_copy,
            (char *)get_useragent(),
            (char *)get_proxy()  /* hostname or IP */
        );
        //strlcpy(get_serverip(), get_proxy(), MAX_SERVER_STR_LEN);
        if (get_useip4()) {
            set_serverip4(get_proxy(), MAX_SERVER_STR_LEN);
        } else {
            set_serverip6(get_proxy(), MAX_SERVER_STR_LEN);
        }
    }
    free(spa_data_copy);

    if (get_testing()) {
        if (get_verbose()) {
           sprintf(logmsg, "testing only: SPA packet (%s)", http_buf);
           logit(_LOG_INFO, logmsg, 0);
        }

        logit(_LOG_INFO, "Test mode enabled, SPA packet not actually sent", 0);
        return 0;
    }

    return send_spa_packet_tcp_or_udp_inviz(http_buf, strlen(http_buf), destination, (struct sockaddr *)remote, specified_socket);
}

/* Function used to send the SPA data.
*/
int
send_spa_packet_inviz(fko_ctx_t ctx, char * destination, struct sockaddr *remote, int specified_socket) {
    int                 res, sd_len;
    char               *spa_data;

    struct sockaddr_in  saddr, daddr;

#ifdef WIN32
    WSADATA wsa_data;
#endif

    /* Get our spa data here.
    */
    res = fko_get_spa_data(ctx, &spa_data);

    if(res != FKO_SUCCESS) {
        fprintf(stderr, "send_spa_packet: Error #%i from fko_get_spa_data: %s\n", res, fko_errstr(res));
        return(-1);
    }

    sd_len = strlen(spa_data);

#ifdef WIN32
    /* Winsock needs to be initialized...
    */
    res = WSAStartup( MAKEWORD(1,1), &wsa_data );
    if( res != 0 )
    {
        fprintf(stderr, "Winsock initialization error %d\n", res );
        return(-1);
    }
#endif

    errno = 0;

    if (get_usetcp() || get_useudp()) {
        res = send_spa_packet_tcp_or_udp_inviz(spa_data, sd_len, destination, (struct sockaddr *)remote, specified_socket);
    } else if (get_usehttpproxy()) {
        res = send_spa_packet_http_inviz(spa_data, sd_len, destination, (struct sockaddr *)remote, specified_socket);
    } else if (get_useraw() || get_useicmp()) {
        memset(&saddr, 0, sizeof(saddr));
        memset(&daddr, 0, sizeof(daddr));

        saddr.sin_family = AF_INET;
        daddr.sin_family = AF_INET;

        /* Set source address and port
        */
        if (get_srcport() > 0) {
            saddr.sin_port = htons(get_srcport());
        } else {
            saddr.sin_port = INADDR_ANY;  /* default */
        }

        if (get_spoofip() && (strlen(get_spoofip()) > 0)) {
            saddr.sin_addr.s_addr = inet_addr((char *)get_spoofip());
        } else
            saddr.sin_addr.s_addr = INADDR_ANY;  /* default */

        if (saddr.sin_addr.s_addr == -1) {
            fprintf(stderr, "Could not set source IP.\n");
            exit(EXIT_FAILURE);
        }

        /* Set destination address and port
        */
        daddr.sin_port = htons(get_pkport());
        daddr.sin_addr.s_addr = inet_addr((char *)get_serverip());

        if (daddr.sin_addr.s_addr == -1) {
            fprintf(stderr, "Could not set destination IP.\n");
            exit(EXIT_FAILURE);
        }

        if (get_useraw()) {
            res = send_spa_packet_raw_inviz(spa_data, sd_len, &saddr, &daddr, specified_socket);
        } else {
            res = send_spa_packet_icmp_inviz(spa_data, sd_len, &saddr, &daddr, specified_socket);
        }
    } else {
        /* --DSS XXX: What to we really want to do here? */
        logit(_LOG_ERROR, "unknown protocol in send_spa_packet", 0);
        res = -1;
    }

    return res;
}

/* Function to write SPA packet data to the filesystem
*/
int
write_spa_packet_data_inviz(fko_ctx_t ctx) {
    FILE   *fp;
    char   *spa_data;
    int     res;
    char    logmsg[512];

    res = fko_get_spa_data(ctx, &spa_data);

    if(res != FKO_SUCCESS) {
        fprintf(stderr, "write_spa_packet_data: Error #%i from fko_get_spa_data: %s\n", res, fko_errstr(res));
        return(-1);
    }

    if (get_pksavefilename() && (strlen(get_pksavefilename()) > 0) && get_append()) {
        fp = fopen(get_pksavefilename(), "a");
    } else {
        unlink(get_pksavefilename());
        fp = fopen(get_pksavefilename(), "w");
    }

    if(fp == NULL) {
        sprintf(logmsg, "unable to open pkfilename %s write_spa_packet_data [%s]", (char *)get_pksavefilename(), strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        return(-1);
    }

    fprintf(fp, "%s\n", (spa_data == NULL) ? "<NULL>" : spa_data);

    fclose(fp);

    return(0);
}
//
// based on resolve_ip_http() from fwknop/client
//
int
resolve_ip_http_inviz(char * resolve_host, char * resolve_url) {
    int     sock, res, error, http_buf_len, i;
    int     o1, o2, o3, o4;
    struct  addrinfo *result, *rp, hints;
    char    http_buf[HTTP_MAX_REQUEST_LEN];
    char    http_response[HTTP_MAX_RESPONSE_LEN] = {0};
    char   *ndx;
    char logmsg[512];
    char public_ip4[MAX_IP4_LEN];
    char public_ip6[MAX_IP6_LEN];
    //
    memset(public_ip4, 0, MAX_IP4_LEN);
    memset(public_ip6, 0, MAX_IP6_LEN);
    //
#ifdef WIN32
    WSADATA wsa_data;

    /* Winsock needs to be initialized...
    */
    res = WSAStartup( MAKEWORD(1,1), &wsa_data );
    if( res != 0 )
    {
        fprintf(stderr, "Winsock initialization error %d\n", res );
        return(-1);
    }
#endif
    //
    /* Build our HTTP request to resolve the external IP (this is similar to
     * to contacting whatismyip.org, but using a different URL).
    */
    snprintf(http_buf, HTTP_MAX_REQUEST_LEN,
        "GET %s HTTP/1.0\r\nUser-Agent: %s\r\nAccept: */*\r\n"
        "Host: %s\r\nConnection: Keep-Alive\r\n\r\n",
        (char *)resolve_url, (char *)get_useragent(), (char *)resolve_host);
    //
    http_buf_len = strlen(http_buf);
    //
    memset(&hints, 0, sizeof(struct addrinfo));
    //
    if (get_useip4()) {
        hints.ai_family = AF_INET;
    } else if (get_useip6()) {
        hints.ai_family = AF_INET6;
    } else {
        hints.ai_family = AF_UNSPEC; /* Allow IPv4 or IPv6 */
    }
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    //
    error = getaddrinfo((char *)resolve_host, "80", &hints, &result);
    if (error != 0) {
        fprintf(stderr, "error in getaddrinfo: %s\n", gai_strerror(error));
        return(-1);
    }
    //
    for (rp = result; rp != NULL; rp = rp->ai_next) {
        sock = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
        if (sock < 0) {
            continue;
        }
        if (get_verbose()) {
            sprintf(logmsg, "attempting to get external address via %s ...", (char *)resolve_host);
            logit(_LOG_INFO, logmsg, 0);
        }
        if ((error = (connect(sock, rp->ai_addr, rp->ai_addrlen) != -1))) {
            break;  /* made it */
        }
        //
#ifdef WIN32
        closesocket(sock);
#else
        close(sock);
#endif
    }
    //
    if (rp == NULL) {
        perror("resolve_ip_http: Could not create socket: ");
        return(-1);
    }

    freeaddrinfo(result);

    res = send(sock, http_buf, http_buf_len, 0);

    if(res < 0) {
        perror("resolve_ip_http: write error: ");
    } else if(res != http_buf_len) {
        fprintf(stderr,
            "[#] Warning: bytes sent (%i) not spa data length (%i).\n",
            res, http_buf_len
        );
    }
    //
    res = recv(sock, http_response, HTTP_MAX_RESPONSE_LEN, 0);
    http_response[HTTP_MAX_RESPONSE_LEN-1] = '\0';
    //
#ifdef WIN32
    closesocket(sock);
#else
    close(sock);
#endif
    //
    /* Move to the end of the HTTP header and to the start of the content.
    */
    ndx = strstr(http_response, "\r\n\r\n");
    if(ndx == NULL) {
        sprintf(logmsg, "unrecognized HTTP header response (from http://%s%s)", (char *)resolve_host, (char *)resolve_url);
        logit(_LOG_ERROR, logmsg, 0);
        return(-1);
    }
    ndx += 4;
    //
    /* Walk along the content to try to find the end of the IP address.
     * Note: We are expecting the content to be just an IP address
     *       (possibly followed by whitespace or other not-digit value).
    */
    for(i=0; i<MAX_IP_STR_LEN; i++) {
        if(! isdigit(*(ndx+i)) && *(ndx+i) != '.') {
            break;
        }
    }
    //
    /* Terminate at the first non-digit and non-dot.
    */
    *(ndx+i) = '\0';
    //
    /* Now that we have what we think is an IP address string.  We make
     * sure the format and values are sane.
    */
    if((sscanf(ndx, "%u.%u.%u.%u", &o1, &o2, &o3, &o4)) == 4
        && o1 >= 0 && o1 <= 255
	&& o2 >= 0 && o2 <= 255
	&& o3 >= 0 && o3 <= 255
	&& o4 >= 0 && o4 <= 255) {
        strlcpy(public_ip4, ndx, MAX_IP_STR_LEN);
        //
	sprintf(logmsg, "resolved external/public IP as %s (via http://%s%s)",
                public_ip4, (char *)resolve_host, (char *)resolve_url);
        logit(_LOG_INFO, logmsg, 0);
        set_externalip4(public_ip4, strlen(public_ip4));
        return(0);
    } else {
        sprintf(logmsg, "unrecognized IP address (%s) in HTTP response (from http://%s%s)",
                ndx, (char *)resolve_host, (char *)resolve_url);
        logit(_LOG_ERROR, logmsg, 0);
        return(-1);
    }
}
