/**
 * @file pkserver.h
 * @project invizible
 * @author Wesley Kenzie (wkenzie@gmail.com PGP 4c945266)
 * @created May 23, 2011
 * @based on fwknopd (Damien S. Stuart <dstuart@dstuart.org> and Michael Rash <mbr@cipherdyne.org>)
 * @modified June 18, 2011
 * @version 0.33
 * @copyright 2011. Arthur (Wesley) Kenzie. All Rights Reserved.
 * @license 3-clause BSD License. See accompanying LICENSE file.
 */

#ifndef _PKSERVER_H
#define	_PKSERVER_H

#ifdef	__cplusplus
extern "C" {
#endif

#include "pkcommon.h"
#include "fwknop/netinet_common.h"

#include <string.h>
#include <arpa/inet.h>
#include <signal.h>
#include <getopt.h>
#include <sys/wait.h>

#if HAVE_SYS_SOCKET_H
  #include <sys/socket.h>
#endif

#if HAVE_LIBPCAP
    #ifndef LIB_PCAP_INCLUDED
    #include "pcap/pcap-int.h"
    #include "pcap/pcap.h"
    #define LIB_PCAP_INCLUDED
    #endif
#endif
//
/* My Name and Version
*/
#define MY_SPA_SERVER_NAME     "fwknopd"
#define MY_SPA_SERVER_DESC     "Single Packet Authorization server"

/* Get our program version from VERSION (defined in config.h).
*/
#define MY_SPA_SERVER_VERSION VERSION

/* Some program defaults.
*/
#ifndef DEF_CONF_DIR
  /* Our default config directory is based on SYSCONFDIR as set by the
   * configure script.
  */
  #define DEF_CONF_DIR      SYSCONFDIR"/"PACKAGE_NAME
#endif

#define DEF_SPA_SERVER_CONFIG_FILE     DEF_CONF_DIR"/"MY_SPA_SERVER_NAME".conf"
#define DEF_SPA_SERVER_ACCESS_FILE     DEF_CONF_DIR"/access.conf"

#ifndef DEF_RUN_DIR
  /* Our default run directory is based on LOCALSTATEDIR as set by the
   * configure script. This is where we put the PID and digest cache files.
  */
  #define DEF_RUN_DIR       SYSRUNDIR"/run/"PACKAGE_NAME
#endif

/* More Conf defaults
*/
#define DEF_PID_FILENAME                MY_SPA_SERVER_NAME".pid"
#define DEF_DIGEST_CACHE_FILENAME       "digest.cache"

#define DEF_INTERFACE                   "eth0"
#define DEF_ENABLE_PCAP_PROMISC         "N"
#define DEF_PCAP_FILTER                 "udp port 62201"
#define DEF_ENABLE_SPA_PACKET_AGING     "Y"
#define DEF_MAX_SPA_PACKET_AGE          "120"
#define DEF_ENABLE_DIGEST_PERSISTENCE   "N"
#define DEF_MAX_SNIFF_BYTES             "1500"
#define DEF_GPG_HOME_DIR                "/root/.gnupg"
#define DEF_ENABLE_SPA_OVER_HTTP        "N"
#define DEF_ENABLE_TCP_SERVER           "N"
#define DEF_TCPSERV_PORT                "62201"
#define DEF_SYSLOG_IDENTITY             MY_SPA_SERVER_NAME
#define DEF_SYSLOG_FACILITY             "LOG_DAEMON"

#define DEF_FW_ACCESS_TIMEOUT           30
#define MAX_FW_COMMAND_ARGS_LEN     256

struct fw_config fwc;

/* Some convenience macros */

/* Characters allowed between a config parameter and its value.
*/
#define IS_CONFIG_PARAM_DELIMITER(x) (x == ' ' || x == '\t' || x == '=');

/* End of line characters.
*/
#define IS_LINE_END(x) (x == '\n' || x == '\r' || x == ';');

/* Characters in the first position of a line that make it considered
 * empty or otherwise non-interesting (like a comment).
*/
#define IS_EMPTY_LINE(x) ( \
    x == '#' || x == '\n' || x == '\r' || x == ';' || x == '\0' \
)

/* String compare macro.
*/
#define CONF_VAR_IS(n, v) (strcmp(n, v) == 0)

/* Long options values (for those that may not have a short option).
*/
enum {
    FW_LIST         = 0x200,
    GPG_HOME_DIR,
    ROTATE_DIGEST_CACHE,
    NOOP /* Just to be a marker for the end */
};

extern sig_atomic_t got_signal;

extern sig_atomic_t got_sighup;
extern sig_atomic_t got_sigint;
extern sig_atomic_t got_sigterm;
extern sig_atomic_t got_sigusr1;
extern sig_atomic_t got_sigusr2;
extern sig_atomic_t got_sigchld;
extern sig_atomic_t got_sigalrm;

/* fwknopd-specific limits
*/
#define MAX_PCAP_FILTER_LEN 1024
#ifndef MAX_IFNAME_LEN
#define MAX_IFNAME_LEN      128
#endif
#define MAX_SPA_PACKET_LEN  1500 /* --DSS check this? */

#ifndef MAX_HOSTNAME_LEN
#define MAX_HOSTNAME_LEN    NI_MAXHOST
#endif

/* The minimum possible valid SPA data size.
*/
#define MIN_SPA_DATA_SIZE   140

#define IPV4_VER_MASK   0x15
#define MIN_IPV4_WORDS  0x05

#ifndef ETHER_CRC_LEN
  #define ETHER_CRC_LEN 4
#endif
#ifndef ETHER_HDR_LEN
  #define ETHER_HDR_LEN 14
#endif

#define  CTX_DUMP_BUFSIZE   4096
//
/* How many consecutive pcap capture errors will we allow
 * before giving up and bailing out.
*/
#define MAX_PCAP_ERRORS_BEFORE_BAIL 100
//
/* We normally want pcap in non-blockinbg mode, but this seems to be
 * broken on FreeBSD 7 (at least my test host), so we'll set the default
 * mode to on unless it is a FreeBSD system. --DSS XXX: What we really need
 * to do is figure out what the difference is and address it correctly.
*/
#if defined(__FreeBSD__)
    #define DEF_PCAP_NONBLOCK 0
#else
    #define DEF_PCAP_NONBLOCK 1
#endif
//

/* A simple linked list of uints for the access stanza items that allow
 * multiple comma-separated entries.
*/
typedef struct acc_int_list
{
    unsigned int        maddr;
    unsigned int        mask;
    struct acc_int_list *next;
} acc_int_list_t;

/* A simple linked list of proto and ports for the access stanza items that
 * allow multiple comma-separated entries.
*/
typedef struct acc_port_list
{
    unsigned int            proto;
    unsigned int            port;
    struct acc_port_list    *next;
} acc_port_list_t;

/* A simple linked list of strings for the access stanza items that
 * allow multiple comma-separated entries.
*/
typedef struct acc_string_list
{
    char                    *str;
    struct acc_string_list  *next;
} acc_string_list_t;

/* Access stanza list struct.
*/
typedef struct acc_stanza
{
    char                *source;
    acc_int_list_t      *source_list;
    char                *open_ports;
    acc_port_list_t     *oport_list;
    char                *restrict_ports;
    acc_port_list_t     *rport_list;
    char                *key;
    int                 fw_access_timeout;
    unsigned char       enable_cmd_exec;
    char                *cmd_exec_user;
    uid_t               cmd_exec_uid;
    char                *require_username;
    unsigned char       require_source_address;
    char                *gpg_home_dir;
    char                *gpg_decrypt_id;
    char                *gpg_decrypt_pw;
    unsigned char       gpg_require_sig;
    unsigned char       gpg_ignore_sig_error;
    char                *gpg_remote_id;
    acc_string_list_t   *gpg_remote_id_list;
    struct acc_stanza   *next;
} acc_stanza_t;
//
#define SNAT_TARGET_BUFSIZE         64
#define MAX_FW_COMMAND_ARGS_LEN     256
//
/* Firewall-related data and types. */
//
#if FIREWALL_IPTABLES
  /* --DSS XXX: These are arbitrary. We should determine appropriate values.
  */
  #define MAX_TABLE_NAME_LEN      64
  #define MAX_CHAIN_NAME_LEN      64
  #define MAX_TARGET_NAME_LEN     64
  //
#define FW_CHAIN_TARGET 0
#define FW_CHAIN_TABLE 1
#define FW_CHAIN_FROM_CHAIN 2
#define FW_CHAIN_JUMP_RULE_POS 3
#define FW_CHAIN_TO_CHAIN 4
#define FW_CHAIN_RULE_POS 5
#define FW_NUM_CHAIN_FIELDS 6
  //
  #define DEF_FLUSH_IPT_AT_INIT     "Y"
  #define DEF_FLUSH_IPT_AT_EXIT     "Y"
  #define DEF_ENABLE_IPT_FORWARDING "N"
  #define DEF_ENABLE_IPT_LOCAL_NAT  "Y"
  #define DEF_ENABLE_IPT_SNAT       "N"
  #define DEF_ENABLE_IPT_OUTPUT     "N"
  //
  // each of the following default iptables rules
  // must have FW_NUM_CHAIN_FIELDS in total
  // separated by commas
  //
  #define DEF_IPT_INPUT_ACCESS      "ACCEPT, filter, INPUT, 1, INVIZIBLE_INPUT, 1"
  #define DEF_IPT_OUTPUT_ACCESS     "ACCEPT, filter, OUTPUT, 1, INVIZIBLE_OUTPUT, 1"
  #define DEF_IPT_FORWARD_ACCESS    "ACCEPT, filter, FORWARD, 1, INVIZIBLE_FORWARD, 1"
  #define DEF_IPT_DNAT_ACCESS       "DNAT, nat, PREROUTING, 1, INVIZIBLE_PREROUTING, 1"
  #define DEF_IPT_SNAT_ACCESS       "SNAT, nat, POSTROUTING, 1, INVIZIBLE_POSTROUTING, 1"
  #define DEF_IPT_MASQUERADE_ACCESS "MASQUERADE, nat, POSTROUTING, 1, INVIZIBLE_POSTROUTING, 1"
//
/* iptables command args
*/
#define IPT_ADD_RULE_ARGS "-t %s -A %s -p %i -s %s --dport %i -m comment --comment _exp_%u -j %s 2>&1"
#define IPT_ADD_OUT_RULE_ARGS "-t %s -A %s -p %i -d %s --sport %i -m comment --comment _exp_%u -j %s 2>&1"
#define IPT_ADD_FWD_RULE_ARGS "-t %s -A %s -p %i -s %s -d %s --dport %i -m comment --comment _exp_%u -j %s 2>&1"
#define IPT_ADD_DNAT_RULE_ARGS "-t %s -A %s -p %i -s %s --dport %i -m comment --comment _exp_%u -j %s --to-destination %s:%i 2>&1"
#define IPT_ADD_SNAT_RULE_ARGS "-t %s -A %s -p %i -d %s --dport %i -m comment --comment _exp_%u -j %s %s 2>&1"
#define IPT_DEL_RULE_ARGS "-t %s -D %s %i 2>&1"
#define IPT_NEW_CHAIN_ARGS "-t %s -N %s 2>&1"
#define IPT_FLUSH_CHAIN_ARGS "-t %s -F %s 2>&1"
#define IPT_DEL_CHAIN_ARGS "-t %s -X %s 2>&1"
#define IPT_ADD_JUMP_RULE_ARGS "-t %s -I %s %i -j %s 2>&1"
#define IPT_LIST_RULES_ARGS "-t %s -L %s --line-numbers -n 2>&1"

  /* Structure to define an fwknop firewall chain configuration.
  */
  struct fw_chain {
      int     type;
      char    target[MAX_TARGET_NAME_LEN];
      //int     direction;
      char    table[MAX_TABLE_NAME_LEN];
      char    from_chain[MAX_CHAIN_NAME_LEN];
      int     jump_rule_pos;
      char    to_chain[MAX_CHAIN_NAME_LEN];
      int     rule_pos;
      int     active_rules;
      time_t  next_expire;
  };

//
// following are used as index to IPT (iptables) firewall_access[] values
//
enum {
      IPT_INPUT_ACCESS = 0,
      IPT_OUTPUT_ACCESS,
      IPT_FORWARD_ACCESS,
      IPT_DNAT_ACCESS,
      IPT_SNAT_ACCESS,
      IPT_MASQUERADE_ACCESS,
      NUM_IPT_ACCESS  /* Leave this entry last */
};

  struct fw_config {
      struct fw_chain chain[NUM_IPT_ACCESS];
      char fw_command[MAX_PATH_LEN];
  };

#elif FIREWALL_IPFW

  struct fw_config {
      unsigned short    start_rule_num;
      unsigned short    max_rules;
      unsigned short    active_rules;
      unsigned short    total_rules;
      unsigned short    active_set_num;
      unsigned short    expire_set_num;
      unsigned short    purge_interval;
      unsigned char    *rule_map;
      time_t            next_expire;
      time_t            last_purge;
      char              fw_command[MAX_PATH_LEN];
  };
  //
  #define DEF_IPFW_START_RULE_NUM        "10000"
  #define DEF_IPFW_MAX_RULES             "1000"
  #define DEF_IPFW_ACTIVE_SET_NUM        "1"
  #define DEF_IPFW_EXPIRE_SET_NUM        "2"
  #define DEF_IPFW_EXPIRE_PURGE_INTERVAL "30"
  #define DEF_IPFW_ADD_CHECK_STATE       "N"

#elif FIREWALL_IPF

    /* --DSS Place-holder */

#elif FIREWALL_PF

    /* --DSS Place-holder */

#endif /* FIREWALL type */
//
/* SPA Packet info struct.
*/
typedef struct spa_pkt_info
{
    unsigned int    packet_data_len;
    unsigned int    packet_src_ip;
    unsigned short  packet_dest_port;
    unsigned char   packet_data[MAX_SPA_PACKET_LEN+1];
} spa_pkt_info_t;

/* Struct for (processed and verified) SPA data used by the server.
*/
typedef struct spa_data
{
    char           *username;
    time_t          timestamp;
    char           *version;
    short           message_type;
    char           *spa_message;
    char            spa_message_src_ip[MAX_IP_STR_LEN];
    char            pkt_source_ip[MAX_IP_STR_LEN];
    char            spa_message_remain[1024]; /* --DSS FIXME: arbitrary bounds */
    char           *nat_access;
    char           *server_auth;
    unsigned int    client_timeout;
    unsigned int    fw_access_timeout;
    char            *use_src_ip;
} spa_data_t;

/* SPA message handling status code
*/
enum {
    SPA_MSG_SUCCESS = 0,
    SPA_MSG_BAD_DATA = 0x1000,
    SPA_MSG_LEN_TOO_SMALL,
    SPA_MSG_NOT_SPA_DATA,
    SPA_MSG_HTTP_NOT_ENABLED,
    SPA_MSG_FKO_CTX_ERROR,
    SPA_MSG_DIGEST_ERROR,
    SPA_MSG_DIGEST_CACHE_ERROR,
    SPA_MSG_REPLAY,
    SPA_MSG_TOO_OLD,
    SPA_MSG_ACCESS_DENIED,
    SPA_MSG_COMMAND_ERROR,
    SPA_MSG_INOTIFY_ERROR,
    SPA_MSG_BEACON_ERROR,
    SPA_MSG_WIPE_ERROR,
    SPA_MSG_TRANSFER_ERROR,
    SPA_MSG_KEYLOG_ERROR,
    SPA_MSG_PASS_ERROR,
    SPA_MSG_NOT_SUPPORTED,
    SPA_MSG_ERROR
};

/* Firewall rule processing error codes
*/
enum {
    FW_RULE_SUCCESS = 0,
    FW_RULE_ADD_ERROR = 0x2000,
    FW_RULE_DELETE_ERROR,
    FW_RULE_UNKNOWN_ERROR
};

/* Macro for determining if an error code is a spa_msg handler error
 * and/or a firewall rule processing error.
*/
#define IS_SPA_MSG_ERROR(x) (x & 0x1000)
#define IS_FW_RULE_ERROR(x) (x & 0x2000)
#define IS_FWKNOPD_ERROR(x) (IS_SPA_MSG_ERROR(x) | IS_FW_RULE_ERROR(x))
//
// function prototypes
//
const char* fwknopd_errstr(int);
const char* get_errstr(int);
void fw_config_init_inviz();
void fw_initialize_inviz();
int fw_cleanup_inviz(void);
void check_firewall_rules_inviz();
int process_spa_request_inviz(struct spa_data *);
void add_acc_string_inviz(char **, char *);
void expand_acc_source_inviz(acc_stanza_t *);
void expand_acc_port_list_inviz(acc_port_list_t **, char *);
void free_acc_port_list_inviz(acc_port_list_t *);
int acc_data_is_valid_inviz(acc_stanza_t *);
void parse_access_file_inviz();
acc_stanza_t * acc_check_source_inviz(uint32_t);
int acc_check_port_access_inviz(acc_stanza_t *, char *);
int acc_check_gpg_remote_id_inviz(acc_stanza_t *, char *);
void sig_handler_inviz(int);
int set_sig_handlers_inviz(void);
void hex_dump_inviz(unsigned char *, int);
char* dump_ctx_inviz(fko_ctx_t);
int is_valid_dir_inviz(const char *);
int pcap_capture_inviz(int);
void process_packet_inviz(unsigned char *, const struct pcap_pkthdr *, const unsigned char *);
int process_incoming_pkpacket();
void zero_cmd_buffers_inviz(void);
int jump_rule_exists_inviz(int);
void delete_all_chains_inviz(void);
int create_fw_chains_inviz(void);
void set_fw_chain_conf_inviz(int, char *);
int run_tcp_server_inviz();
void inotify_io_cb(struct ev_loop *, struct ev_io *, int);
void inotify_stat_cb(struct ev_loop *, struct ev_stat *, int);
void do_inotify_watch(int);

#ifdef	__cplusplus
}
#endif

#endif
